require 'bomberman.scripts.base'
require 'base.scripts.shader'	--for BombWarp
require 'base.scripts.lighttex'	--for light on explosion

--[[
I need a custom uniform for this shader ... "starttime" ...
 and it only needs to be set once - when the object is created ...
 in the mean time, just use the circular light tex for our sprite, and just scale it for a sense of distortion
--]]
animLoad('bombwarp', 'lights/highres/circular_intense.bmp')	-- this is in base
local bombWarpShader = renderShader{ fragmentCode = advectionFragmentCode }

animLoad('bomb', 'bomb.png')
animLoad('redbomb', 'redbomb.png')

entclass('Bomb', Base)

Bomb.animGroup = 'bomb'
Bomb.bbox = box2:new(.45)	--block players going down single tiled aisles
Bomb.speed = 10

Bomb.duration = 2
Bomb.chainTime = .2

Bomb.STATE = enum{'IDLE', 'BOOM'}
Bomb.state = Bomb.STATE.IDLE
Bomb.seed = 0

function Bomb:ctor(owner, so)
	Bomb.super.ctor(self)

	owner.bombs:insert(self)		-- keep track of all bombs an owner owns for their shot's sake

	--snap to closest half tile
	self:setPos(so.pos)
	-- spawn params
	self.bombLength = so.bombLength
	self.incinerator = so.incinerator
	self.redbomb = so.redbomb

	if self.redbomb then self:setAnimGroup('redbomb') end

	self.boomTime = getTime() + self.duration
	self.owner = owner

	-- net hack vars (for less ent alloc and hopefully faster render time)
	self.boomDists = vec4:new()
	self.seed = math.random(65535)	-- yeah, dirty hack, dirty dirty hack
end

function Bomb:think()
	Bomb.super.think(self)
	if not self:isActive() then return end

	local t = getTime() - self.boomTime
	if self.state == Bomb.STATE.IDLE then
		self:setScale(vec2:new(math.cos(2*t*math.pi)*.2+.9))
		if t >= 0 then 	-- t minus ... like NASA time!
			self:explode()
		end
	elseif self.state == Bomb.STATE.BOOM then
		-- assert that t>=0.  or just wing it.
		if t > 1 then
			self:delete()
			return
		end
		t = 1
		-- all we do is chill til we die
	else
		assert(false, 'got a bad bomb state: '..self.state)
	end
end

local sides = {vec2:new(0,-1), vec2:new(0,1), vec2:new(-1,0), vec2:new(1,0) }

-- TODO for net optimization:
--  send across the network only the following (no particles)
--	x min, max
--	y min, max
--	pos is useful but it'll already be provided
--  particle seed or something
--  explosion start time
function Bomb:explode()
	-- start at our current position
	-- cycle in all directions
	local pos = vec2:new()
	for side,dir in pairs(sides) do
		pos:set(self:getPos())
		local len = 0
		while true do
			-- if you want, we can trace as far as we can go, wrap a bbox around it, then just do one test to all ents
			-- if there's a guy within .5 linf then KILL THEM
			local hit = false	--bailout on hitting an entity
			for i,e in pairs(bombables) do
				if e ~= self then
					local snappos = vec2:new(e:getPos()):snap()
					-- if e snaps to the same tile that the bomb would then kill it
					if (pos - snappos):lenSq() < .4*.4 then
						if e.hitByBomb then
							e:hitByBomb()
							if e.blocksExplosions then
								hit = true
							end
						end
					end
				end
			end
			-- break here or else we'll place flames past our hit wall
			if not self.redbomb and hit then break end
			-- move
			len = len + 1	-- keep track of how many len increments you make
			if not self.incinerator and len > self.bombLength then
				len = self.bombLength
				break
			end
			pos:addTo(dir)
			-- if the map type is solid then stop
			if mapGetType(unpack(pos)) == MAPTYPE.Solid then break end
		end
		self.boomDists[side] = len
	end

	--remove unique form owner.bombs
	for i,v in ipairs(self.owner.bombs) do
		if v == self then
			self.owner.bombs:remove(i)	--remove operates by index
			break		-- assume it's unique and stop here (otherwise we'd have to mess with the (readonly) iterator)
		end
	end

	self.state = Bomb.STATE.BOOM
	self:setAnimGroup('null')
	self:setNoPhysics('true')	-- no collidable too? or is that innate w/nophys=true?
end

function Bomb:hitByBomb()
	-- give us a .1 second delay to die
	self.boomTime = getTime() + self.chainTime
end

local function clientMakeSpark(t,basepos)
	for i=1,3 do
		local c = math.random()
		local vel = vec2:new(math.random()*2-1, math.random()*2-1)
		-- pos is floor(self.pos) + .5 + dir * dist
		local color = vec4:new(1, c, c * math.random() * math.random(), 1)
		local scale =  math.random() + .5
		local life = math.random() * .5 + .5
		if t <= life then
			local frac = 1 - t / life
			color.w = frac
			local pos = vec2:new(basepos):addTo(vel[1] * t, vel[2] * t)
			-- now draw particle geometry here
			drawSprite(
				pos,
				0,		--angle
				vec2:new(scale,scale),
				color,
				0,		--shader
				BLEND.ADD,	--blend
				false,		--postFBO
				true,	-- lighting
				false,		--left
				'particle',	--group
				'idle',		--seq
				0,		--frame
				Renderable.zBias + Renderable.zScale * pos[2])	--depth
		end
	end
end

local function clientMakePuff(t,basepos)
	for i=1,10 do
		local s = math.random() * 2
		local pos = vec2:new(basepos)
		local vel = vec2:new(math.random()*2-1, math.random()*2-1)
		local scale = vec2:new(s + 1)
		local life = 5 - s
		if t <= life then
			pos:addTo(vel[1] * t, vel[2] * t)
			local frac = t / life
			local alpha = (1-frac)*(1-frac)
			local color = vec4:new(1,1,1,alpha)
			drawSprite(
				pos,
				0,	--angle
				scale,	
				color,
				0,
				BLEND.DECAL,
				false,
				true,	-- lighting
				false,
				'cloud',
				'idle',
				0,
				Renderable.zBias + Renderable.zScale * pos[2])
		end
	end
end

function Bomb.clientUpdate(classObj)
	local state = net.state
	if state == Bomb.STATE.IDLE then
		-- just render the sprite
		Bomb.super.clientUpdate(classObj)
	elseif state == Bomb.STATE.BOOM then
		-- do an explosion!  (do a barrel roll)
		-- draw a bomb warp sprite
		-- and draw some explosions along the tracelines
		local t = getTime() - net.boomTime
		if t < 0 then t = 0
		elseif t > 1 then t = 1
		end
		-- don't forget the seed ... hmm ...
		local pushseed = math.random(65535)		-- I need a seed getter function ...
		math.randomseed(net.seed)

		-- now for all tiles from the origin
		local pos = vec2:new(net.pos)
		clientMakeSpark(t,pos)
		clientMakePuff(t,pos)
		for side,dir in pairs(sides) do
			for dist=1,net.boomDists[side] do
				pos[1] = net.pos[1] + dir[1] * dist
				pos[2] = net.pos[2] + dir[2] * dist	-- offset to our boom position
				clientMakeSpark(t,pos)
			end
		end
		-- while we're here make a puff too
		-- and a bomb warp effect
		local bombWarpTime = .3
		if t <= bombWarpTime then
			drawSprite(
				net.pos,			-- pos
				0,				-- angle
				vec2:new(10*t/bombWarpTime),	-- scale
				{0,0,0,0},		-- color
				bombWarpShader,			-- shader
				BLEND.NONE,			-- blend
				true,				-- postFBO
				true,	-- lighting
				false,				-- left
				'bombwarp',			-- group
				'idle',				-- seq
				0,				-- frame
				classObj.zScale * pos[2] - 10)	-- zBias
		end
		-- and our light
		do 
			local lf = 1 - t
			addLight(
				net.pos,		--pos
				vec4:new(2*lf),		--color
				circularLightTex,	--tex
				box2:new(10))		--bbox
		end
		math.randomseed(pushseed)
	else
		print('Bomb.render got a bad state')
	end
end

Bomb:createNetTable{
	state = 'char';		-- it's really just a flag of whether we're exploding or not ...
	boomDists = 'vec4s';	-- up,down,left,right distances that our bomb explosion has travelled
	boomTime = 'float';	-- explosion time
	seed = 'ushort';
}

