/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <assert.h>
#include <stdio.h>
#include <math.h>	//floor() atleast

#include "system/gl.h"
#include "system/singletons.h"
#include "system/memory.h"

#include "base/base.h"

#include "gc/gc.h"

#include "inspiration/game/main.h"	//game
#include "inspiration/texloader.h"
#include "inspiration/main.h"
#include "inspiration/spawnobj.h"
#include "inspiration/client.h"
#include "inspiration/program.h"
#include "inspiration/world.h"
#include "inspiration/draw.h"
#include "inspiration/entalloc.h"
#include "inspiration/animation.h"
#include "inspiration/blend.h"
#include "inspiration/main.h"
#include "inspiration/maptex.h"

#include "editor.h"
#include "menu.h"
#include "impl.h"
#include "func.h"
#include "wndent.h"
#include "wndresize.h"
#include "wndtag.h"
#include "wndtex.h"
#include "wndtile.h"
#include "input.h"

#ifdef WIN32	//MSVC only
#include <direct.h>
#else
#include <sys/unistd.h>
#endif

#include "menulib.h"

using namespace std;
using namespace pwn;


//***********************
//EDITOR WINDOW FUNCTIONS
//***********************

#define EDITOR_INFL_SCROLL_MIN		0
#define EDITOR_INFL_SCROLL_MAX		5
#define EDITOR_RISE_SCROLL_MIN		0
#define EDITOR_RISE_SCROLL_MAX		5

//static int editor_first_tex_disp = 0;

EditorImpl::EditorImpl()
:	rootmenu(gc::sys->createPtr<Object>()),
	active(0),
	method(-1),
	paintflags(0),
	snapEnabled(1),
	snapDist(0.5f),
	hideEnts(0),
	layer(0),
	tagKey(0),
	tagType(0),
	tex(0),
	mapKey(0),
	movePreserve(1),
	spawnObj(NULL),
	spawnPair(-1),
	paintArea(1),
	paintDissolve(1.f),
	drawDepth(-10),
	sortDepth(-1999),
	func(this),
	menu(this),
	tagWnd(this),
	texWnd(this),
	tileWnd(this),
	input(this)
{
	for (int i = 0; i < numberof(editWnd); i++) {
		editWnd[i] = new gc::GlobalPointer<MenuEditorWindow>(gc::sys->createPtr<MenuEditorWindow>());
	}
}

EditorImpl::~EditorImpl() {
	for (int i = 0; i < numberof(editWnd); i++) {
		delete editWnd[i];
	}
}

//TODO - time to make selects like html
//give them name->value mappings instead of just name mappings ...
void EditorImpl::refreshTextureSelect(pwn::Select *m) {
	int i, count;
	count = world->texs->texs.size();
	m->resetContent();
	for (i = 0; i < count; i++) {
		m->addString(world->texs->texs[i]->texName.c_str());
	}
	//this value is typically -1
	//is the texture setter capable of transforming all OOB values into null textures?
	m->addString("<none>");
}

// cant think of anywhere else good to put this
//todo - make a util file?

void EditorImpl::trimFloatString(char *string, int length) {
	bool decimal_found = false;
	char *c;
	for (c = string; *c && (c - string) < length; c++) {
		if (*c == '.') decimal_found = true;
	}

	if (decimal_found) {
		for (c--; c >= string && *c == '0'; c--) {
			*c = '\0';
		}
		if (*c == '.') *c = '\0';
	}
}

void EditorImpl::setFloatTextFields(const float *fv, TextField **mv, int count) {
	char buffer[256];
	for (; count>0; count--, fv++, mv++) {
		sprintf(buffer, "%f", *fv);
		trimFloatString(buffer, sizeof(buffer));
		mv[0]->setText(buffer);
	}
}

bool EditorImpl::nameIsValid(const char *buffer) {
	//no empty strings!
	if (buffer[0] == 0) return false;
	return true;
}

//these functions reference stuff below - therefore a potential cause of lockups
void EditorImpl::refreshAllTextures() {
	texWnd.refreshAllTextures();	//when the texture list gets changed
	tagWnd.refreshAllTextures();
	tileWnd.refreshAllTextures();
}

void EditorImpl::setMethod(int newmethod) {

	if (method == newmethod) return;
	method = newmethod;

	//if we switch to ENT_SELECT mode then deselect all entities
	//we've been having some problems with lots of random ents being selected
	if (method == EDITOR_ENT_SELECT) {
		for (int i = 0; i < spawnObjSys.count(); i++) {
			SpawnObj *s = spawnObjSys[i];
			assert(s);
			s->sel = false;
		}
	}
}

void EditorImpl::refreshAll() {
	editorWndEnt.refresh();
	tileWnd.refresh();
	tagWnd.refresh();
	texWnd.refresh();
}

void EditorImpl::deleteEntity() {
	//go backwards so, as we remove, it will not copy the next obj back past our index
	for (int i = spawnObjSys.count()-1; i >= 0;i--) {
		SpawnObj *s = spawnObjSys[i];
		assert(s);
		if (!s->sel) continue;

		//kill any assoc ents
		if (s->ent.i != -1) game.deleteEntity(s->ent);

		int nextIndex = spawnObjSys.removeIndex(i);
		if (s == spawnObj) {
			spawnObj = spawnObjSys[nextIndex];
		}
	}

	spawnType = "";
	if (spawnObj) spawnObj->rw("classname") >> spawnType;

	editorWndEnt.refreshSelected();
}

//*********************************
// NON-WINDOWS EDITOR FUNCTIONALITY 
//*********************************

bool EditorImpl::sysInit() {
//	landscapeHistory.Create();

	//create the root menu from which all other editors feed off of
	rootmenu = MenuEditorRootSys::create(this);
	if (!rootmenu) return false;
	assert(menuSys.getRoot());
	menuSys.getRoot()->addChild(rootmenu);

	editorWndEnt.init();
	tileWnd.init();
	texWnd.init();
	menu.init();
	//used to be modalless
	tagWnd.init();
	resizeWnd.init(this);

	//default
	//setMethod(EDITOR_VERTEX_HEIGHT);
	return true;
}

void EditorImpl::sysKill() {
	//deleting rootmenu will cascade and del all child menus
	if (rootmenu) rootmenu->unlinkFromParent();
	rootmenu = NULL;

	//shutdown history
//	landscapeHistory.Destroy();	
}

		//the editor input interface

#define EDITOR_SPAWNOBJ_SIZE	1.f

vec2f EditorImpl::getSpawnObjPos(SpawnObj *id) {
	assert(id);
	vec2f dest;
	if (!(id->rw("pos") >> dest)) {
		con << "failed to get spawnobj's pos!\n";
		dest.x = dest.y = (float)spawnObjSys.getIndex(id) * (EDITOR_SPAWNOBJ_SIZE * 2);
		id->rw("pos") << dest;
	}
	return dest;
}

int EditorImpl::isActive() { return active; }
												
static vec2f fontSize(.3f);	//for fonts displayed in the map

void EditorImpl::displayText(const char *text, float x, float y, float sx, float sy, int color) {
	int len = strlen(text);
	const char *p = text;
	vec2f s(sx,sy);
	vec2f vtx[4];
	
	glPushMatrix();
	glTranslatef(x, y, sortDepth - .1f);

	//good ol' menu system ...
	//use position 0 since i just translated anyways
	//note that the font y scale isn't being used here.
	menuDisplayText(vec2f(0.f), vec2f(100.f), text, color, vec2f(sx,sy));

	glColor3f(1,1,1);
	glPopMatrix();
}

void EditorImpl::drawTags(const box2f &viewarea) {
	box2f csc = viewarea.scale(vec2f(1.f / (float)world->tags->tagsize.x, 1.f / (float)world->tags->tagsize.y));
	box2i b(vec2i(floor_ftoi(csc.min.x), floor_ftoi(csc.min.y)),
			vec2i(floor_ftoi(csc.max.x), floor_ftoi(csc.max.y)));

	box2i tagbox(vec2i(),world->tags->getSize()-1);
	if (!b.touchesE(tagbox)) return;	//in regards to equality also
	b.shrink(tagbox);

	glDisable(GL_TEXTURE_2D);

	glLineWidth(2);
	glBegin(GL_LINES);
	glColor3f(0,1,0);
	{
		for (int x = b.min.x; x <= b.max.x+1; x++) {
			for (int y = b.min.y; y <= b.max.y; y++) {
				if ((*world->tags)(x, y) != (*world->tags)(x-1, y)) {
					//draw a line on the right side of x1 - between x1 and x2
					glVertex3f((float)(x * world->tags->tagsize.x), (float)(y * world->tags->tagsize.y), drawDepth);
					glVertex3f((float)(x * world->tags->tagsize.x), (float)((y+1) * world->tags->tagsize.y), drawDepth);
				}
			}
		}
	}
	{
		for (int y = b.min.y; y <= b.max.y+1; y++) {
			for (int x = b.min.x; x <= b.max.x; x++) {
				if ((*world->tags)(x, y) != (*world->tags)(x, y-1)) {
					//draw a line on the bottom side of y1 - between y1 and y2
					glVertex3f((float)(x * world->tags->tagsize.x), (float)(y * world->tags->tagsize.y), drawDepth);
					glVertex3f((float)((x+1) * world->tags->tagsize.x), (float)(y * world->tags->tagsize.y), drawDepth);
				}
			}
		}
	}
	glEnd();
	glLineWidth(1);
	glEnable(GL_TEXTURE_2D);

	if (method == EDITOR_EDIT_TAGS) {
		for (int y = b.min.y; y <= b.max.y; y++) {
			for (int x = b.min.x; x <= b.max.x; x++) {
				int tag = (*world->tags)(x, y) - MapTagKey::minRange;
				char buffer[255];
				sprintf(buffer, "%.4x", tag);
				displayText(buffer, (x * world->tags->tagsize.x) + 0.5f, (y * world->tags->tagsize.y) + 0.5f, 1.f, 1.f);
			}
		}
	}
}

void EditorImpl::display(const box2f &viewarea) {
	if (!world) return;
	glClear(GL_DEPTH_BUFFER_BIT);

	//window hotkeys
	{
		for (int i = 0; i <= 3; i++) {	//ent,tile,tag,tex -- main needs to be removed?  script needs to be revamped
			if (kbd.keystate[Key::F3 + i] == BUTTON_STATE_PRESS) {
				func.showWindow(i + EW_ENT, -1);
			}
		}
	}

	//this should be placed in a more-appropriately-named 'Update' function... but we dont have one
	if (kbd.keystate[Key::F7] == BUTTON_STATE_PRESS) {
		active = !active;
	}

	//do an editor overlay - for usefulness and such
	if (!active) return;	
	if (hideEnts) return;

	drawTags(viewarea);

	//show all entity spawnpoint instances
	for (int e = 0; e < spawnObjSys.count(); e++) {
		SpawnObj *s = spawnObjSys[e];
		assert(s);
		vec2f v = getSpawnObjPos(s);

//now that i'm generalizing types
//the only for-sure entity pointers i can display
//are those that we already know
#ifdef USE_SPAWN_KEYS
		if (method == EDITOR_ENT_CREATE ||
			method == EDITOR_ENT_SELECT ||
			method == EDITOR_ENT_LINK)
		{
			for (int j = 0; j < s->getPairCount(); j++) {
				ESpawnPair *pair = s->getPair(j);
				assert(pair);
				assert(pair->key);
				if (pair->key->type != VARTYPE_OBJ) continue;
				int destIndex;
				if (!pair->getData(&destIndex)) continue;
				SpawnObj *dest = spawnObjSys.getExt(destIndex);
				if (!dest) continue;
				vec2f dest_v = getSpawnObjPos(dest);

				glDisable(GL_TEXTURE_2D);
				glLineWidth(3);
				glBegin(GL_LINES);
				glColor3f(0.2f, 0.2f, 1.f);
				glVertex3fv(v);
				glColor3f(1.f, 0.2f, 0.2f);
				glVertex3fv(dest_v);
				glEnd();
				glLineWidth(1);

				dest_v = vecLerp(v, dest_v, 0.25f);

				displayText(pair->key->name, dest_v.x, dest_v.y, VEC2ELEM(fontSize));
			}
		}
#endif

		map<string,string>::iterator it;

		//ESpawnType *type = s->getTypeExt();
//		//draw the object's renderer
		//if (type)
		{

			int anim_group = -1;
			
			string group;
			if (s->rw("group") >> group) {
				anim_group = animSys.getGroupIndex(group.c_str());
			} else {
				anim_group = animSys.getGroupIndex(s->getTypeGroup());
			}
			if (anim_group != -1) {

				float angle = 0.f;
				s->rw("angle") >> angle;
				
				vec2f scale(1.f);
				s->rw("scale") >> scale;

				vec4f color(1.f);
				s->rw("color") >> color;

				animSys.render(
					VEC2ELEM(v),
					sortDepth,
					angle,
					VEC2ELEM(scale),
					(vec4uc)(255.f * color),
					0,	//shader
					Blend(),
					true,	//postfbo
					false,	//uselighting
					false,	//flip
					AnimState(anim_group,0,0).getAbsFrame());
			}
		}

		box2f bbox = s->getTypeBBox();
		s->rw("bbox") >> bbox;		//if we get a param override ...

		//draw the bounding box
		{
			glDisable(GL_TEXTURE_2D);

			box2f b(bbox);
			b += v;

			glColor3f(1, !s->sel, 0);
			glLineWidth(2);
			glBegin(GL_LINE_LOOP);
			glVertex3f(b.min.x, b.min.y, drawDepth-.15f);
			glVertex3f(b.min.x, b.max.y, drawDepth-.15f);
			glVertex3f(b.max.x, b.max.y, drawDepth-.15f);
			glVertex3f(b.max.x, b.min.y, drawDepth-.15f);
			glEnd();
			glLineWidth(1);
			glEnable(GL_TEXTURE_2D);
		}

		if (method == EDITOR_ENT_CREATE ||
			method == EDITOR_ENT_SELECT ||
			method == EDITOR_ENT_LINK)
		{
			string classname;
			if (s->rw("classname") >> classname && classname.length()) {
				displayText(classname.c_str(), v.x + bbox.min.x, v.y + bbox.max.y + 0.1f, VEC2ELEM(fontSize));
			}

			string name;
			if (s->rw("name") >> name && name.length()) {
				displayText(name.c_str(), v.x + bbox.min.x, v.y + bbox.max.y + 0.7f, VEC2ELEM(fontSize), 0xFFFFFF7F);
			}
		}
	}
}

void EditorImpl::setMapKeySel(int index) {
	if (index != mapKey) {
		mapKey = index;
		tileWnd.refreshKey();
	}
}

void EditorImpl::gameKill() { refreshAll(); }

bool EditorImpl::sessionInit() { refreshAll(); return true; }
	
void EditorImpl::clearLevelVars() {
	layer = 0;
	tagKey = 0;
	tagType = 0;
	tex = 0;
	mapKey = 0;
	spawnType = "";			//the current selected entity type
	spawnObj = NULL;			//the current selected entity
	spawnPair = -1;
}

void EditorImpl::setWorld(World *world_) {
	world = world_;
	clearLevelVars();
}


bool EditorImpl::getInput() {
	if (!world) return false;
	input.getInput();
	return false;
}

void EditorImpl::updateMouse(float x, float y) { input.updateMouse(vec2f(x,y)); }


Editor::Editor() : impl(new EditorImpl()) {}
Editor::~Editor() { delete impl; }
bool Editor::sysInit() { return impl->sysInit(); }
void Editor::sysKill() { impl->sysKill(); }
void Editor::gameKill() { impl->gameKill(); }
bool Editor::sessionInit() { return impl->sessionInit(); }	
void Editor::display(const box2f &viewarea) { impl->display(viewarea); }
bool Editor::getInput() { return impl->getInput(); }
void Editor::refreshAll() { impl->refreshAll(); }
int Editor::isActive() { return impl->isActive(); }
void Editor::updateMouse(float x, float y) { impl->updateMouse(x,y); }
void Editor::setWorld(World *world) { impl->setWorld(world); }
Editor editor;



//the cvar interface fades more by the day ...

void CFunc_editor_mapkey_sel(int,char**) {
	editor->tileWnd.refreshKey();
}

static CVarw *local_cvars[] = {
	new CVarw(&editor->active,				"editor"),
	
	new CVarw(&editor->method,				"editor_method"),
	new CVarw(&editor->paintflags,			"editor_paintflags"),
	new CVarw(&editor->snapEnabled,			"editor_snap_enable"),
	new CVarw(&editor->snapDist,					"editor_snap"),
	//world vars
	new CVarw(&editor->layer,				"editor_layer"),
	//tag vars
	new CVarw(&editor->tagKey,			"editor_tagkey_sel"),
	new CVarw(&editor->tagType,			"editor_tagtype_sel"),
	//tex vars
	new CVarw(&editor->tex,				"tex"),
	//paint vars
	new CVarw(&editor->mapKey,			"editor_mapkey_sel", CFunc_editor_mapkey_sel),
	new CVarw(&editor->movePreserve,		"editor_move_preserve"),
	//ent vars
//	new CVarw(&editor->spawnType.i,	"editor_spawn_obj_type_sel"),
//	new CVarw(&editor->spawnObj.i,		"editor_spawn_obj_sel"),
	//brush vars
	new CVarw(&editor->paintArea,					"editor_area"),
	new CVarw(&editor->paintDissolve,				"editor_dissolve"),
	new CVarw(&editor->drawDepth,			"drawDepth"),
	new CVarw(&editor->sortDepth,			"sortDepth"),
};
