/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <stdio.h>
#include <iostream>

#if 0	//def WIN32	- haven't got windows open yet?
#include <commctrl.h>
#endif

#include "system/gl.h"
#include "system/memory.h"
#include "system/exception.h"
#include "base/base.h"

#include "inspiration/world.h"
#include "inspiration/maptex.h"

#include "impl.h"
#include "func.h"
#include "wndtile.h"
#include "menulib.h"

using namespace pwn;
using namespace std;

// ************************* MenuTile prototype ************************* 

class MenuTile
:	public MenuEditorWindow,
	public SelectListener,
	public TextListener,
	public CheckboxListener
{
	typedef MenuEditorWindow super;
protected:
	EditorImpl *ed;
public:

	gc::Pointer<Text> surfacePaint, surfaceFill, surfaceMove, surfaceMovePreserve;
	gc::Pointer<Scrollbar> sliderArea, sliderDissolve;
	gc::Pointer<Select> selectKey, selectTexture, selectType, selectBorderTexture;
	gc::Pointer<Object> texPreview;
	gc::Pointer<Object> borderTexPreview;
	gc::Pointer<TextField> shader;
	gc::Pointer<Checkbox> postFBO;
	gc::Pointer<Text> createKey, removeKey;
	gc::Pointer<Text> *sides[4];
	
	gc::Pointer<TextField> *friction[4];
	gc::Pointer<TextField> bounce;
	gc::Pointer<TextField> pull;
	gc::Pointer<TextField> *texgen[6];
	gc::Pointer<TextField> borderThickness, borderUVThickness;
	gc::Pointer<TextField> group, seq;
	gc::Pointer<TextField> *bbox[4];
	gc::Pointer<TextField> zBias;
	gc::Pointer<TextField> *planes[12];	//4 sets of 3 components
	
	MenuTile(EditorImpl *);
	virtual ~MenuTile();
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
	virtual bool display();
	virtual void callback_userSetSelected(Select *, int newSel, int oldSel);
	virtual void callback_onTextChange(TextField *);
	virtual void callback_onSetChecked(Checkbox *m, bool isChecked);
};

class EditorWndTileImpl {
protected:
	EditorImpl *ed;
public:
	gc::GlobalPointer<MenuTile> tileWnd;

	EditorWndTileImpl(EditorImpl *ed_) : tileWnd(gc::sys->createPtr<MenuTile>()), ed(ed_) {}
	
	void refreshTexturePreview();
	void refreshComboTextureCurSel();
	void refreshComboTextureContents();
	void refreshAllTextures();
	void refreshComboMapTypeCurSel();
	void refreshComboMapTypes();
	void refreshComboKeyCurSel();
	void refreshComboKeyContents();
	void refreshSides();
	void refreshKey();
	void refresh();
	void init();
};


//******************
//TILE EDITOR WINDOW
//******************

void EditorWndTileImpl::refreshTexturePreview() {
	MapTex *tex = NULL;
	MapTex *borderTex = NULL;
	if (ed->world->keys->isValid(editor->mapKey)) {
		MapKey *k = (*ed->world->keys)[editor->mapKey];
		if (k) {
			tex = k->tex;
			borderTex = k->borderTex;
		}
	}


	//of course, since the menu system is based upon TexPtr's, we dont have too many options . . .
	const TexLoad *texImg = tex ? texSys.getObj(tex->getCurrentImage()) : NULL;
	tileWnd->texPreview->setBackgroundColor(texImg ? 0xFFFFFFFF : 0);
	tileWnd->texPreview->setTextureID(texImg ? texImg->tex.getTextureID() : 0);

	const TexLoad *borderTexImg = borderTex ? texSys.getObj(borderTex->getCurrentImage()) : NULL;
	tileWnd->borderTexPreview->setBackgroundColor(borderTexImg ? 0xFFFFFFFF : 0);
	tileWnd->borderTexPreview->setTextureID(borderTexImg ? borderTexImg->tex.getTextureID() : 0);
}

void EditorWndTileImpl::refreshComboTextureCurSel() {
	int texIndex = -1;
	int borderTexIndex = -1;
	if (ed->world->keys->isValid(editor->mapKey)) {
		MapKey *k = (*ed->world->keys)[editor->mapKey];
		if (k) {
			texIndex = ed->world->texs->getIndex(k->tex);
			borderTexIndex = ed->world->texs->getIndex(k->borderTex);
		}
	}
	//if this is set then tileWnd is set, but not necessarily vice versa
	tileWnd->selectTexture->setSelected(texIndex);
	tileWnd->selectBorderTexture->setSelected(borderTexIndex);
	refreshTexturePreview();
}

void EditorWndTileImpl::refreshComboTextureContents() {
	editor->refreshTextureSelect(tileWnd->selectTexture);
	editor->refreshTextureSelect(tileWnd->selectBorderTexture);	
	refreshComboTextureCurSel();
}

void EditorWndTileImpl::refreshAllTextures() {
	refreshComboTextureContents();
}

void EditorWndTileImpl::refreshComboMapTypeCurSel() {
	if (!ed->world->keys->isValid(editor->mapKey)) return;
	MapKey *k = (*ed->world->keys)[editor->mapKey];
	if (!k) return;
	//select
	tileWnd->selectType->setSelected(k->type);
}

void EditorWndTileImpl::refreshComboMapTypes() {
	//clear
	tileWnd->selectType->resetContent();
	//fill
	for (int i = 0; i < mapTypes.size(); i++) {
		tileWnd->selectType->addString(mapTypes[i].c_str());
	}
	refreshComboMapTypeCurSel();
}

void EditorWndTileImpl::refreshComboKeyCurSel() {
	int tobesel = -1;
	int i,j = 0;
	for (i = 0; i < ed->world->keys->size(); i++) {
		if (i && !ed->world->keys->isValid(i)) continue;	//i==0 can pass 
		if (i && !(*ed->world->keys)[i]) continue;	//allow i == 0 to pass through as our empty tile
		if (i == editor->mapKey) {
			tobesel = j;
		}
		j++;
	}
	tileWnd->selectKey->setSelected(tobesel);
}

void EditorWndTileImpl::refreshComboKeyContents() {
	char name[256];
	int i, count;
	count = ed->world->texs->texs.size();
	tileWnd->selectKey->resetContent();
	int tobesel = -1;
	int j = 0;
	for (i = 0; i < ed->world->keys->size(); i++) {
		//i == 0 go through
		if (i && !ed->world->keys->isValid(i)) continue;	//allow i==0 to go through
		if (i && !(*ed->world->keys)[i]) continue;			//once again, i==0 continue

		sprintf(name, "%d", i);
		//can we stagger 'i' like this? no. therefore we'll have to read in the text upon click to find which ID we have
		tileWnd->selectKey->addString(name);
		if (i == editor->mapKey) {
			tobesel = j;
		}
		j++;
	}
	tileWnd->selectKey->setSelected(tobesel);
}

void EditorWndTileImpl::refreshSides() {
	char sides = -1;
	if (ed->world->keys->isValid(editor->mapKey)) {
		MapKey *k = (*ed->world->keys)[editor->mapKey];
		if (k) sides = k->sides;
	}
	for (int i = 0; i < numberof(tileWnd->sides); i++) {
		(*tileWnd->sides[i])->setFontColor((sides & bitflag(i)) ? 0xFF0000FF : 0xFFFFFFFF);
	}
}

//refresh everything related to keys . . .
void EditorWndTileImpl::refreshKey() {

	refreshComboKeyCurSel();
	refreshComboTextureCurSel();
	refreshComboMapTypeCurSel();
	refreshSides();

	//refresh friction texts
	MapKey k;	//use an obj so it gets filled with defaults up front
	//only copy the right vals in if we find them
	if (ed->world->keys->isValid(editor->mapKey)) {
		MapKey *rk = (*ed->world->keys)[editor->mapKey];
		if (rk) {
			k = *rk;
		}
	}
	
	TextField *friction[] = {
		*tileWnd->friction[0],
		*tileWnd->friction[1],
		*tileWnd->friction[2],
		*tileWnd->friction[3]
	};
	editor->setFloatTextFields(k.friction, friction, numberof(friction));

	//refresh bounce
	float bounce = k.restitution;
	TextField *m = tileWnd->bounce;
	editor->setFloatTextFields(&bounce, &m, 1);
	
	float pull = k.pull;
	m = tileWnd->pull;
	editor->setFloatTextFields(&pull, &m, 1);

	//texmat
	TextField *texgenU[] = {
		*tileWnd->texgen[0],
		*tileWnd->texgen[1],
		*tileWnd->texgen[2],
	};
	editor->setFloatTextFields(k.texu, texgenU, numberof(texgenU));

	TextField *texgenV[] = {
		*tileWnd->texgen[3],
		*tileWnd->texgen[4],
		*tileWnd->texgen[5],
	};
	editor->setFloatTextFields(k.texv, texgenV, numberof(texgenV));
	
	//bbox
	TextField *bbox[] = {
		*tileWnd->bbox[0],
		*tileWnd->bbox[1],
		*tileWnd->bbox[2],
		*tileWnd->bbox[3],
	};
	editor->setFloatTextFields(&k.bbox.min.x, bbox, numberof(bbox));
	
	m = tileWnd->zBias;
	editor->setFloatTextFields(&k.z, &m, 1);

	tileWnd->shader->setText(k.shaderName);
	tileWnd->postFBO->setChecked(k.postFBO);
	
	m = tileWnd->borderThickness;
	editor->setFloatTextFields(&k.borderThickness, &m, 1);

	m = tileWnd->borderUVThickness;
	editor->setFloatTextFields(&k.borderUVThickness, &m, 1);
	
	tileWnd->group->setText(k.groupname);
	tileWnd->seq->setText(k.seqname);
	
	TextField *planes[numberof(tileWnd->planes)];
	for (int i = 0; i < numberof(tileWnd->planes); i++) {
		planes[i] = *tileWnd->planes[i];
	}
	editor->setFloatTextFields((float*)k.planes, planes, numberof(planes));
}

void EditorWndTileImpl::refresh() {
	//area slider
	tileWnd->sliderArea->setSliderPos(editor->paintArea);
	//dissolve slider
	tileWnd->sliderDissolve->setSliderPos(editor->paintDissolve);

	//todo 
	//only do this upon...
	//init, _create, and _remove
	refreshComboKeyContents();

	//refresh texture
	refreshComboTextureContents();

	//refresh map types
	refreshComboMapTypes();
	
	//refreshes all contents related to the current key
	//calls refreshComboTextureCurSel again
	refreshKey();

}

void EditorWndTileImpl::init() {
	(*editor->editWnd[EW_TILE]) = tileWnd = new MenuTile(ed);
	if (!tileWnd) throw Exception() << "EditorWndTile::init failed";
	editor->rootmenu->addChild(tileWnd);
}

MenuTile::MenuTile(EditorImpl *ed_)
:	super(vec2f(1,1),vec2f(25,41)),
	surfacePaint(createPtr<Text>()),
	surfaceFill(createPtr<Text>()),
	surfaceMove(createPtr<Text>()),
	surfaceMovePreserve(createPtr<Text>()),
	sliderArea(createPtr<Scrollbar>()),
	sliderDissolve(createPtr<Scrollbar>()),
	selectKey(createPtr<Select>()),
	selectTexture(createPtr<Select>()),
	selectBorderTexture(createPtr<Select>()),
	selectType(createPtr<Select>()),
	texPreview(createPtr<Object>()),
	borderTexPreview(createPtr<Object>()),
	shader(createPtr<TextField>()),
	postFBO(createPtr<Checkbox>()),
	createKey(createPtr<Text>()),
	removeKey(createPtr<Text>()),
	//gc::Pointer<Text> *sides[4];
	//gc::Pointer<TextField> *friction[4];
	bounce(createPtr<TextField>()),
	pull(createPtr<TextField>()),
	//gc::Pointer<TextField> *texgen[6];
	borderThickness(createPtr<TextField>()),
	borderUVThickness(createPtr<TextField>()),
	group(createPtr<TextField>()),
	seq(createPtr<TextField>()),
	//gc::Pointer<TextField> *bbox[4],
	zBias(createPtr<TextField>()),
	//gc::Pointer<TextField> *planes[12];
	ed(ed_)
{
	for (int i = 0; i < numberof(sides); i++) { sides[i] = new gc::Pointer<Text>(createPtr<Text>()); }
	for (int i = 0; i < numberof(friction); i++) { friction[i] = new gc::Pointer<TextField>(createPtr<TextField>()); }
	for (int i = 0; i < numberof(texgen); i++) { texgen[i] = new gc::Pointer<TextField>(createPtr<TextField>()); }
	for (int i = 0; i < numberof(bbox); i++) { bbox[i] = new gc::Pointer<TextField>(createPtr<TextField>()); }
	for (int i = 0; i < numberof(planes); i++) { planes[i] = new gc::Pointer<TextField>(createPtr<TextField>()); }

	visible = false;
	addChild(surfacePaint = new Text(vec2f(1,1), "Paint"));
	addChild(surfaceFill = new Text(vec2f(1,3), "Fill"));
	addChild(surfaceMove = new Text(vec2f(13,1), "Move"));
	addChild(surfaceMovePreserve = new Text(vec2f(10,2), "(Preserve)"));

	//btw these arent even used yet!
	addChild(new Text(vec2f(1,5), "Area"));
	addChild(sliderArea = new Scrollbar(vec2f(10,5), 13, true, 1, 10, editor->paintArea));	//call refreshPos every time it updates
	addChild(new Text(vec2f(1,7), "Dissolve"));
	addChild(sliderDissolve = new Scrollbar(vec2f(10,7), 13, true, 0, 1, editor->paintDissolve));

	addChild(new Text(vec2f(1,9), "Key"));
	addChild(selectKey = new Select(vec2f(6,9), 10));
	selectKey->addSelectListener(this);

	addChild(createKey = new Text(vec2f(17,9), "Create"));
	addChild(removeKey = new Text(vec2f(17,10), "Remove"));
	
	addChild(new Text(vec2f(1,11), "Tex"));
	addChild(selectTexture = new Select(vec2f(6,11), 10));
	selectTexture->addSelectListener(this);

	addChild(texPreview = new Object(vec2f(17, 12), vec2f(2,2)));
	texPreview->setBackgroundColor(0xFFFFFFFF);	//then apply textures later

	addChild(new Text(vec2f(1,12), "Border"));
	addChild(selectBorderTexture = new Select(vec2f(6,12), 10));
	selectBorderTexture->addSelectListener(this);

	addChild(borderTexPreview = new Object(vec2f(20, 12), vec2f(2,2)));
	borderTexPreview->setBackgroundColor(0xFFFFFFFF);
	
	addChild(new Text(vec2f(1,13), "Type"));
	addChild(selectType = new Select(vec2f(6,13), 10));
	selectType->addSelectListener(this);
	
	addChild(new Text(vec2f(1,16), "Col.Sides:"));
	addChild(*sides[0] = new Text(vec2f(14,15), "U"));
	addChild(*sides[1] = new Text(vec2f(14,17), "D"));
	addChild(*sides[2] = new Text(vec2f(13,16), "L"));
	addChild(*sides[3] = new Text(vec2f(15,16), "R"));
	
	addChild(new Text(vec2f(1,18), "Friction:"));
	addChild(*friction[0] = new TextField(vec2f(12,18), 7));
	addChild(*friction[1] = new TextField(vec2f(12,20), 7));
	addChild(*friction[2] = new TextField(vec2f(8,19),  7));
	addChild(*friction[3] = new TextField(vec2f(17,19), 7));
	for (int i = 0; i < numberof(friction); i++) (*friction[i])->addTextListener(this);
	
	addChild(new Text(vec2f(1,21), "Bounce:"));
	addChild(bounce = new TextField(vec2f(9,21), 7));
	bounce->addTextListener(this);
	
	addChild(new Text(vec2f(1,22), "Pull:"));
	addChild(pull = new TextField(vec2f(9,22), 7));
	pull->addTextListener(this);
	
	addChild(new Text(vec2f(1,23), "Tex Gen Matrix:"));
	addChild(*texgen[0] = new TextField(vec2f(1,24), 7));
	addChild(*texgen[1] = new TextField(vec2f(9,24), 7));
	addChild(*texgen[2] = new TextField(vec2f(17,24), 7));
	addChild(*texgen[3] = new TextField(vec2f(1,25), 7));
	addChild(*texgen[4] = new TextField(vec2f(9,25), 7));
	addChild(*texgen[5] = new TextField(vec2f(17,25), 7));
	for (int i = 0; i < numberof(texgen); i++) (*texgen[i])->addTextListener(this);

	addChild(new Text(vec2f(1, 26), "Border:"));
	addChild(borderThickness = new TextField(vec2f(11, 26), 10));
	borderThickness->addTextListener(this);

	addChild(new Text(vec2f(1, 27), "BorderUV:"));
	addChild(borderUVThickness = new TextField(vec2f(11, 27), 10));
	borderUVThickness->addTextListener(this);
	
	addChild(new Text(vec2f(1, 28), "Group:"));
	addChild(group = new TextField(vec2f(11, 28), 10));
	group->addTextListener(this);

	addChild(new Text(vec2f(1, 29), "Seq:"));
	addChild(seq = new TextField(vec2f(11, 29), 10));
	seq->addTextListener(this);
	
	addChild(new Text(vec2f(1, 30), "Z Bias:"));
	addChild(zBias = new TextField(vec2f(11, 30), 10));
	zBias->addTextListener(this);

	addChild(new Text(vec2f(1, 31), "Shader:"));
	addChild(shader = new TextField(vec2f(11,31),10));
	shader->addTextListener(this);

	addChild(postFBO = new Checkbox(vec2f(11,32), "Post-FBO?"));	
	postFBO->addCheckboxListener(this);
	
	addChild(new Text(vec2f(1,33),"BBox:"));
	addChild(new Text(vec2f(6,33),"Min:"));
	addChild(*bbox[0] = new TextField(vec2f(10,33), 5));
	addChild(*bbox[1] = new TextField(vec2f(15,33), 5));
	addChild(new Text(vec2f(6,34),"Max:"));
	addChild(*bbox[2] = new TextField(vec2f(10,34), 5));
	addChild(*bbox[3] = new TextField(vec2f(15,34), 5));
	for (int i = 0; i < numberof(bbox); i++) (*bbox[i])->addTextListener(this);

	addChild(new Text(vec2f(1, 35), "Planes:"));
	for (int i = 0; i < numberof(planes); i++) {
		addChild(*planes[i] = new TextField(vec2f(1 + 8 * (i%3), 36 + (i/3)), 7));
		(*planes[i])->addTextListener(this);
	}
}

MenuTile::~MenuTile() {
	for (int i = 0; i < numberof(sides); i++) { delete sides[i]; }
	for (int i = 0; i < numberof(friction); i++) { delete friction[i]; }
	for (int i = 0; i < numberof(texgen); i++) { delete texgen[i]; }
	for (int i = 0; i < numberof(bbox); i++) { delete bbox[i]; }
	for (int i = 0; i < numberof(planes); i++) { delete planes[i]; }
}

//handle buttons here (Text)
//this method is causing the MenuTile to scroll twice as far as other dialog windows 
//that are based upon menu_window_t . . .
eventResponse_t MenuTile::mouseEvent(MouseEvent e, const vec2f &v) {

	//grab our current key just in case ...
	MapKey *k = NULL;
	if (ed->world->keys->isValid(editor->mapKey)) k = (*ed->world->keys)[editor->mapKey];

	if (e.lpress) {
		if (selectedChild.ptr() == surfacePaint) {
			editor->setMethod(EDITOR_SURFACE_PAINT);
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == surfaceFill) {
			editor->setMethod(EDITOR_SURFACE_FILL);
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == surfaceMove) {
			editor->setMethod(EDITOR_SURFACE_MOVE);
			//return RESPONSE_STOP;
		} else if (selectedChild.ptr() == surfaceMovePreserve) {
			editor->movePreserve ^= 1;
			//return RESPONSE_STOP;
		} else if (selectedChild.ptr() == createKey) {
			int newkey = ed->world->keys->create();
			if (newkey == -1) {
				con.print("failed to create map key\n");
			} else {
				ed->tileWnd->refreshComboKeyContents();
				editor->setMapKeySel(newkey);
				//return RESPONSE_STOP;
			}
		} else if (selectedChild.ptr() == removeKey) {
			ed->tileWnd->refreshComboKeyContents();
			editor->setMapKeySel(ed->world->keys->remove(editor->mapKey));
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == *sides[0]) {
			if (k) {
				k->sides ^= bitflag(SIDE_UP);
				ed->tileWnd->refreshSides();
			}
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == *sides[1]) {
			if (k) {
				k->sides ^= bitflag(SIDE_DOWN);
				ed->tileWnd->refreshSides();
			}
			ed->tileWnd->refreshSides();
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == *sides[2]) {
			if (k) {
				k->sides ^= bitflag(SIDE_LEFT);
				ed->tileWnd->refreshSides();
			}
			ed->tileWnd->refreshSides();
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == *sides[3]) {
			if (k) {
				k->sides ^= bitflag(SIDE_RIGHT);
				ed->tileWnd->refreshSides();
			}
			ed->tileWnd->refreshSides();
			return RESPONSE_STOP;
		}
	}
	
	if (selectedChild.ptr() == sliderArea) {
		editor->paintArea = (int)sliderArea->getSliderPos();
		return RESPONSE_STOP;
	} else if (selectedChild.ptr() == sliderDissolve) {
		editor->paintDissolve = sliderDissolve->getSliderPos();
		return RESPONSE_STOP;
	}
	
	super::mouseEvent(e, v);
	return RESPONSE_STOP;
}

bool MenuTile::display() {
	//update colors
	surfacePaint->setFontColor((editor->method == EDITOR_SURFACE_PAINT) ? 0xFF0000FF : 0xFFFFFFFF);
	surfaceFill->setFontColor((editor->method == EDITOR_SURFACE_FILL)  ? 0xFF0000FF : 0xFFFFFFFF);
	surfaceMove->setFontColor((editor->method == EDITOR_SURFACE_MOVE)  ? 0xFF0000FF : 0xFFFFFFFF);
	surfaceMovePreserve->setFontColor((editor->movePreserve) ? 0xFF0000FF : 0xFFFFFFFF);
	//display rect
	return super::display();
}

void MenuTile::callback_userSetSelected(Select *m, int newSel, int oldSel) {
	if (newSel == oldSel) return;

	if (!ed->world->keys->isValid(editor->mapKey)) return;
	MapKey *k = (*ed->world->keys)[editor->mapKey];
	if (!k) return;

	if (m == ed->tileWnd->tileWnd->selectKey) {
		editor->setMapKeySel(atoi(m->getString(newSel)));
	} else if (m == ed->tileWnd->tileWnd->selectTexture) {
		k->tex = (newSel >= 0 && newSel < ed->world->texs->texs.size()) ? ed->world->texs->texs[newSel] : NULL;
		ed->tileWnd->refreshTexturePreview();
	} else if (m == ed->tileWnd->tileWnd->selectBorderTexture) {
		k->borderTex = (newSel >= 0 && newSel < ed->world->texs->texs.size()) ? ed->world->texs->texs[newSel] : NULL;
		ed->tileWnd->refreshTexturePreview();
	} else if (m == ed->tileWnd->tileWnd->selectType) {
		k->type = newSel;
	}
}

void MenuTile::callback_onTextChange(TextField *text) {
	if (!ed->world->keys->isValid(editor->mapKey)) return;
	MapKey *k = (*ed->world->keys)[editor->mapKey];
	if (!k) return;
	
	for (int i = 0; i < numberof(friction); i++) {
		if (text == *friction[i]) {
			k->friction[i] = atof(text->getText().c_str());
			return;
		}
	}
	
	for (int i = 0; i < 3; i++) {
		if (text == *texgen[i]) {
			k->texu[i] = atof(text->getText().c_str());
			return;
		}
		if (text == *texgen[i+3]) {
			k->texv[i] = atof(text->getText().c_str());
			return;
		}
	}
	
	for (int i = 0; i < numberof(bbox); i++) {
		if (text == *bbox[i]) {
			k->bbox[i>>1][i&1] = atof(text->getText().c_str());
			return;
		}
	}
	
	for (int i = 0; i < 12; i++) {
		if (text == *planes[i]) {
			k->planes[i/3][i%3] = atof(text->getText().c_str());
			return;
		}
	}

	if (text == bounce)			k->restitution = atof(text->getText().c_str());
	else if (text == pull)		k->pull = atof(text->getText().c_str());
	else if (text == zBias)		k->z = atof(text->getText().c_str());
	else if (text == borderThickness) k->borderThickness = atof(text->getText().c_str());
	else if (text == borderUVThickness) k->borderUVThickness = atof(text->getText().c_str());
	else if (text == group)		k->setGroup(text->getText().c_str());
	else if (text == seq)		k->setSeq(text->getText().c_str());
	else if (text == shader)	k->setShader(shader->getText().c_str());
}

void MenuTile::callback_onSetChecked(Checkbox *m, bool isChecked) {
	if (!ed->world->keys->isValid(editor->mapKey)) return;
	MapKey *k = (*ed->world->keys)[editor->mapKey];
	if (!k) return;

	if (m == postFBO) k->postFBO = m->isChecked();
}

EditorWndTile::EditorWndTile(EditorImpl *ed) : impl(new EditorWndTileImpl(ed)) {}
EditorWndTile::~EditorWndTile() { delete impl; }
void EditorWndTile::init() { impl->init(); }
void EditorWndTile::refresh() { impl->refresh(); }
void EditorWndTile::refreshKey() { impl->refreshKey(); }
void EditorWndTile::refreshAllTextures() { impl->refreshAllTextures(); }
