/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <math.h>
#include <assert.h>

#include "program.h"
#include <sstream>
using namespace std;

#include "system/gl.h"
#include "system/exception.h"
#include "system/singletons.h"
#include "base/base.h"
#include "menu/sys.h"
#include "editor/editor.h"

#include "world.h"	//only for Display_World, which is the only clientside func in world.cpp
#include "entalloc.h"
#include "client.h"
#include "main.h"
#include "texloader.h"
#include "lighting.h"
#include "draw.h"
#include "network.h"
#include "nettables.h"
#include "animation.h"
#include "blend.h"
#include "delta.h"
#include "game/main.h"
#include "game/client.h"
#include "game/programmacros.h"

float viewTileSize = 16;	//set to 40.f for SNES scale, 24 for original

static CVari client_force_remote_update("client_force_remote_update", 0);
static CVari snapCameraToScreen("snapCameraToScreen", 0);
static CVari snapCameraToTags("snapCameraToTags", 1);
static CVari disableMouse("disableMouse", 0);

static CVarf editorPanSpeed("editorPanSpeed", 50);
static CVarw *local_cvars[] = {
	new CVarw(&viewTileSize,	"viewTileSize"),
};

static int drawLayerFlags = -1;
static CVarw cv_drawLayerFlags(&drawLayerFlags, "drawLayerFlags");

static CVari clientShowPackageContents("clientShowPackageContents", 0);
static CVari client_report_changes("client_report_changes", 0);

// ***************************** local client system

//configurable information - except the viewpos convergence variables
class LocalClient {
public:
	LocalClient() : visible(false), window(vec2f(0,0), vec2f(1,1)) {
		memset(inkey, 0, sizeof(inkey));
	}

	//configuration - what input keys to use
	short inkey[MAX_INPUT_KEYS];
//keyscan uses bits 0-2 in the high-byte, leaving bit 7, the negative sign, untouched
//therefore mouse cmds are gonna be represented via negative values

	bool visible;	//whether the window is visible or not

	//which area of the screen to render through
	box2f window;

	//current viewport tracking coordinates
	vec2f viewpos;
};

class LocalClientSystem {
public:

	//should we std::vector this?
	int clientCount;
	LocalClient clients[MAX_LOCAL_CLIENTS];
		
	LocalClientSystem() {
		clientCount = 0;
	}

	//called upon a connection =)
	bool initViewports(int n) {	//bool Init_LocalClients(int n) {
		//reset
		//memset(clients, 0, sizeof(clients));
	
		if (n < 1) n = 1;
		if (n > MAX_LOCAL_CLIENTS) n = MAX_LOCAL_CLIENTS;
		
		lua_State *L = progSys.getState();
		int top = lua_gettop(L);
		try {
			lua_getGameAPIfunction(L, "setupViewports");	//[setupViewports] [GameAPI]
			lua_pushinteger(L, n);	//number of viewports
			if (lua_pcall(L, 2, MAX_LOCAL_CLIENTS, 0)) throw Exception() << "failed to run.  got errors:\n" << lua_tostring(L,-1);
			int top2 = lua_gettop(L);
			
			const char *boxFields[] = {"min", "max"};
			
			for (int i = 0; i < n; i++) {
				//decode bbox at stack position top-1
				int loc = top2 + i - (MAX_LOCAL_CLIENTS-1);
				if (lua_isnil(L, loc)) {
					clients[i].visible = false;
				} else {
					for (int j = 0; j < numberof(boxFields); j++) {
						lua_getfield(L, loc, boxFields[j]);
						clients[i].window[j] = luaReadVec2f(L);
						lua_pop(L,1);
					}
					clients[i].visible = true;
				}
			}
		} catch (Exception &t) {
			con << "LocalClientSystem::initViewports failed() " << t << "\n";
		}
		lua_settop(L,top);
	
		clientCount = n;
		return true;
	}
};

class ClientConnection {
public:

	ClientConnection() {
		ent.i = -1;
		input.cmd = 0;
		input.mouseActive = 0;
		input.mousePos.set(0.f);
	}

	entptr_t	ent;		//index of the entity to track as the client
	ClientInput	input;	//last input sent to the server - to only send updates =)

	//todo - connection type (non-loopback)
};

//for ent sending messages I should keep a map from client entities to their objects
//.... and for client-side input lookups

class ClientSystemImpl {
protected:
	//client data arrays
	char *local_cl_info;
	char *local_cl_entity;

	//join vars
	bool waitingForJoin;
	double waitForJoinStartTime;
	double waitForJoinLastTime;

	//socket vars
	SocketListen *cl_listener;
	SocketSend *cl_sender;

	//will this call the default ctor on the ents?
	ClientConnection cl_connection[MAX_LOCAL_CLIENTS];

	//the current operating camera 
	box2f viewarea;
	
	//for passing incomplete messages
	vector<string> previousMessageParams;

public:	
	//who is responsible for handling input during menus
	//	only used for its inkey
	//the rest circumvented
	LocalClient menuClient;
	ClientInput menuClientInput;

	//local system.  public only for a cfunc
	LocalClientSystem local;

	ClientSystemImpl();
	
	bool gameInit();
	void gameKill();
	
	bool sessionInit();
	void sessionKill();

	void update();
	void setClientKey(int clientIndex, int keyIndex, int keyCode);
	short *getKeyPtr(int clientIndex);
	short *getMenuKeyPtr();
	void checkDisconnect();
	void setLocalArrays(int index);
	void connect(int count, const char *dest);
	void sendMessage(entptr_t ep, const std::vector<std::string> &params);
	bool getClientInput(entptr_t ep, ClientInput *input);
	bool getMenuClientInput(ClientInput *input);
	void clearMenuKeys();
	box2f getViewArea();

protected:
	void startWaitForJoin();
	void stopWaitForJoin();
	void waitForJoin();
	void updateClient(World *world, int index);
	void updateMenuClient();
	void gatherMessages();
};

ClientSystemImpl::ClientSystemImpl() {
	local_cl_info = NULL;
	local_cl_entity = NULL;

	waitingForJoin = false;
	waitForJoinStartTime = -100;
	waitForJoinLastTime = -99;

	cl_listener = NULL;
	cl_sender = NULL;
}

bool ClientSystemImpl::gameInit() {
	viewTileSize = 16;
	//reset cvars. TODO - make them map/game params
	snapCameraToTags = 1;
	return true;
}

void ClientSystemImpl::gameKill() {
}
	
bool ClientSystemImpl::sessionInit() {	
	//init all client vars here
	local_cl_info = entAllocSys.getClientInfoArray();
	local_cl_entity = entAllocSys.getClientEntityArray();
	return true;
}
	
void ClientSystemImpl::sessionKill() {

	//disconnect... somehow =)
	//this happens upon 'reset' ... hmm... we wanna shutdown
	checkDisconnect();

	if (cl_listener) {
		delete cl_listener;
		cl_listener = NULL;
	}
	if (cl_sender) {
		delete cl_sender;
		cl_sender = NULL;
	}
}

	// ***************************** client updates
	
		//input
//if you switch display/state update, then be sure to reconsult the lighting.h file for light order
void ClientSystemImpl::update() {
	PROFILE();

	//TODO - pick the world per update, or pick the world per client?
	World *world = worldSys.getWorld();

	//can connect go here?
	if (kbd.keystate[Key::F2] == BUTTON_STATE_PRESS) {
		//shift means server reset (reload level, scripts, etc)
		if (kbd.keystate[Key::LSHIFT] == BUTTON_STATE_DOWN ||
			kbd.keystate[Key::RSHIFT] == BUTTON_STATE_DOWN)
		{
			gameState = GS_MENU;
			mainSys.gameReset();
		}
		//no shift means just a session restart (no level reload -- no scripts too?)
		else {
			mainSys.sessionReset();
		}
	}
	if (kbd.keystate[Key::ESC] == BUTTON_STATE_PRESS) {
		gameState = GS_MENU;
	}
	
	//if (gameState == GS_MENU)
	updateMenuClient();

	waitForJoin();		//for connections, in specific
	gatherMessages();	//for all else (merge these two?)

	//game_client running only (connected remotely)
	if (gameflags & GAME_CLIENT) {
		int i;
		LocalClient *c;
		for (i = 0, c = local.clients; i < local.clientCount; i++, c++) {
			if (c->visible) {
			
				//reset lighting system
				lightSys.sysReset();

				//TODO - client windows need to be specified by the mod
				const box2f &window = c->window;
		
				//set the viewport via the local client window
				box2f b = window.scale((vec2f)disp.getSize());
				box2i ib = b;
				vec2i size = ib.size();
				int width = size.x;
				int height = size.y;
		
		////always assume we have a width of 320
		//todo - min rectangle inside the client window -- only zoom in, no zoom out
				vec2f projHalfSize(320.f, 320.f * (float)size.y / (float)size.x);
				//old comment: = (vec2f)size * 0.5f
		
				//prereq to display functions / mouse input
				viewarea.set(
					c->viewpos - projHalfSize / viewTileSize,
					c->viewpos + projHalfSize / viewTileSize);
		
				//display
		
				//set the camera position to the local client entity location
				glMatrixMode(GL_MODELVIEW);
				glLoadIdentity();
				glScalef(viewTileSize,viewTileSize,1.f);
				glTranslatef(-c->viewpos.x, -c->viewpos.y, 0);
		
		//next viewport for render scene, and texture-overlay lighting
				glViewport(ib.min.x, ib.min.y, size.x, size.y);
				glMatrixMode(GL_PROJECTION);
				glLoadIdentity();
				glOrtho(-projHalfSize.x, projHalfSize.x, projHalfSize.y, -projHalfSize.y, -2000, 2000);
				glMatrixMode(GL_MODELVIEW);
		
				//accum
				entptr_t ep = cl_connection[i].ent;
		
				vec2f v;
		
				//if our entity is active then use its position for our display
				//else - complain - and use the screen center for displaying
				if (ep.i >= 0 && ep.i < ENTITY_BLOCK_COUNT) {
					ClientEntity *cl = (ClientEntity *)(local_cl_entity + ep.i * sizeof(ClientEntity));
					v = cl->pos;
				} else {
					con << "client " << i << " lost its entity!\n";
					gameflags &= ~GAME_CLIENT;
					mainSys.sessionKill();
					return;
				}
		
				{
					//here we are asking for the following mapping:
					//tagdata -> tagkey -> tagtype
					int tagKeyID = world->tags->maploc( floor_ftoi(v.x), floor_ftoi(v.y) );
					map<short,MapTagKey*>::iterator it = world->tagkeys->keys.find(tagKeyID);
					if (it != world->tagkeys->keys.end()) {
						MapTagType *tt = it->second->tagType;
						if (tt) {
				
							//update the lighting here or somewhere else?
							lightSys.setAmbientLevel(tt->light);
			
							//display world background
							worldSys.displayBackground(tt, viewarea);
							
							//todo
							//- only display entities once - with a z index
							//- display each map at a certain (scaled by layer #) z index
							//binary tree sort all sprites in the end.  n log n time
							// orrrr ... just sort the entities n log n per layer
							//group all w/layer <0 into the first set, all w/layer >= layer_count-1 into the last
			
							game.clientUpdateEntities((ClientEntity *)local_cl_entity, ENTITY_BLOCK_COUNT, cl_sender);
							
							//I'm changing clientUpdateEntities to become a generic client update for each entity
							//entity updates are already scriptable, so why not push it over the edge?
							//while I'm at it I'm going to have a final delta compression for sending any changed information back to the server
						
							for (int j = 0; j < world->getMapSize().z; j++) {
								if (drawLayerFlags & bitflag(j)) {
									worldSys.displayLayer(j, viewarea);
								}
							}
						}
					}
				}

				//post-accum (render)
				renderSys.display(viewarea);
				
				//before i wouold reset only just before the client update code
				//but i'm negotiating how to use the editor
				//esp with its gl code in there
				//i think this might be the safest way to allow its gl code to overlay
				//while its render code gets collected into the renderer
				//note to self - editor display will mess up with multiple local clients
		
				//pre-accum
				renderSys.reset();
			}
			
			//state
			if (gameState == GS_RUNNING) {	//work on this paused/running deal
				updateClient(world, i);			//turn off update client so we dont send packets out and overflow our loopback buffer =)
			}

			//regardless of whether its running or not, gather editor input here:
			if (editor.isActive()) {
				if (kbd.keystate[Key::UP])		c->viewpos.y -= editorPanSpeed * system_deltatime;
				if (kbd.keystate[Key::DOWN])	c->viewpos.y += editorPanSpeed * system_deltatime;
				if (kbd.keystate[Key::LEFT])	c->viewpos.x -= editorPanSpeed * system_deltatime;
				if (kbd.keystate[Key::RIGHT])	c->viewpos.x += editorPanSpeed * system_deltatime;
			}
		}

		//cycle the animations.  should these be sync'd by the server?
		//(so clients can hide behind animated sprites - relative to their own view and others' )
		if (gameState == GS_RUNNING) {	//work on this paused/running deal
			worldSys.update();
		}
	} else {
		vec2f c_viewpos = vec2f(0.f);

		//set the viewport via the local client window
		box2f b = box2f(vec2f(0.f), vec2f(1.f)).scale((vec2f)disp.getSize());
		box2i ib = b;
		vec2i size = ib.size();
		int width = size.x;
		int height = size.y;

////always assume we have a width of 320
//todo - min rectangle inside the client window -- only zoom in, no zoom out
		vec2f projHalfSize(320.f, 320.f * (float)size.y / (float)size.x);
		//old comment: = (vec2f)size * 0.5f
		

		//prereq to display functions / mouse input
		viewarea.set(
			c_viewpos - projHalfSize / viewTileSize,
			c_viewpos + projHalfSize / viewTileSize);

		//display

		//set the camera position to the local client entity location
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glScalef(viewTileSize,viewTileSize,1.f);
		glTranslatef(-c_viewpos.x, -c_viewpos.y, 0);

//next viewport for render scene, and texture-overlay lighting
		glViewport(ib.min.x, ib.min.y, size.x, size.y);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(-projHalfSize.x, projHalfSize.x, projHalfSize.y, -projHalfSize.y, -2000, 2000);
		glMatrixMode(GL_MODELVIEW);
			
		renderSys.display(viewarea);
		renderSys.reset();
	}

		//only needed thanx to the console
	//main_display->ResetViewport();
	glViewport(0,0,VEC2ELEM(disp.getSize()));	//because our base views might not be handling our stuff ...
	//i think this is done once per frame anyways ...

	//now that we have all the viewport, display all overlays
	editor.display(viewarea);

	//see i have display code in the input funcitons
	//i know its bad but who cares
	//either way, this was before entity stuff so the editor could capture input
	//but ... i clearly am not doing that at the moment so i might as well do that here
	//todo: determine whether we get editor input and allow that to bypass ent input
	//and then move this code back up before the ent input code
	//and then make sure to move all the display code in editor input outside to editor display
	{
		if (editor.isActive()) {
			editor.updateMouse(
				viewarea.min.x * (1.f - mouse.getFPos().x) + viewarea.max.x * mouse.getFPos().x,
				viewarea.max.y * (1.f - mouse.getFPos().y) + viewarea.min.y * mouse.getFPos().y);
		}
	}

	//i dont know where else to connect this...
	//for now i'll make it so if the menu system gets input then it trumps the editor's
	//this is especially for the menu-based editor system
	//but if - later - we want <editor menu> -> <editor base> -> <game menu>
	// - then we can make the MenuSystem_GetInput based upon 
	if (!pwn::menuSys.getInput()) {
		editor.getInput();		//editor/game interactions
		//dont do display_editor here - it needs to be called after client rendering =)
	}

	pwn::menuSys.display();
}

void ClientSystemImpl::setClientKey(int clientIndex, int keyIndex, int keyCode) {
	if (clientIndex < 0 || clientIndex >= MAX_LOCAL_CLIENTS) return;
	LocalClient &cl = local.clients[clientIndex];
	if (keyIndex < 0 || keyIndex >= MAX_INPUT_KEYS) return;
	cl.inkey[keyIndex] = keyCode;
}

short *ClientSystemImpl::getKeyPtr(int clientIndex) { 
	if (clientIndex < 0 || clientIndex >= MAX_LOCAL_CLIENTS) return NULL;
	LocalClient &cl = local.clients[clientIndex];
	return cl.inkey;
}

short *ClientSystemImpl::getMenuKeyPtr() {
	return menuClient.inkey;
}

// ******************** client waiting for server connection message . . .

//called by client main loop
//only does anything if we're waiting for a join game reply

void ClientSystemImpl::startWaitForJoin() {	//init the wait for join here:
//		con.print("startWaitForJoin()\n");
	waitingForJoin = true;
	waitForJoinStartTime = timer.getTime();
	waitForJoinLastTime = waitForJoinStartTime - 1;
}
	
void ClientSystemImpl::stopWaitForJoin() {
//		con.print("stopWaitForJoin()\n");
	waitingForJoin = false;
}
	
void ClientSystemImpl::waitForJoin() {	//called from clientSys.update()
	if (!waitingForJoin) return;

	double thistime = timer.getTime();

	if (thistime > waitForJoinStartTime + 15.0) {
		con << "waitForJoin():  failed to get a join message" << endl;
		stopWaitForJoin();
		return;
	}

	if (!cl_listener || !cl_listener->isListening()) {
		con << "waitForJoin(): called when WE'RE NOT EVEN LISTENING!" << endl;
		stopWaitForJoin();
		return;
	}

	if (!cl_sender->isSending()) {
		con << "waitForJoin():  someone cut off our sending!" << endl;
		stopWaitForJoin();
		return;
	}

	if (waitForJoinLastTime < waitForJoinStartTime && thistime >= waitForJoinStartTime) {
		con << "waitForJoin(): trying..." << endl;
	}
	if (waitForJoinLastTime < waitForJoinStartTime + 5 && thistime >= waitForJoinStartTime + 5) {
		con << "waitForJoin(): still trying..." << endl;
	}
	if (waitForJoinLastTime < waitForJoinStartTime + 10 && thistime >= waitForJoinStartTime + 10) {
		con << "waitForJoin(): believe it or not we're still Trying..." << endl;
	}

	waitForJoinLastTime = thistime;

	Message<PACKET_LENGTH> m;
//	char buffer[PACKET_LENGTH];
//	int bufferlen = sizeof(buffer);
	if (cl_listener->recv(m, NULL) == 1) {	//got something
//		const char *buffer = m.data;
//		int bufferlen = m.size;
//		con.print("cl_listener->recv got a command!\n");

//TODO - if a server accidently sends us a non-join command
//when we're waiting on a join - we will complain!
//so instead, ignore it!
//TODO - i just copied this all into the default client packet handling code
//for those times when the server changes levels...
//so move all code down there! and make a state engine on handling dif packets at dif times...

		char cmd = m.readb();
		if (cmd == NETCMD_SV_SET_CLIENTS) {

			int count = m.readb();
			if (count != local.clientCount) {
				con << "waitForJoin(): expected " << local.clientCount << " clients, only got " << count << " back from the server..." << endl;
			}

//				con.print("waitForJoin(): reading join request recieved for %d clients\n", count);
			
			if (!count) {
				con << "waitForJoin(): couldn't find any free space for our clients on this server" << endl;
				cl_sender->kill();
				stopWaitForJoin();
				return;
			}

			//short *s = (short *)(buffer + 2);
			for (int i = 0; i < count; i++) {
				cl_connection[i].ent.i = m.reads();//*s;
				//s++;
				
//connection_s::ent write
//					con.print("waitForJoin(): setting local client %d entity to %d\n", i, cl_connection[i].ent.i);
			}

//				con.print("waitForJoin(): done connecting\n");

			//done
			gameflags |= GAME_CLIENT;
			stopWaitForJoin();
		}
	}
	
	if (m.overflow) con << "waitForJoin() found a chopped packet!" << endl;
}

	// ******************** client update
	
void ClientSystemImpl::updateClient(World *world, int index) {
	LocalClient *cl = local.clients + index;

	//use client data
	if (cl_connection[index].ent.i < 0 || cl_connection[index].ent.i >= ENTITY_BLOCK_COUNT) return;
	ClientEntity *local_ent = (ClientEntity *)(local_cl_entity + cl_connection[index].ent.i * sizeof(ClientEntity));
	bool local_active = local_ent->active;
	vec2f localpos = local_ent->pos;
	
	//SEND OUT GATHERED INPUT DATA
	//
	//notice - this is all the client can send: input commands

	vec2f mousepos;
	if (local_active && local.clientCount == 1) {
		//I *heart* LERPs
		//mousepos.x = viewarea.min.x * (1.f - mouse.getFPos().x) + viewarea.max.x * mouse.getFPos().x;
		//mousepos.y = viewarea.max.y * (1.f - mouse.getFPos().y) + viewarea.min.y * mouse.getFPos().y;
		mousepos.x = mouse.getFPos().x;
		mousepos.y = 1.f - mouse.getFPos().y;
	}

	//loopback - direct write
	//gather mouse button input
	bool mouseinput = false;

	//if we're doing split screen
	//then NO ONE GETS A MOUSE!
	if (local.clientCount == 1 && !disableMouse) {
		mouseinput = true;
	} else {
		mousepos = vec2f(.5f, .5f);	//localpos;
	}
		
	//gather keyboard input
	int input = 0;
	for (int i = 0; i < MAX_INPUT_KEYS; i++) {
		unsigned char bits = 0;
		short key = cl->inkey[i];
		
		//TODO - rather than determine state here, just send the up/down info
		//THEN the server can determine the state
		//and that way, press/release states will ONLY LAST ONE FRAME
		//
		//this is actually going on inthe game mod - the 'new' bit is recalculated in the game mod
		//
		//however - would it better impress accuracy of input if we sent all the press/releases instead?
		//like sending the keyboard queue coinciding with the keys we're keeping track of?
		//that way - the game mod would be better at not skipping beats
		//lag would only occur if someone was hammering away at a key rather than holding it down . . .

		//todo - unify the button state enums in the m_system3d?
		if (key < 0) {
			bits = mouse.getButtonState(~key);
		} else {
			bits = kbd.keystate[key];
		}

		//input values takes up two bits:
		input |= (bits & 3) << (i << 1);
	}

	//TODO - combine all into one packet and send at the end of the client system update 
	if (local_active) {
		//and send it...somehow...

		if (!cl_sender || !cl_sender->isSending()) {
			//con.print("got a remote connection when we're not sending!\n");
		} else {

			/*
			hmm... how can we combine sending ClientInput with sending any updates that happen on the clientside of things...
			the obvious is just run another delta compression message for sending the netobj back to the server
			but why bother have two separate means of doing the same thing?
			why not put the ClientInput inside the netTable and just run delta compression back and forth between client and server?
			
			well the catch there is right now the GameAPI handles
			(a) client connection and
			(b) setting client input
			right now the correlation between entities and connections is pretty loose
			we do already have 1-1 associations of entities to connections
			we'd have to enforce a certain entity class for connections to jump to the action of putting the ClientInput inside the netTable
			...and doing so would have an arbitrary attachment on the already ambiguous "entity hierarchy" on the server side
			... which is nothing more than Enity -> Renderable -> Collidable
			
			
			*/
			ClientInput in;
			
			in.cmd = input;
			in.mousePos = mousepos;
			in.mouseActive = mouseinput;

			if (memcmp(&in, &cl_connection[index].input, sizeof(ClientInput))) {
				//write it thru the client connection.  lots of asserts 'cuz this is a first for me doing it
				Message<> m;
				m << (char)NETCMD_CL_SEND_INPUT;
				m << (short)cl_connection[index].ent.i;
				m << (int)in.cmd;
				m << (char)in.mouseActive;
				m << (float)in.mousePos.x;
				m << (float)in.mousePos.y;
				assert(!m.overflow);
				cl_sender->send(m);
				//and update the input
				memcpy(&cl_connection[index].input, &in, sizeof(ClientInput));
			}
		}
	}

	if (!editor.isActive()) {	//dont change viewport if the editor is active
	
		//TODO - move this to mod player code client-side update

		if (local_active) {

			//track view to client position
			//by default use the entity's position
			vec2f dest_viewpos = localpos;
			/*
			hmm ...
			here getClientViewPos is being called
			but its being called from the _ents[] array
			and that is only going to be relevant for the server
			(the server holds the main _ents copy and only sends netobj info to clients
			 so clients don't get any _ents info at all, 
				unless they decide to allocate it in their render method
					which is potential for error since server ents might overwrite them ... )

			so how should client ents get their viewpos info ?
			the only info the client gets to see is in the netobj atm

			we can have this read from the net object of the client's entity
			and just have all client entities expected to have a "clientView" vec2f net object ...
			*/

			{
				ClientEntity *e = (ClientEntity *)local_cl_entity + cl_connection[index].ent.i;
				EntClass *ec = netTablesSys.getEntClass(e->entClassID);
				if (ec && ec->netTableID != -1) {
					//set our net table ptr correctly
					NetTableInstance net(ec->netTableID, e, -1);

					void *ptr = net.rawAccess("viewpos");
					if (ptr) {
						dest_viewpos = *(vec2f*)ptr;
					}
				}
			}

#if 1
			//where again is viewarea computed? in clientSys.update()
			//we need to give the mod access to the view position code
			//and for that, the mod needs access to clip the viewpos to tags
			//and for that, the camera bbox needs to have been computed
			vec2f camsize = viewarea.size();

			//HACK TO ALIGN CAMERA PER SCREEN
			//which is no longer working...
			if (snapCameraToScreen) {

				//hmm. . . 640x480 isn't always the same
				//don't forget, this depends upon the client connect windows

				//int bx = (int)(dest_viewpos_x / (640.f / viewTileSize));
				//int by = (int)(dest_viewpos_y / (480.f / viewTileSize));
				//dest_viewpos_x = ((float)bx + 0.5f) * (640.f / viewTileSize);
				//dest_viewpos_y = ((float)by + 0.5f) * (480.f / viewTileSize);
				int bx = (int)(dest_viewpos.x / (camsize.x / viewTileSize));
				int by = (int)(dest_viewpos.y / (camsize.y / viewTileSize));
				dest_viewpos.x = ((float)bx + 0.5f) * (camsize.x / viewTileSize);
				dest_viewpos.y = ((float)by + 0.5f) * (camsize.y / viewTileSize);
			}

			//update the tag info:
			vec2i localposi(floor_ftoi(localpos.x), floor_ftoi(localpos.y));
			short userTag = world->tags->maploc(VEC2ELEM(localposi));

			//align to tags
			if (snapCameraToTags) {
				vec2f camHalfSize = camsize * 0.5f;
				box2i tagbox(
					vec2i(
						floor_ftoi(dest_viewpos.x - camHalfSize.x),
						floor_ftoi(dest_viewpos.y - camHalfSize.y)),
					vec2i(
						floor_ftoi(dest_viewpos.x + camHalfSize.x),
						floor_ftoi(dest_viewpos.y + camHalfSize.y)));
						
				//if the lower bounds tag does not match then push camera upwards
				short lower = world->tags->maploc(localposi.x, tagbox.max.y);
				if (lower != userTag) {
					//snap local_ent->y to tag_tile_height, store in dest_viewpos_y
					//offset by the half vector
					dest_viewpos.y = floor_ftoi(localpos.y / (float)world->tags->tagsize.y + 1.f) * (float)world->tags->tagsize.y - camHalfSize.y;
					//dest_viewpos.y = floor_ftoi(dest_viewpos.y / (float)tag_tile_height + 1.f) * (float)tag_tile_height - camHalfSize.y;
				}
				
				//if upper bounds tag does not match then push the camera downwards
				short upper = world->tags->maploc(localposi.x, tagbox.min.y);
				if (upper != userTag) {
					//snap local_ent->y to tag_tile_height, store in dest_viewpos_y
					//offset by the half vector
					dest_viewpos.y = floor_ftoi(localpos.y / (float)world->tags->tagsize.y) * (float)world->tags->tagsize.y + camHalfSize.y;
					//dest_viewpos.y = floor_ftoi(dest_viewpos.y / (float)tag_tile_height) * (float)tag_tile_height + camHalfSize.y;
				}

				//if upper bounds tag does not match then push the camera downwards
				short left = world->tags->maploc(tagbox.min.x, localposi.y);
				if (left != userTag) {
					//snap local_ent->x to tag_tile_width, store in dest_viewpos_x
					//offset by the half vector
					dest_viewpos.x = floor_ftoi(localpos.x / (float)world->tags->tagsize.x) * (float)world->tags->tagsize.x + camHalfSize.x;
					//better yet ... just go right one?
					//mind you this whole system is janked
					//dest_viewpos.x = floor_ftoi(dest_viewpos.x / (float)tag_tile_width) * (float)tag_tile_width + camHalfSize.x;
				}

				//if upper bounds tag does not match then push the camera downwards
				short right = world->tags->maploc(tagbox.max.x, localposi.y);
				if (right != userTag) {
					//snap local_ent->x to tag_tile_width, store in dest_viewpos_x
					//offset by the half vector
					dest_viewpos.x = floor_ftoi(localpos.x / (float)world->tags->tagsize.x + 1.f) * (float)world->tags->tagsize.x - camHalfSize.x;
					//dest_viewpos.x = floor_ftoi(dest_viewpos.x / (float)tag_tile_width + 1.f) * (float)tag_tile_width - camHalfSize.x;
				}
			}
#endif
			cl->viewpos = dest_viewpos;
		}
	}
}

void ClientSystemImpl::updateMenuClient() {
	//gather keyboard input
	menuClientInput.cmd = 0;
	for (int i = 0; i < MAX_INPUT_KEYS; i++) {
		unsigned char bits = 0;
		short key = menuClient.inkey[i];
		if (key < 0) {
			bits = mouse.getButtonState(~key);
		} else {
			bits = kbd.keystate[key];
		}
		menuClientInput.cmd |= (bits & 3) << (i << 1);
	}
	
	menuClientInput.mousePos.x = mouse.getFPos().x;
	menuClientInput.mousePos.y = 1.f - mouse.getFPos().y;
	menuClientInput.mouseActive = true;		
}

void ClientSystemImpl::clearMenuKeys() {
	menuClientInput.cmd = 0;
}


//loopback optimization
// - called by the server when we get a loopback connection  
//  to tell the client which connection index to switch over to
void ClientSystemImpl::setLocalArrays(int index) {
	if (index < 0 || index >= numberof(serverConnSys.connections)) index = 0;
//with one buffer just leave local pointing to the same place
#ifndef ONE_DELTA_ENTITY_BUFFER
	local_cl_entity = entAllocSys.getClientEntityArray() + sizeof(ClientEntity) * index * CL_ENTITY_NUM_PER_CLIENT;
#endif
}

void ClientSystemImpl::connect(int count, const char *dest) {

	checkDisconnect();

	//prep viewports/input keys
//		con.print("connect(): Init_LocalClients(%d)\n", count);
	local.initViewports(count);

	//determine our connection type depending upon 'dest'
	//depending upon 'dest',
	//	establish a connection communication scheme
	//then, send a message through it:
	//	"i am connecting with # people"
	//and wait several seconds for a reply:
	//	"i hear you, here are your entity #'s: # # # . . ."
	//then we are good to go.

#ifndef OMIT_NETWORKING

	if (dest) {	//remote connect

		//clear the "running a server!" flag - just run the client
		gameflags &= ~GAME_SERVER;
		con.print("GAME_SERVER cleared\n");

		//make sure our local entity stats are clear
		//prep it to recieve correct stats from the server

		//if we are doing an inet connection then
		//make sure we have an inet cl_listener & cl_sender

		if (cl_listener && cl_listener->getType() != SOCKET_INET) {
			delete cl_listener;
			cl_listener = NULL;
		}
		if (!cl_listener) {
			cl_listener = new SocketListenINET();
		}

		if (cl_sender && cl_sender->getType() != SOCKET_INET) {
			delete cl_sender;
			cl_sender = NULL;
		}
		if (!cl_sender) {
			cl_sender = new SocketSendINET();
		}
	} else

#endif //OMIT_NETWORKING

	{	//local connect

		//ALL JUST LIKE ABOVE EXCEPT DONT CLEAR THE SERVER FLAG!

		if (cl_listener && cl_listener->getType() != SOCKET_LOOPBACK) {
			delete cl_listener;
			cl_listener = NULL;
		}
		if (!cl_listener) {
			cl_listener = new SocketListenLoopback();
		}

		if (cl_sender && cl_sender->getType() != SOCKET_LOOPBACK) {
			delete cl_sender;
			cl_sender = NULL;
		}
		if (!cl_sender) {
			cl_sender = new SocketSendLoopback();
		}
	}

	//now does the client have to start off listening in the first place?
	//how do we reply to the client from the server?
	if (cl_listener->isListening()) {
		con.print("connect(): cl_listener %d is already listening\n", cl_listener->getType());
	} else {
//			con.print("connect(): init'ing listening...\n");
#ifndef OMIT_NETWORKING
		if (dest) {
			((SocketListenINET *)cl_listener)->init(0);
		} else
#endif //OMIT_NETWORKING
		{
			((SocketListenLoopback*)cl_listener)->init(LOOPBACK_TYPE_CLIENT);
		}
	}

//establish connection

//		con.print("connect(): connecting to dest name %s\n", dest);

	if (cl_sender->isSending()) {
		con.print("connect(): already connected - killing old connection...\n");
		cl_sender->kill();
	}

	bool sender_init_result = false;
#ifndef OMIT_NETWORKING
	if (dest) {
		sender_init_result = ((SocketSendINET*)cl_sender)->init(dest,0);
	} else
#endif //OMIT_NETWORKING
	{
		sender_init_result = ((SocketSendLoopback*)cl_sender)->init( LOOPBACK_TYPE_CLIENT );
	}

	if (!sender_init_result) {
		con.print("connect(): failed to init new connection\n");
		return;
	}

//		con.print("connect(): connected!\n");

//then tell the server to spawn an entity

	//(first clear our own entity set)
	//(i think all we need to clear is the client data...)
	entAllocSys.clearClientData();

//		con.print("connect(): sending join request for %d clients\n", count);
	Message<2> m;
	m << (char)NETCMD_CL_JOIN_REQUEST;
	m << (char)count;
	//char connect_cmd[256];
	//connect_cmd[0] = NETCMD_CL_JOIN_REQUEST;
	//connect_cmd[1] = count;
	cl_sender->send(m);	// connect_cmd, 2);

//and read back which entity it was that the server spawned

	startWaitForJoin();
}
	
// ******************** client connection goods

//needs to be public for cfunc 
void ClientSystemImpl::checkDisconnect() {
	if (!(gameflags & GAME_CLIENT)) return;

	if (!cl_sender || !cl_sender->isSending()) {
		con << "tried to disconnect when we weren't sending!" << endl;
	} else {
		con << "checkDisconnect disconnecting...\n";
		//send a "disconnect me" message to the connected server
		char cmd = NETCMD_CL_DISCONNECT;
		cl_sender->send(&cmd, 1);
	}

	gameflags &= ~GAME_CLIENT;
}
	
// ******************** client system messages
	
void ClientSystemImpl::gatherMessages() {
	//char buffer[PACKET_LENGTH];

	if (waitingForJoin) return;	//of the method which i should merge this all with.

	if (!cl_listener || !cl_listener->isListening()) return;	//can't listen if we're not listening =) obviously

//	if (gameflags & GAME_SERVER) return;	//no need to gather client messages if a server is running here (localhost)
//	con.print("gatherMessages...\n");

	Message<PACKET_LENGTH> bm;

	for(;;) {
		//int bufferlen = sizeof(buffer);
		if (cl_listener->recv(bm, NULL) != 1) break;
		if (!bm.size) break;	//stop at the zero length packets
		
		if (clientShowPackageContents) {
			con << "got packet: ";
			for (char *p = bm.data; p < bm.data + bm.size; p++) {
				con.print("%2x", (unsigned char)*p);
			}
			con << "\n";
		}

		//pre - check the listen port is from the server
//		con.print("gatherMessages() got a message %d\n", buffer[0]);

		//update all this to messages once we get some good net to test on . . .
		for (char *p = bm.data; p < bm.data + bm.size; ) {
			int len = bm.size - (int)(p - bm.data);
		
			switch (p[0]) {

			case NETCMD_SV_SET_CLIENTS:
				p++;
				if (len < 2) {
					con << "are you sure you got set clients message?  there isn't any room for the number of clients (len = " << len << ")\n";
				} else {
					int count = *p; p++;
					if (count != local.clientCount) {
						con.print("gatherMessages(): expected %d clients, only got %d back from the server...\n",
							local.clientCount, count);
					}

					con.print("reading join request recieved for %d clients\n", count);
					
					if (!count) {
						con.print("couldn't find any free space for our clients on this server\n");
						//cl_sender->kill();	only cl_sender->kill if we're joining . . .
						stopWaitForJoin();	//change connection state to 'sittin on my bum...'
						return;
					}

					for (int i = 0; i < count; i++) {
						con.print("setting local client %d entity to %d\n", i, *(short *)p);
						//connection_s::ent write
						cl_connection[i].ent.i = *(short *)p; p += 2;
					}

					con << "done connecting\n";

					//done
					gameflags |= GAME_CLIENT;

					stopWaitForJoin();
				}
				break;

			case NETCMD_SV_SETLEVEL:

				//notice - during this - we're losing some important packets
				//because we're busy putting the level together
				// like . . . who this client /is/ amidst all the client data
				//so . . .
				// should we have a "ok server, start your entity dump" message to 
				//be sent after each changelevel?
				//sure . . .

				p++;	//skip the cmd
				{
					char levelname[256];
					memset(levelname, 0, sizeof(levelname));
					int _max = sizeof(levelname)-1;
					if (_max > len-1) _max = len-1;
					strncpy(levelname, p, _max);
					con.print("remote request to load world %s\n", levelname);
#pragma warning TODO - looks like the client module needs ... well ... either a copy of the world ... or ... for me to think this through better
					if (worldSys.loadWorld(levelname)) {
						con << "...success\n";
					} else {
						con << "...failed\n";
					}
				}
				p += strlen(p) + 1;	//skip the string and the '\0'

				{
					con << "sending ready message...\n";

					//send a 'we're ready' packet
					char sendmsg[1];
					sendmsg[0] = NETCMD_CL_READY;
					cl_sender->send(sendmsg, 1);
				}
				break;

			case NETCMD_SV_UPDATE_ENT:
				p++;
				if (len < 5) {
					con << "are you sure you got an update ent message?  there isn't any room for the ent index and message (len = " << len << ")\n";
				} else {

					int index = *(short *)p; p += 2;
					
					if (client_report_changes) con << "updating ent " << index << endl;

					if (index < 0 || index >= ENTITY_BLOCK_COUNT) {
						con << "got an update for an out of bounds entity " << index << endl;
					} else {
						char *dest = local_cl_entity + index * sizeof(ClientEntity);
						ClientEntity *dest_ent = (ClientEntity *)dest;
#if 0
						con.print("sv->cl recv ");
						for (char *u = p;;) {
							int ofs = *u; u++;
							int size = *u; u++;
							if (!size) break;
							con.print(" %2x:", ofs);
							for (int j = 0; j < size; j++) {
								con.print(" %2x->%2x", ((char*)dest)[ofs+j], *u); u++;
							}
						}
						con.print("\n");
#endif
						if (!deltaUncompress(p, dest_ent, 0, sizeof(ClientEntity))) {
							con << "tried to write an out of bound block to the entity " << index << endl;
							dest_ent->active = false;
						}
					}
				}
				break;
			case NETCMD_SV_REMOVE_ENT:
				p++;
				if (len < 3) {
					con << "are you sure you got a remove ent message?  there isn't any room for the ent index (len=" << len << ")\n";
				} else {
					int index = *(short *)p; p += 2;

					if (client_report_changes) con << "updating ent " << index << "\n";

					if (index < 0 || index >= ENTITY_BLOCK_COUNT) {
						con << "NETCMD_SV_REMOVE_ENT got an update for an out of bounds entity " << index << "\n";
					} else {
						ClientEntity *dest = (ClientEntity *)(local_cl_entity + index * sizeof(ClientEntity));
						memset(dest, 0, sizeof(ClientEntity));
					}
				}
				break;
			case NETCMD_SV_SEND_MESSAGE:
				p++;
				if (len < 3) {
					con << "got a NETCMD_SV_SEND_MESSAGE with no room for an ent (len=" << len << ")\n";
				} else {
					int index = *(short*)p; p += 2;
					if (index < 0 || index >= ENTITY_BLOCK_COUNT) {
						con << "NETCMD_SV_SEND_MESSAGE got an update for an out of bounds entity " << index << "\n";
					} else {
						ClientEntity *e = (ClientEntity *)(local_cl_entity + index * sizeof(ClientEntity));
						vector<string> params;
						bool broken = false;
						for (;;) {
#if 1	//handle splits
							if (*p == 1) {
								// ... then there's more to come ...
								broken = true;
								for (vector<string>::iterator i = params.begin(); i != params.end(); ++i) {
									previousMessageParams.push_back(*i);
								}
								p+=2;
								break;
							} else
#endif
							{
								int len = strlen(p);
								if (!len) {	//a \0 in place of a string is the signal that it's over
									p++;
									break;
								}
								params.push_back(p);
								p += len+1;
								if (p > bm.data + bm.size) {
									con << "NETCMD_SV_SEND_MESSAGE got a broken message" << endl;
									broken = true;
									break;
								}
							}
						}
						if (!broken) {
							if (previousMessageParams.size()) {
								//only mess with this if we're doing a split message
								for (vector<string>::iterator i = params.begin(); i != params.end(); ++i) {
									previousMessageParams.push_back(*i);
								}
								params = previousMessageParams;
							}
							previousMessageParams.clear();
						
							lua_State *L = progSys.getState();
							int top = lua_gettop(L);
							lua_getglobal(L, "_entClasses");		//[_entClasses]
							lua_pushinteger(L, e->entClassID);		//[_entClasses] [e->entClassID]
							lua_gettable(L, -2);					//[_entClasses] [class]
							lua_getfield(L, -1, "clientRecvMessage");	//[_entClasses] [class] [class.clientUpdateRemote]
							if (!lua_isfunction(L, -1)) {
								con << "NETCMD_SV_SEND_MESSAGE class did not possess clientRecvMessage member function" << endl;
							} else {
								for (vector<string>::iterator i = params.begin(); i != params.end(); ++i) {
									lua_pushstring(L, i->c_str());
								}
								EntClass *ec = netTablesSys.getEntClass(e->entClassID);
								if (!ec) {
									con << "NETCMD_SV_SEND_MESSAGE tried to send to an entity with an unrecognizable ent class" << endl;
								} else {
									int tableID = ec->netTableID;
									NetTableInstance net(tableID, e, index);

									if (lua_pcall(L, params.size(), 0, 0)) {
										con << "clientRecvMessage: lua_pcall got error: " << lua_tostring(L, -1) << endl;
									}
								}
							}
							lua_settop(L, top);
						}
					}
				}
				break;
			case NETCMD_SV_DISCONNECT:
				p++;
				checkDisconnect();
				break;
			default:
				con.print("got an unknown command from the server: %d\n", (int)p[0]);
				p += bm.size;
				break;
			}
		}
	}
}
	
void ClientSystemImpl::sendMessage(entptr_t ep, const std::vector<std::string> &params) {
	for (ClientConnection *conn = cl_connection; conn < endof(cl_connection); conn++) {
		if (conn->ent.i == ep.i) {
			Message<> m;

			m << (char)NETCMD_CL_SEND_MESSAGE;
			m << (short)ep.i;
			for (vector<string>::const_iterator j = params.begin(); j != params.end(); ++j) {
				m << j->c_str();
			}
			m << '\0';
			if (m.overflow) {
				con << "ClientSys::sendMessage got an overflow" << endl;
			} else {
				cl_sender->send(m);
			}
			return;
		}
	}
	con << "ClientSys::sendMessage unable to find connection for entity " << netTablesSys.getCurrentInstance()->getEntIndex() << endl;
}

bool ClientSystemImpl::getClientInput(entptr_t ep, ClientInput *input) {
	for (ClientConnection *conn = cl_connection; conn != endof(cl_connection); conn++) {
		if (conn->ent.i != ep.i) continue;
		*input = conn->input;
		return true;
	}
	return false;
}

bool ClientSystemImpl::getMenuClientInput(ClientInput *input) {
	*input = menuClientInput;
	return true;
}

box2f ClientSystemImpl::getViewArea() { return viewarea; }


ClientSystem::ClientSystem() : impl(new ClientSystemImpl()) {}
ClientSystem::~ClientSystem() { delete impl; }
bool ClientSystem::gameInit() { return impl->gameInit(); }
void ClientSystem::gameKill() { impl->gameKill(); }
bool ClientSystem::sessionInit() { return impl->sessionInit(); }
void ClientSystem::sessionKill() { impl->sessionKill(); }
void ClientSystem::update() { impl->update(); }
void ClientSystem::setLocalArrays(int index) { impl->setLocalArrays(index); }
void ClientSystem::connect(int count, const char *dest) { impl->connect(count, dest); }
void ClientSystem::setClientKey(int clientIndex, int keyIndex, int keyCode) { impl->setClientKey(clientIndex, keyIndex, keyCode); }
short *ClientSystem::getKeyPtr(int clientIndex) { return impl->getKeyPtr(clientIndex); }
short *ClientSystem::getMenuKeyPtr() { return impl->getMenuKeyPtr(); }
void ClientSystem::sendMessage(entptr_t ep, const std::vector<std::string> &params) { impl->sendMessage(ep, params); }
bool ClientSystem::getClientInput(entptr_t ep, ClientInput *input) { return impl->getClientInput(ep, input); }
bool ClientSystem::getMenuClientInput(ClientInput *input) { return impl->getMenuClientInput(input); }
void ClientSystem::clearMenuKeys() { impl->clearMenuKeys(); }
box2f ClientSystem::getViewArea() { return impl->getViewArea(); }
ClientSystem clientSys;

//pass client index and key index in the game->clientKeys array
//	return key ... scancode or key string? string.
int PF_getClientKeyValue(lua_State *L) {
	int argc = lua_gettop(L);
	if (argc < 2) {
		luaL_error(L, "usage: keyName = getClientKeyValue clientIndex, clientKeyIndex, newKeyName");
	}
	LocalClient *cl = NULL;
	if (lua_isnumber(L, 1)) {
		int clientIndex = lua_tonumber(L, 1);
		if (clientIndex >= 0 && clientIndex < numberof(clientSys->local.clients)) {
			cl = clientSys->local.clients + clientIndex;
		}
	} else if (lua_isstring(L, 1)) {
		cl = &clientSys->menuClient;
	}
	if (!cl) {
		luaL_error(L, "getClientKeyValue got a bad first arg.  expected clientIndex or 'menu'.");
	}
	
	int clientKeyIndex = lua_tonumber(L, 2);
	if (clientKeyIndex < 0 || clientKeyIndex > MAX_INPUT_KEYS) {
		luaL_error(L, "getClientKeyValue got a bad second arg.  expected clientKeyIndex");
	}
	
	short keycode = cl->inkey[clientKeyIndex];
	
	if (argc >= 3) {
		if (lua_isnil(L, 3)) {	//only hit this if 'nil' is explicitly passed as arg[3] ... not if it's left blank (hence the 'argc >= 3')
			cl->inkey[clientKeyIndex] = 0;
		} else {
			const char *str = lua_tostring(L, 3);
			if (!str) luaL_error(L, "param 3 expected to be a string.");
			int newkeycode = getInputKeyCodeForName(str);
			if (!newkeycode) luaL_error(L, "unable to translate key code name %s\n", str);
			
			cl->inkey[clientKeyIndex] = newkeycode;
		}
	}
	
	//todo - getNameForInputKeyCode ?
	if (keycode < 0) {
		static const char *mouseButtonNames[3] = {
			"mouseleft",
			"mouseright",
			"mousemiddle"
		};
		keycode = ~keycode;
		if (keycode >= numberof(mouseButtonNames)) {
			lua_pushnumber(L, keycode);
		} else {
			lua_pushstring(L, mouseButtonNames[keycode]);
		}
	} else {
		if (keycode >= Key::COUNT) {
			lua_pushnumber(L, keycode);
		} else {
			lua_pushstring(L, Key::name[keycode]);
		}
	}
	
	return 1;
}

//cfunc
void CFunc_clientConnect(int argc, char **argv) {
	int count = 0;
	const char *dest = NULL;

	if (argc > 1) count = atoi(argv[1]);
	if (!count) count = clientSys->local.clientCount;
//	if (!count) count = 1;
	if (argc > 2) dest = argv[2];

	clientSys->connect(count, dest);
}
static CVarw cv_clientConnect(0, "connect", CFunc_clientConnect);

void CFunc_clientDisconnect(int argc, char **argv) {
	//todo - in the future - allow selective local client disconnections?
	//so its not just an all-or-none disconnection
	clientSys->checkDisconnect();
}
static CVarw cv_clientDisconnect(0, "disconnect", CFunc_clientDisconnect);


static void CVar_EditResetPos(int, char**) {
	if (editor.isActive()) {
		int i = 0;
		LocalClient *c = NULL;
		for (i = 0, c = clientSys->local.clients; i < clientSys->local.clientCount; i++, c++) {
			c->viewpos = 0.f;
		}
	}
}
static CVarw cvarEditorResetPos(0, "editorResetPos", CVar_EditResetPos);
