/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
//TODO - test visibility of sprites/lights to screen before
//adding them to the buffer.  if fail then dont even add them.

#include <iostream>

#include "system/gl.h"
#include <math.h>

#include "system/memory.h"
#include "math/vec.h"
#include "base/base.h"	//cvar
#include "shader/shaderext.h"

#include "main.h"		//game_thistime - for shaders
#include "client.h"
#include "draw.h"
#include "lighting.h"
#include "texloader.h"

using namespace std;

//cvars
static CVari useBumpMapping("useBumpMapping", 1);
static CVari useLighting("useLighting", 1);
static CVari showLightMap("showLightMap", 0);
static CVari renderMaxLights("renderMaxLights", -1);
static CVarf depthClear("depthClear", -5.f);
static CVarf depthWrite("depthWrite", 5.f);
static CVarf lightZ("lightZ", 0.f);
static CVari renderMethod("renderMethod", 0);		//2 for good bump mapping.  requires dest alpha.  THANKS SDL =P.  3 for our unified light & transparency system (go fbos!) but doesn't work with bump-mapping quite yet ...
static CVari geomOutsideScreen("geomOutsideScreen", 0);
static CVari geomCounter("geomCounter", 0);
static CVari showTangentSpace("showTangentSpace", 0);
static CVari printRenderInfo("printRenderInfo", 0);

static CVari overrideDepth("overrideDepth", 0);
static CVarf overrideDepthValue("overrideDepthValue", 1.f);	//scaled by the geom's index.  1 for fwd, -1 for backwds rendering

static vec4f templight_diffuse(1.f);
static vec4f templight_pos(0,0,0,1.f);
static vec3f templight_dir(1.f,0,0);	//dont use Z
static CVarf templight_cutoff("templight_cutoff", 180.f);
static CVarf templight_attn_const("templight_attn_const", 1.f);
static CVarf templight_attn_lin("templight_attn_lin", 0.f);
static CVarf templight_attn_quad("templight_attn_quad", 0.f);

static CVarw cv1(&templight_diffuse, "templight_diffuse");
static CVarw cv2(&templight_pos, "templight_pos");
static CVarw cv3(&templight_dir, "templight_dir");

//textureid
static GLuint normalizeCubeMap = 0;

#define GEOM_COUNT		2048

class Geometry : public EGeometry {
public:
	//all below is aux data computed after addGeom is done

//screen bounds
	box2f bbox;
//bintree linking - based upon depth (for sorting's sake) - sorted least to greatest
	Geometry *left, *right;
	int treeDepth;
//hmm, can I hack a 2-way linked list into the tree?
	Geometry *prev, *next;
};

static Geometry *binTreeRoot = NULL;	//null every frame.  built upon DrawGeometry
static Geometry *linkedListHead = NULL;	//same deal here
static Geometry *linkedListTail = NULL;

static int numGeom = 0;
static Geometry g_buffer[GEOM_COUNT];

static Geometry *GeometryBuffer() {
	if (numGeom >= GEOM_COUNT) {
		con.print("geometry buffer overflow!\n");
		numGeom = GEOM_COUNT-1;
	}
	Geometry *g = g_buffer + numGeom;
	numGeom++;
	return g;
}


// *************************** MD2SHADER RIP

static void getCubeVector(int i, int cubesize, int x, int y, float *v) {
	float s, t, sc, tc, mag;

	s = ((float)x + 0.5f) / (float)cubesize;
	t = ((float)y + 0.5f) / (float)cubesize;
	sc = s*2.0f - 1.0f;
	tc = t*2.0f - 1.0f;

	switch (i) {
	case 0:    v[0] = 1.0f;	v[1] = -tc;		v[2] = -sc;		break;
	case 1:    v[0] = -1.0f;	v[1] = -tc;		v[2] = sc;		break;
	case 2:    v[0] = sc;		v[1] = 1.0f;	v[2] = tc;		break;
	case 3:    v[0] = sc;		v[1] = -1.0f;	v[2] = -tc;		break;
	case 4:    v[0] = sc;		v[1] = -tc;		v[2] = 1.0f;	break;
	case 5:    v[0] = -sc;	v[1] = -tc;		v[2] = -1.0f;	break;
	}

	mag = 1.0f/(float)sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
	v[0] *= mag;
	v[1] *= mag;
	v[2] *= mag;
}

/* Initialize a cube map texture object that generates RGB values
 * that when expanded to a [-1,1] range in the register combiners
 * form a normalized vector matching the per-pixel vector used to
 * access the cube map.
 */
bool makeNormalizeVectorCubeMap(int size)
{
	float v[3];
	int i, x, y;
	GLubyte *pixels;

	pixels = (GLubyte*)MEMALLOCNAMED(size*size*3, "Normalized Vector Cube Map");
	if (pixels == NULL) {
		con.print("normalmap couldnt be allocated\n");
		return false;
	}

	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	for (i = 0; i < 6; i++) {
	for (y = 0; y < size; y++) {
		for (x = 0; x < size; x++) {
			getCubeVector(i, size, x, y, v);
			pixels[3*(y*size+x) + 0] = 128 + (unsigned char)(127.f*v[0]);
			pixels[3*(y*size+x) + 1] = 128 + (unsigned char)(127.f*v[1]);
			pixels[3*(y*size+x) + 2] = 128 + (unsigned char)(127.f*v[2]);
		}
	}
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB+i, 0, GL_RGB8,
		size, size, 0, GL_RGB, GL_UNSIGNED_BYTE, pixels);
	}

	memFree(pixels);

	return true;
}

// *************************** END MD2SHADER RIP


//needed constants
#define GL_COMBINE_ARB			0x8570
#define GL_COMBINE_RGB_ARB		0x8571
#define GL_COMBINE_ALPHA_ARB	0x8572
#define GL_CONSTANT_ARB			0x8576
#define GL_PRIMARY_COLOR_ARB	0x8577
#define GL_PREVIOUS_ARB			0x8578
#define GL_SOURCE0_RGB_ARB		0x8580
#define GL_SOURCE1_RGB_ARB		0x8581
#define GL_SOURCE0_ALPHA_ARB	0x8588
#define GL_SOURCE1_ALPHA_ARB	0x8589
#define GL_OPERAND0_RGB_ARB		0x8590
#define GL_OPERAND1_RGB_ARB		0x8591
#define GL_OPERAND0_ALPHA_ARB	0x8598
#define GL_OPERAND1_ALPHA_ARB	0x8599
#define GL_DOT3_RGB_ARB			0x86AE
#define GL_DOT3_RGBA_ARB		0x86AF


// ***************************************************************************
// ************************* RENDER ALGORITHM 1 CODE *************************
// ***************************************************************************

// no bump mapping
// but accelerated lighting

static void Display_Render1(const box2f &viewarea) {

	unsigned int l;
	int i;
	Light *light;
	Geometry *g;
	const vec2f *v, *t;

	if (glActiveTextureARB) glActiveTextureARB(GL_TEXTURE0_ARB);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);

	if (!useLighting) {
		//fullbright the light
		glClearColor(1,1,1,0);
	} else {
		//yeah, i know, bad idea - this doesnt clear until the next pass
		//fine since for the most part light_ambient_level is constant
		vec3f ambient = lightSys.getAmbientLevel();
		glClearColor(VEC3ELEM(ambient), 0);

		//first - additive blend all lights into destination color buffer
		glDisable(GL_DEPTH_TEST);
		glBlendFunc(GL_ONE, GL_ONE);
		for (l = 0, light =	lightSys.getBase();
			l < lightSys.getCount() && l < (unsigned int)renderMaxLights;
			l++, light++)
		{
			if (printRenderInfo) con << " light " << l << " render depth " << (float)depthClear << "\n";

			//and do the dirty work...
			glColor3fv(light->color);

			TexLoad *tex = texSys.getObj(light->tex);
			if (tex) glBindTexture(GL_TEXTURE_2D, tex->tex.getTextureID());
			else glDisable(GL_TEXTURE_2D);

			glBegin(GL_QUADS);
			glTexCoord2f(0, 0);	glVertex3f(light->getCoords()[0][0], light->getCoords()[0][1], depthClear);
			glTexCoord2f(1, 0);	glVertex3f(light->getCoords()[1][0], light->getCoords()[1][1], depthClear);
			glTexCoord2f(1, 1);	glVertex3f(light->getCoords()[2][0], light->getCoords()[2][1], depthClear);
			glTexCoord2f(0, 1);	glVertex3f(light->getCoords()[3][0], light->getCoords()[3][1], depthClear);
			glEnd();
			//shutdown

			if (!tex) glEnable(GL_TEXTURE_2D);
		}

		glEnable(GL_DEPTH_TEST);
	}
	
	if (showLightMap) return;

	glBlendFunc(GL_DST_COLOR, GL_ZERO);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glAlphaFunc(GL_GREATER, 0.5);	//for when its used
	
	int s = numGeom-1;

	for (g = linkedListHead; g; g = g->next, s--) {	//prev means towards the head, right?
//	for (g = g_buffer; g < g_buffer + numGeom; g++, s--) {

		if (printRenderInfo)
			con << "geom " << (g - g_buffer)
				<< " tree depth " << g->treeDepth
				<< " sprite depth " << g->depth
				<< " geom vtxs "	<< g->vertex[0]
							<< " "	<< g->vertex[1]
							<< " "	<< g->vertex[2]
							<< " "	<< g->vertex[3]
				<< "\n";

		//only alpha test when we need to
		if (g->blend != Blend()) continue;
		if (g->use_alpha_test) glEnable(GL_ALPHA_TEST);
		if (g->use_lighting) glEnable(GL_BLEND);		//only light what we want to:

		//otherwise we always write it thru
		glColor4ubv(g->color);
		glBindTexture(GL_TEXTURE_2D, g->colormap);
		glBegin(GL_QUADS);
		for (i = 0, v = g->vertex, t = g->texcoord; i < 4; i++, v++, t++) {
			glTexCoord2fv(*t);
			glVertex3f(v->x, v->y, (float)s);
		}
		glEnd();
		//cleanup afterwards
		glDisable(GL_BLEND);
		glDisable(GL_ALPHA_TEST);
		
//		break;
	}

	s = 0;
	glDepthFunc(GL_ALWAYS);
	glEnable(GL_BLEND);

	//and once again for the additive blends & transparencies
	for (g = linkedListTail; g; g = g->prev, s++) {	//prev means towards the head, right?
//	for (g = g_buffer; g < g_buffer + numGeom; g++, s--) {

		if (printRenderInfo) 
			con << "alpha geom " << (g - g_buffer)
				<< " tree depth " << g->treeDepth
				<< " sprite depth " << g->depth
				<< " geom vtxs "	<< g->vertex[0]
							<< " "	<< g->vertex[1]
							<< " "	<< g->vertex[2]
							<< " "	<< g->vertex[3]
				<< "\n";

		if (g->blend == Blend()) continue;		
		glBlendFunc(g->blend.src(), g->blend.dst());

		//otherwise we always write it thru
		glColor4ubv(g->color);
		glBindTexture(GL_TEXTURE_2D, g->colormap);
		glBegin(GL_QUADS);
		for (i = 0, v = g->vertex, t = g->texcoord; i < 4; i++, v++, t++) {
			glTexCoord2fv(*t);
			glVertex3f(v->x, v->y, (float)s);
		}
		glEnd();

		//cleanup afterwards
		glDisable(GL_ALPHA_TEST);
	}

	glDisable(GL_BLEND);
	glDepthFunc(GL_LEQUAL);	//default

	if (printRenderInfo) printRenderInfo = 0;
}

// ***************************************************************************
// *********************** END RENDER ALGORITHM 1 CODE ***********************
// ***************************************************************************

// ***************************************************************************
// ************************* RENDER ALGORITHM 2 CODE *************************
// ***************************************************************************

//burlier
//requires dest alpha
//(not supported by sdl osx)

static void Display_Render2(const box2f &viewarea) {

	unsigned int l;
	int s,i;
	Light *light;
	Geometry *g;
	float tstc[3];
	const vec2f *v, *t;

	//yeah, i know, bad idea - this doesnt clear until the next pass
	//fine since for the most part light_ambient_level is constant
	vec3f ambient = lightSys.getAmbientLevel();
	glClearColor(VEC3ELEM(ambient), 0);

	//first - before all else - write correct depth values to depth buffer
	//dont write any colors or alphas at all...
	//cycle thru all sprites
	glColorMask(0,0,0,0);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glDepthMask(1);
	glAlphaFunc(GL_GREATER, 0.5);

	//alright, alpha testing on all is messing up - keeping the background from rendering
	glEnable(GL_ALPHA_TEST);

//	for (s = numGeom-1, g = &g_buffer[numGeom-1]; s >= 0; s--, g--) {
	for (s = numGeom-1, g = linkedListHead; g; g = g->next, s--) {
		//only alpha test when we need to
//		if (g->use_alpha_test) glEnable(GL_ALPHA_TEST);
		//otherwise we always write it thru
		glColor4ubv(g->color);
		glBindTexture(GL_TEXTURE_2D, g->colormap);
		glBegin(GL_QUADS);
		for (i = 0, v = g->vertex, t = g->texcoord; i < 4; i++, v++, t++) {
			glTexCoord2fv(*t);
			glVertex3f(v->x, v->y, (float)s);
		}
		glEnd();
		//cleanup afterwards
//		glDisable(GL_ALPHA_TEST);
	}
	glDisable(GL_ALPHA_TEST);

	//from now on... either use GL_EQUAL depth testing or none at all
	//and DONT WRITE to the depth mask
	glDepthMask(0);

	if (useLighting) {
		for (l = 0, light = lightSys.getBase();
			l < lightSys.getCount() && l < (unsigned int)renderMaxLights;
			l++, light++)
		{
			//if the light has no color then bail out
			if (light->color[0] == 0 &&
				light->color[1] == 0 &&
				light->color[2] == 0) continue;

			//if the light doesnt cross viewspace then continue
			if (!light->getBBox().touches(viewarea)) continue;

			//unit1 - enable cube mapping for dot3 pass
			if (useBumpMapping) {

				glColor3f(1,1,1);

				glActiveTextureARB(GL_TEXTURE1_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_DOT3_RGBA_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_PREVIOUS_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
				glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, normalizeCubeMap);
				glEnable(GL_TEXTURE_CUBE_MAP_ARB);

				//unit0 - set blending to replace
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
				glEnable(GL_TEXTURE_2D);

				//only let pre-touched tiles drawn
				glDepthFunc(GL_EQUAL);

				//only write to alpha
				glColorMask(0,0,0,1);

if (showTangentSpace) {
//set to replace w/the cubemap normal texture
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
	glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_REPLACE);
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE);
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
	glActiveTextureARB(GL_TEXTURE0_ARB);
//enable write with all color channels
	glColorMask(1,1,1,1);
}

				//cycle thru all sprites
//				for (s = numGeom-1, g = &g_buffer[numGeom-1]; s >= 0; s--, g--) {
				for (s = numGeom-1, g = linkedListHead; g; g = g->next, s--) {

					//if the light doesnt cross sprite space then continue
					if (!light->getBBox().touches(g->bbox)) continue;

					if (g->use_bump_mapping && g->use_lighting) {
						//enable the cube map
						glActiveTextureARB(GL_TEXTURE1_ARB);
						glEnable(GL_TEXTURE_CUBE_MAP_ARB);
						//enable the bump map
						glActiveTextureARB(GL_TEXTURE0_ARB);
						glEnable(GL_TEXTURE_2D);
						glBindTexture(GL_TEXTURE_2D, g->bumpmap);
					} else {
						//disable cube mapping
						glActiveTextureARB(GL_TEXTURE1_ARB);
						glDisable(GL_TEXTURE_CUBE_MAP_ARB);
						//disable bump mapping
						glActiveTextureARB(GL_TEXTURE0_ARB);
						glDisable(GL_TEXTURE_2D);
					}
					//only write alpha thru
					glBegin(GL_QUADS);
					for (i = 0, v = g->vertex, t = g->texcoord; i < 4; i++, v++, t++) {
						if (g->use_bump_mapping) {
							//x,y - flip
							tstc[0] = (v->x - light->spot.x) * g->flip;
							tstc[1] = (v->y - light->spot.y);
							//cplxmul
							tstc[2] = tstc[0] * g->dir.x - tstc[1] * g->dir.y;		//use 'z' as the reg for cplx mul
							tstc[1] = tstc[0] * g->dir.y + tstc[1] * g->dir.x;		//inverse-rotate (negative sinom)
							tstc[0] = tstc[2];
							//z
							tstc[2] = lightZ - 10.f;

							glMultiTexCoord2fvARB(GL_TEXTURE0_ARB, *t);
							glMultiTexCoord3fvARB(GL_TEXTURE1_ARB, tstc);
						}
						glVertex3f(v->x, v->y, (float)s);
					}
					glEnd();
				}
			}

if (!showTangentSpace) {

			glDepthFunc(GL_ALWAYS);

			//disable cube mapping - but dont switch texenv's just yet...
			glActiveTextureARB(GL_TEXTURE1_ARB);
			glDisable(GL_TEXTURE_CUBE_MAP_ARB);
			//set up our lighting blend write to dest alpha...
			glActiveTextureARB(GL_TEXTURE0_ARB);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glEnable(GL_BLEND);
			if (useBumpMapping)	glBlendFunc(GL_DST_ALPHA, GL_ONE);
			else					glBlendFunc(GL_ONE, GL_ONE);
			glColorMask(1,1,1,1);
			//and do the dirty work...
			glColor3fv(light->color);

			TexLoad *tex = texSys.getObj(light->tex);
			glBindTexture(GL_TEXTURE_2D, tex ? tex->tex.getTextureID() : 0);

			glBegin(GL_QUADS);
			glTexCoord2f(0, 0);	glVertex3f(light->getCoords()[0][0], light->getCoords()[0][1], depthClear);
			glTexCoord2f(1, 0);	glVertex3f(light->getCoords()[1][0], light->getCoords()[1][1], depthClear);
			glTexCoord2f(1, 1);	glVertex3f(light->getCoords()[2][0], light->getCoords()[2][1], depthClear);
			glTexCoord2f(0, 1);	glVertex3f(light->getCoords()[3][0], light->getCoords()[3][1], depthClear);
			glEnd();
			//shutdown
			glDisable(GL_BLEND);
}

		}
	}

	glColorMask(1,1,1,1);

if (!showTangentSpace) {
	if (!showLightMap) {

		glDepthFunc(GL_EQUAL);
		//blend sprite color w/dest alpha
		glBlendFunc(GL_DST_COLOR, GL_ZERO);
//		for (s = numGeom-1, g = &g_buffer[numGeom-1]; s >= 0; s--, g--) {
		for (s = numGeom-1, g = linkedListHead; g; g = g->next, s--) {
			if (useLighting && g->use_lighting) glEnable(GL_BLEND);
			glColor4ubv(g->color);
			glBindTexture(GL_TEXTURE_2D, g->colormap);
			glBegin(GL_QUADS);
			for (i = 0, v = g->vertex, t = g->texcoord; i < 4; i++, v++, t++) {
				glTexCoord2fv(*t);
				glVertex3f(v->x, v->y, (float)s);
			}
			glEnd();
			glDisable(GL_BLEND);
		}
	}
}
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glDisable(GL_TEXTURE_CUBE_MAP_ARB);
	glDisable(GL_TEXTURE_2D);
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glEnable(GL_TEXTURE_2D);

	glDepthMask(1);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_ALPHA_TEST);
	glDepthFunc(GL_ALWAYS);
}

// **************************a*************************************************
// *********************** END RENDER ALGORITHM 2 CODE ***********************
// ***************************************************************************


//for the render method 3's ...
//todo - subclass alot
//that way we could just subclass out the fbo render algorithms
//and that way we could subclass the endGeom()'s (and add in multiple linked lists ... but that'd require a Geometry structure change atm)

//afaik you only need one fbo object and from which you can attach-reattach dif texs to it
//so long as all the channels match up
GLuint lightFboID = 0;
const int lightFboSize = 512;
Texture lightFboTex;

GLuint screenFboID = 0;
vec2i screenFboSize(1024,512);	//this is why games don't allow in-game resizing
	//set to the two tex's above (im too lazy to free them so i do this instead) - and ref'd for reading by screenFboTex[!screenFboWriteBuf]
Texture screenFboTex[3];	//method3 uses 2.  method4 uses 3.
static int screenFboWriteBuf = 0;	//whether we're on the front or back buffer of the screen FBO

static bool checkFBO(unsigned int fbo) {
	int fboStatus = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (fboStatus != GL_FRAMEBUFFER_COMPLETE_EXT) {
		con << "checkfbo failed for fbo " << fbo << " with status " << fboStatus << "\n";
		return false;
	}
	return true;
}

// ***************************************************************************
// ************************* RENDER ALGORITHM 3 CODE *************************
// ***************************************************************************

static CVari useFBOFX("useFBOFX", 1);

//here's my catch:
//swapping screen fbos between two textures is fast
//so fast that when we're writing to the first, we're reading from the second before the second is even done
//and as a result we don't get all the rendered information (unless we stop the process by setting useFBOFX to 0)
//so how do we ensure flushes? or how are we supposed to be doing fbo double-buffered swaps?
//separate pre-and post-fbo accumulation

//i got a bad slowdown with my shader water, but its not from this ...
#define SEPARATE_FBO_AND_POST

void displayRender3_drawSprites(bool postFBO, const vec4f &texgenu, const vec4f &texgenv) {

	glActiveTextureARB(GL_TEXTURE1_ARB);
	glBindTexture(GL_TEXTURE_2D, lightFboTex.getTextureID());

	//bind unit3 to the read fbo
	glActiveTextureARB(GL_TEXTURE3_ARB);
	glBindTexture(GL_TEXTURE_2D, screenFboTex[!screenFboWriteBuf].getTextureID());

	glActiveTextureARB(GL_TEXTURE0_ARB);

	for (Geometry *g = linkedListTail; g; g = g->prev) {	//prev means towards the head, right?

#ifdef SEPARATE_FBO_AND_POST
		if (g->postFBO != postFBO) continue;
#endif

		if (printRenderInfo)
			con << "geom " << (g - g_buffer)
				<< " tree depth " << g->treeDepth
				<< " sprite depth " << g->depth
				<< " geom vtxs "	<< g->vertex[0]
							<< " "	<< g->vertex[1]
							<< " "	<< g->vertex[2]
							<< " "	<< g->vertex[3]
				<< "\n";
				
		//only alpha test when we need to
		if (g->use_alpha_test) glEnable(GL_ALPHA_TEST);

		if (g->blend != Blend()) {	//if we have a non-default value ...
			glEnable(GL_BLEND);
			glBlendFunc(BlendSFactor(g->blend), BlendDFactor(g->blend));
		}
		
		if (g->shader) {
			const ShaderObj *so = g->shader;
			glUseProgramObjectARB(so->getProg());

			//see if we're expecting any engine presets
			//todo - prefix these to let them know they're engine-hardcoded?
			if (so->getTimeLoc() != -1) glUniform1fARB(so->getTimeLoc(), game_thistime);

			//i should have a max texunit test
			//only bind the bumpmap for shaders
			glActiveTextureARB(GL_TEXTURE2_ARB);
			glBindTexture(GL_TEXTURE_2D, g->bumpmap);		//bind but don't enable - solely for shader's sake

		} else {
			//use_lighting is a flag for the fixed functionality pathway so don't run it in the shader
			//the shader still uses the light texunit1, but lucky for us units don't have to be bound to work
			if (g->use_lighting) {
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glEnable(GL_TEXTURE_2D);
			}
		}

		glActiveTextureARB(GL_TEXTURE0_ARB);
		glBindTexture(GL_TEXTURE_2D, g->colormap);

		//otherwise we always write it thru
		glColor4ubv(g->color);
		glBegin(GL_QUADS);	//everything is quads!!! and all our vtxcounts are 4!!! just get rid of it already!
		vec2f *v = g->vertex;
		vec2f *t = g->texcoord;
		for (int i = 0; i < 4; i++, v++, t++) {
			glTexCoord2fv(*t);
			if (g->shader) {
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 
					v->x * texgenu.x + v->y * texgenu.y + texgenu.w,
					v->x * texgenv.x + v->y * texgenv.y + texgenv.w);
			}
			glVertex2fv(*v);
		}
		glEnd();

		if (g->shader) {
			glUseProgramObjectARB(0);
		} else {
			if (g->use_lighting) {
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE0_ARB);
			}
		}

		//cleanup afterwards
		glDisable(GL_ALPHA_TEST);
		glDisable(GL_BLEND);
	}

	//turn off light framebuffer texunit
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glBindTexture(GL_TEXTURE_2D, 0);

	glActiveTextureARB(GL_TEXTURE3_ARB);
	glBindTexture(GL_TEXTURE_2D, 0);

	glActiveTextureARB(GL_TEXTURE0_ARB);
}

static void Display_Render3(const box2f &viewarea) {

	glAlphaFunc(GL_GREATER, 0.5);	//this should be done once
	
	glDisable(GL_DEPTH_TEST);		//not used for this method (allsprites aresorted anyways)

	GLint drawBuffer;
	glGetIntegerv(GL_DRAW_BUFFER, &drawBuffer);
	
	//first render all lights into our FBO
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, lightFboID);
	if (checkFBO(lightFboID)) {
		glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);

		vec3f ambient = lightSys.getAmbientLevel();
		glClearColor(VEC3ELEM(ambient), 0);
		glClear(GL_COLOR_BUFFER_BIT);	//lightFBO has no depth components

		//draw to our fbo.  don't forget to setup (push/pop) viewport, etc
		glPushAttrib(GL_VIEWPORT_BIT);
		glViewport(0,0,lightFboSize,lightFboSize);
		//do we need to re-ortho the screen?
		
		glBlendFunc(GL_ONE, GL_ONE);
		glEnable(GL_BLEND);
		
		int l = 0;
		for (Light *light = lightSys.getBase(); l < lightSys.getCount() && l < (unsigned int)renderMaxLights; l++, light++) {
			if (printRenderInfo) con << " light " << l << " render depth " << (float)depthClear << "\n";

			//and do the dirty work...
			glColor3fv(light->color);

			TexLoad *tex = texSys.getObj(light->tex);
			if (tex) glBindTexture(GL_TEXTURE_2D, tex->tex.getTextureID());
			else glDisable(GL_TEXTURE_2D);	//there really had better be a texture with our light

			//huh, this should be quads ... for speed ... then again we're rebinding every frame ...
			glBegin(GL_QUADS);
			glTexCoord2f(0, 0);	glVertex2fv(light->getCoords()[0]);
			glTexCoord2f(1, 0);	glVertex2fv(light->getCoords()[1]);
			glTexCoord2f(1, 1);	glVertex2fv(light->getCoords()[2]);
			glTexCoord2f(0, 1);	glVertex2fv(light->getCoords()[3]);
			glEnd();
			//shutdown

			if (!tex) glEnable(GL_TEXTURE_2D);
		}
		
		glDisable(GL_BLEND);
		
		glPopAttrib();

	}
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	
	//is there any harm using one fbo object to render to two textures?
	//back to back?
	if (useFBOFX) {
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, screenFboID);
		if (!checkFBO(screenFboID)) return;
		glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT+screenFboWriteBuf);

		//now to use the other texture (!screenFBO) for reading

		glPushAttrib(GL_VIEWPORT_BIT);
		glViewport(0,0,VEC2ELEM(screenFboSize));
	} else {
		glDrawBuffer(drawBuffer);	//reset the draw buffer
	}
	
	//this is all set up for use throughout the rest of the render pathway (in- and out-of-fbo)
	vec2f viewsize = viewarea.size();
	vec4f texgenu(1.f/viewsize.x, 0.f, 0.f, -viewarea.min.x/viewsize.x);
	vec4f texgenv(0.f, -1.f/viewsize.y, 0.f, viewarea.max.y/viewsize.y);

	//set up our light framebuffer (right off that bat ... hmm )
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	//it would be cool if we could set up our texgen for unit1 to map stuff from camera coords right into our fbo tex
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGenfv(GL_S, GL_OBJECT_PLANE, texgenu);
	glTexGenfv(GL_T, GL_OBJECT_PLANE, texgenv);	//y is flipped from framebuffer coordinate space
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glActiveTextureARB(GL_TEXTURE0_ARB);
	
	//now we draw all sprites to the screen, normally, full light
	//we don't if you want to just see the lightmap
	if (showLightMap) {
		glBindTexture(GL_TEXTURE_2D, 0);	//if unit0 tex2s is disabled then does it pass the primary to unit1? or does unit1 get cut?
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glEnable(GL_TEXTURE_2D);
		
		//overlay the lightmap (todo - just wire this in place of the fbo-to-screen draw?)
		glColor3f(1,1,1);
		glBegin(GL_QUADS);
		glTexCoord2f(0,0);	glVertex2f(viewarea.min.x, viewarea.min.y);
		glTexCoord2f(1,0);	glVertex2f(viewarea.max.x, viewarea.min.y);
		glTexCoord2f(1,1);	glVertex2f(viewarea.max.x, viewarea.max.y);
		glTexCoord2f(0,1);	glVertex2f(viewarea.min.x, viewarea.max.y);
		glEnd();
		
		glDisable(GL_TEXTURE_2D);
		glActiveTextureARB(GL_TEXTURE0_ARB);
	} else {
		//draw our sprites
		displayRender3_drawSprites(false, texgenu, texgenv);
	}
	

	if (useFBOFX) {
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
		glDrawBuffer(drawBuffer);
		glPopAttrib();

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0,1,0,1,-1,1);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		glColor3f(1,1,1);
		glBindTexture(GL_TEXTURE_2D, screenFboTex[screenFboWriteBuf].getTextureID());
		glBegin(GL_QUADS);
		for (int i = 0; i < 4; i++) {
			int yofs = i>>1;
			int xofs = (i&1) ^ yofs;
			glTexCoord2i(xofs,yofs); glVertex2i(xofs,yofs);
		}
		glEnd();
		
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		//swap fbo texs
		screenFboWriteBuf = !screenFboWriteBuf;
	}

#ifdef SEPARATE_FBO_AND_POST
	displayRender3_drawSprites(true, texgenu, texgenv);
#endif

	glEnable(GL_DEPTH_TEST);
	
	if (printRenderInfo) printRenderInfo = 0;
}

// ***************************************************************************
// *********************** END RENDER ALGORITHM 3 CODE ***********************
// ***************************************************************************

// ***************************************************************************
// ************************* RENDER ALGORITHM 4 CODE *************************
// ***************************************************************************

/*
this one is like 3 except
it pushes stuff back to the read buffer after every SPRITE that uses it (not just every frame)
*/

static void displayRender4_drawSprites(bool postFBO, const vec4f &texgenu, const vec4f &texgenv) {

	//turn on the texgen planes for lighting - for the non-shader sprites
	//but don't turn lighting on itself unless the sprite requests it
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glBindTexture(GL_TEXTURE_2D, lightFboTex.getTextureID());
	
/*
bind screen fbo {

bind screenRead to lastscreentex texunit			-- (if it's not already bound there)
set the drawbuffer to screenWrite

for all sprites
	if our sprite has a shader and it uses the uniform "currentscreentex" {
		bind currentscreentex
		copy screenWrite back into currentscreentex		-- copy what we have so far into the screenRead buffer
	}
	render sprite		-- it ends up going in the screenWrite fbo
} unbind screen fbo

swap screenWrite and screenLast -- since we still need image flow capapbilities (recurrence of texture application - 2-stroke system)

here's the indexing:
screenWrite = screenFboTex[screenFboWriteBuf]
screenLast  = screenFboTex[!screenFboWriteBuf]	just like method 3
screenRead  = screenFboTex[2]
*/
	//unit3 goes to the read fbo
	glActiveTextureARB(GL_TEXTURE3_ARB);
	glBindTexture(GL_TEXTURE_2D, screenFboTex[!screenFboWriteBuf].getTextureID());
	
	//unit4 goes to the current read fbo
	//(which gets re-flushed before running any shader that requests it
	//i.e. request it only if you need it
	glActiveTextureARB(GL_TEXTURE4_ARB);
	glBindTexture(GL_TEXTURE_2D, screenFboTex[2].getTextureID());

	for (Geometry *g = linkedListTail; g; g = g->prev) {	//prev means towards the head, right?
		assert(g->prev != g);	//why would this happen? but it does ...
		if (g->postFBO != postFBO) continue;

		if (printRenderInfo)
			con << "geom " << (g - g_buffer)
				<< " tree depth " << g->treeDepth
				<< " sprite depth " << g->depth
				<< " geom vtxs "	<< g->vertex[0]
							<< " "	<< g->vertex[1]
							<< " "	<< g->vertex[2]
							<< " "	<< g->vertex[3]
				<< "\n";
				
		//only alpha test when we need to
		if (g->use_alpha_test) glEnable(GL_ALPHA_TEST);

		if (g->blend != Blend()) {	//if we have a non-default value ...
			glEnable(GL_BLEND);
			glBlendFunc(BlendSFactor(g->blend), BlendDFactor(g->blend));
		}
		
		if (g->shader) {
			const ShaderObj *so = g->shader;
			glUseProgramObjectARB(so->getProg());
			
			//unit0: colormap will be bound
			//unit1: lightmap is alerady bound
			//unit2: bumpmap will be bound upon request
			//unit3: last framebuffer is already bound
			//unit4: mid-read buffer is already bound and will be refreshed on request)

			//see if we're expecting any engine presets
			//todo - prefix these to let them know they're engine-hardcoded?
			if (so->getTimeLoc() != -1) {
				glUniform1fARB(so->getTimeLoc(), game_thistime);
			}

			//does the sprite want to read from the current framebuffer or the previous?
			if (so->getCurrentScreenTexLoc() != (GLint)-1) {
				//fill buffer2 with our current framebuffer (screenFboWriteBuf)
				//read into our junk
				glActiveTextureARB(GL_TEXTURE4_ARB);
				glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, VEC2ELEM(screenFboSize));
			}

			if (so->getBumpTexLoc() != (GLint)-1) {
				//i should have a max texunit test
				//only bind the bumpmap for shaders
				//i should also only bind this if i need to
				glActiveTextureARB(GL_TEXTURE2_ARB);
				glBindTexture(GL_TEXTURE_2D, g->bumpmap);		//bind but don't enable - solely for shader's sake
			}
		} else {
			//use_lighting is a flag for the fixed functionality pathway so don't run it in the shader
			//the shader still uses the light texunit1, but lucky for us units don't have to be bound to work
			if (g->use_lighting) {
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glEnable(GL_TEXTURE_2D);
			}
		}

		glActiveTextureARB(GL_TEXTURE0_ARB);
		glBindTexture(GL_TEXTURE_2D, g->colormap);

		//otherwise we always write it thru
		glColor4ubv(g->color);
		glBegin(GL_QUADS);	//everything is quads!!! and all our vtxcounts are 4!!! just get rid of it already!
		vec2f *v = g->vertex;
		vec2f *t = g->texcoord;
		for (int i = 0; i < 4; i++, v++, t++) {
			glTexCoord2fv(*t);
			if (g->shader) {
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 
					v->x * texgenu.x + v->y * texgenu.y + texgenu.w,
					v->x * texgenv.x + v->y * texgenv.y + texgenv.w);
			}
			glVertex2fv(*v);
		}
		glEnd();

		if (g->shader) {
			glUseProgramObjectARB(0);
		} else {
			if (g->use_lighting) {
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB(GL_TEXTURE0_ARB);
			}
		}

		//cleanup afterwards
		if (g->use_alpha_test) glDisable(GL_ALPHA_TEST);
		if (g->blend != Blend()) glDisable(GL_BLEND);
	}

	//turn off light framebuffer texunit
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glBindTexture(GL_TEXTURE_2D, 0);	//unbind the texture for fbo's concern's sake
	glDisable(GL_TEXTURE_2D);	//in case whoever is next doesn't need it
	glActiveTextureARB(GL_TEXTURE0_ARB);
}

static void Display_Render4(const box2f &viewarea) {

	glAlphaFunc(GL_GREATER, 0.5);	//this should be done once
	
	glDisable(GL_DEPTH_TEST);		//not used for this method (allsprites aresorted anyways)

	GLint drawBuffer;
	glGetIntegerv(GL_DRAW_BUFFER, &drawBuffer);
	
	//first render all lights into our FBO
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, lightFboID);
	if (checkFBO(lightFboID)) {
		glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);

		vec3f ambient = lightSys.getAmbientLevel();
		glClearColor(VEC3ELEM(ambient), 0);
		glClear(GL_COLOR_BUFFER_BIT);	//lightFBO has no depth components

		//draw to our fbo.  don't forget to setup (push/pop) viewport, etc
		glPushAttrib(GL_VIEWPORT_BIT);
		glViewport(0,0,lightFboSize,lightFboSize);
		//do we need to re-ortho the screen?
		
		glBlendFunc(GL_ONE, GL_ONE);
		glEnable(GL_BLEND);
		
		int l = 0;
		for (Light *light = lightSys.getBase(); l < lightSys.getCount() && l < (unsigned int)renderMaxLights; l++, light++) {
			if (printRenderInfo) con << " light " << l << " render depth " << (float)depthClear << "\n";

			//and do the dirty work...
			glColor3fv(light->color);

			TexLoad *tex = texSys.getObj(light->tex);
			if (tex) glBindTexture(GL_TEXTURE_2D, tex->tex.getTextureID());
			else glDisable(GL_TEXTURE_2D);	//there really had better be a texture with our light

			//huh, this should be quads ... for speed ... then again we're rebinding every frame ...
			glBegin(GL_QUADS);
			glTexCoord2f(0, 0);	glVertex2fv(light->getCoords()[0]);
			glTexCoord2f(1, 0);	glVertex2fv(light->getCoords()[1]);
			glTexCoord2f(1, 1);	glVertex2fv(light->getCoords()[2]);
			glTexCoord2f(0, 1);	glVertex2fv(light->getCoords()[3]);
			glEnd();
			//shutdown

			if (!tex) glEnable(GL_TEXTURE_2D);
		}
		
		glDisable(GL_BLEND);
		
		glPopAttrib();

	}
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	
	//set up our fbo
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, screenFboID);
	if (!checkFBO(screenFboID)) return;
	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT+screenFboWriteBuf);
	
	//for method4 we are going to read from the draw buffer.  is that good?
	GLint readBuffer;
	glGetIntegerv(GL_READ_BUFFER, &readBuffer);
	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT + screenFboWriteBuf);

	//now to use the other texture (!screenFBO) for reading

	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0,0,VEC2ELEM(screenFboSize));
	
	//this is all set up for use throughout the rest of the render pathway (in- and out-of-fbo)
	vec2f viewsize = viewarea.size();
	vec4f texgenu(1.f/viewsize.x, 0.f, 0.f, -viewarea.min.x/viewsize.x);
	vec4f texgenv(0.f, -1.f/viewsize.y, 0.f, viewarea.max.y/viewsize.y);

	//set up our light framebuffer (right off that bat ... hmm )
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	//it would be cool if we could set up our texgen for unit1 to map stuff from camera coords right into our fbo tex
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGenfv(GL_S, GL_OBJECT_PLANE, texgenu);
	glTexGenfv(GL_T, GL_OBJECT_PLANE, texgenv);	//y is flipped from framebuffer coordinate space
	
	//now we draw all sprites to the screen, normally, full light
	//we don't if you want to just see the lightmap
	if (showLightMap) {
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glBindTexture(GL_TEXTURE_2D, 0);	//if unit0 tex2s is disabled then does it pass the primary to unit1? or does unit1 get cut?
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glEnable(GL_TEXTURE_2D);
		
		//overlay the lightmap (todo - just wire this in place of the fbo-to-screen draw?)
		glColor3f(1,1,1);
		glBegin(GL_QUADS);
		glTexCoord2f(0,0);	glVertex2f(viewarea.min.x, viewarea.min.y);
		glTexCoord2f(1,0);	glVertex2f(viewarea.max.x, viewarea.min.y);
		glTexCoord2f(1,1);	glVertex2f(viewarea.max.x, viewarea.max.y);
		glTexCoord2f(0,1);	glVertex2f(viewarea.min.x, viewarea.max.y);
		glEnd();
		
		glDisable(GL_TEXTURE_2D);
		glActiveTextureARB(GL_TEXTURE0_ARB);
	} else {
		displayRender4_drawSprites(false, texgenu, texgenv);
	}

	
	//turn off our fbo
	//restore all the old settings
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glReadBuffer(readBuffer);
	glDrawBuffer(drawBuffer);
	glPopAttrib();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0,1,0,1,-1,1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glColor3f(1,1,1);
	glBindTexture(GL_TEXTURE_2D, screenFboTex[screenFboWriteBuf].getTextureID());
	glBegin(GL_QUADS);
	for (int i = 0; i < 4; i++) {
		int yofs = i>>1;
		int xofs = (i&1) ^ yofs;
		glTexCoord2i(xofs,yofs); glVertex2i(xofs,yofs);
	}
	glEnd();
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	
	//finish it up
	displayRender4_drawSprites(true, texgenu, texgenv);
	

	//fbo swap
	screenFboWriteBuf = !screenFboWriteBuf;

	glEnable(GL_DEPTH_TEST);
	
	if (printRenderInfo) printRenderInfo = 0;
}


// ***************************************************************************
// *********************** END RENDER ALGORITHM 4 CODE ***********************
// ***************************************************************************

static bool hasGeom = false;	//hack to make sure no one makes two grabs before finishing one
//NOTE TO THE OUTSIDE WORLD - DONT CALL newGeom KNOWING YOU STILL NEED TO CALL geomDone() !

class RenderSystemImpl {
public:
	bool sysInit();
	void sysKill();

	EGeometry *newGeom();
	void geomDone(EGeometry *_g);
	void addToTree(Geometry *g);
	void display(const box2f &viewarea);
	void reset();
};

EGeometry *RenderSystemImpl::newGeom() {
	//hack to make sure no two grabs go on without one release between them
	//this'll make sure my bbox.touches(camera) code below doesn't mess anything up
	assert(!hasGeom);
	hasGeom = true;

	//initialize it ... ?
	return GeometryBuffer();
}

/*
viewarea reference means only call this function during the client's render code
*/
void RenderSystemImpl::geomDone(EGeometry *_g) {
	//hack to make sure no two grabs go on without one release between them
	//this'll make sure my bbox.touches(camera) code below doesn't mess anything up
	assert(hasGeom);
	hasGeom = false;
	
	Geometry *g = (Geometry*)_g;
	//test sprite bounds to client view
	g->bbox.set(*g->vertex);
	vec2f *vtxend = g->vertex + 4;
	for (vec2f *v = g->vertex + 1; v < vtxend; v++) g->bbox.stretch(*v);
	if (!g->bbox.touches(clientSys.getViewArea())) {
		//now we can't just 'return' and not add the geom
		//instead we have to cycle the counter back one
		numGeom--;
		//and hope (not mt safe but no worries yet) no one did a newGeom since our own newGeom
		//or we'll be backtracking on good geometry
		//	workaround: an 'active' flag, but who wants to work that into this system?
		return;
	}

	addToTree(g);
}

void RenderSystemImpl::addToTree(Geometry *g) {
//
	g->treeDepth = 0;
	g->left = g->right = NULL;	//dont forget this before attaching it to the tree
	g->prev = g->next = NULL;	//linked list too (nlogn insertion, const-time for-loop access =)
	assert(!!binTreeRoot == !!linkedListHead);
	assert(!!linkedListTail == !!linkedListTail);
	if (!binTreeRoot) {
		binTreeRoot = g;
		linkedListHead = linkedListTail = g;
	} else {
	//todo - get this working
		Geometry *parent = binTreeRoot;
		for(;;) {
			g->treeDepth++;
			//equal cuz keep equality left-most
			//so if we draw back-to-front (greatest depth to lowest) then equality gets drawn later which means drawn closer to the viewer
			float g_depth = g->depth;
			if (overrideDepth) {
				g_depth = overrideDepthValue * numGeom;
			}
			if (g_depth <= parent->depth) {
				if (parent->left) {
					parent = parent->left;
				} else {
					parent->left = g;
					//so we're inserting it at the front of the tree ...
					if (parent->prev) {
						parent->prev->next = g;
					} else {
						linkedListHead = g;
					}
					g->prev = parent->prev;
					g->next = parent;
					parent->prev = g;
					break;
				}
			} else {
				if (parent->right) {
					parent = parent->right;
				} else {
					parent->right = g;
					//linked list
					if (parent->next) {
						parent->next->prev = g;
					} else {
						linkedListTail = g;
					}
					g->next = parent->next;
					g->prev = parent;
					parent->next = g;
					break;
				}
			}
		}
	}
}
	
void RenderSystemImpl::display(const box2f &viewarea) {
	if (geomCounter) con << "numGeom: " << numGeom << "\n";

	switch ((int)renderMethod) {
	case 1:	Display_Render1(viewarea); break;	//as good as it gets without shaders or FBO's.  transparency works at the expense of being lit correctly
	case 2:	Display_Render2(viewarea); break;	//bump mapping.  works at the expense of transparency (only alpha testing works)
	case 3: Display_Render3(viewarea); break;	//render lights into a fbo, then blend it with the rest of the story
	case 4: Display_Render4(viewarea); break;	//just like 3 but by forcing fbo swaps any time the screen fbo is requested this elims the 2-pass requirement to separate image flow from predator invis.
	}
	
	//now that we're done with accum
	//and now that we're done with the chains
	//we can safely clear the iteration vars
	binTreeRoot = NULL;
	linkedListHead = linkedListTail = NULL;
}

void RenderSystemImpl::reset() {
	numGeom = 0;
	geomOutsideScreen = 0;
}

bool RenderSystemImpl::sysInit() {
	renderMethod = 1;
	
	if (found_GL_ARB_texture_cube_map &&
		found_GL_ARB_texture_env_combine)
	{
		//init the light cubemap texture
		glGenTextures(1, &normalizeCubeMap);
		glEnable(GL_TEXTURE_CUBE_MAP_ARB);
		glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, normalizeCubeMap);
		if (!makeNormalizeVectorCubeMap(32)) return false;
		glDisable(GL_TEXTURE_CUBE_MAP_ARB);
		
		renderMethod = 2;
	}
	
	//if we're good to go
	if (found_GL_EXT_framebuffer_object &&
		found_GL_ARB_shader_objects &&
		found_GL_ARB_vertex_shader &&
		found_GL_ARB_vertex_program)
	{
		renderMethod = 4;	//super pathway
		
		//make us a texture while we're at it
		lightFboTex.createEmpty(lightFboSize, lightFboSize, GL_RGB, GL_UNSIGNED_BYTE);
		lightFboTex.setFilter(GL_LINEAR, GL_LINEAR);
		lightFboTex.setWrap2D(GL_REPEAT, GL_REPEAT);
	
		//while we're here, lets make some FBOs ...
		//the light fbo doesn't need a depth buffer, does it?
		glGenFramebuffersEXT(1, &lightFboID);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, lightFboID);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, lightFboTex.getTextureID(), 0);
		checkFBO(lightFboID);	//show any errors
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
		
		for (int i = 0; i < numberof(screenFboTex); i++) {
			screenFboTex[i].createEmpty(VEC2ELEM(screenFboSize), GL_RGB, GL_UNSIGNED_BYTE);		//2 buffers for read/write
			//sets up linear/linear tex filtering by default
			screenFboTex[i].setWrap2D(GL_REPEAT, GL_REPEAT);
			glBindTexture(GL_TEXTURE_2D, 0);
			glDisable(GL_TEXTURE_2D);
		}

		glGenFramebuffersEXT(1, &screenFboID);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, screenFboID);
		for (int i = 0; i < numberof(screenFboTex); i++) {
			glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT + i, GL_TEXTURE_2D, screenFboTex[i].getTextureID(), 0);
		}
		checkFBO(screenFboID);	//show any errors
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
		screenFboWriteBuf = 0;	//write index
	}
	
	con << "using render method " << (int)renderMethod << "\n";

	reset();

	return true;
}

//		how do you kill a framebuffer?
//			with a silver bullet
void RenderSystemImpl::sysKill() {
	//so we dont try doing this when the gl context is gone
	for (int i = 0; i < numberof(screenFboTex); i++) {
		screenFboTex[i].unload();
	}
	
	lightFboTex.unload();
}

RenderSystem::RenderSystem() : impl(new RenderSystemImpl()) {}
RenderSystem::~RenderSystem() { delete impl; }
bool RenderSystem::sysInit() { return impl->sysInit(); }
void RenderSystem::sysKill() { impl->sysKill(); }
void RenderSystem::reset() { impl->reset(); }
void RenderSystem::display(const box2f &viewarea) { impl->display(viewarea); }
EGeometry *RenderSystem::newGeom() { return impl->newGeom(); }
void RenderSystem::geomDone(EGeometry *g) { impl->geomDone(g); }
RenderSystem renderSys;

