/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <typeinfo>
#include <map>
#include <string>

#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "system/gl.h"
#include "system/exception.h"

#include "base/base.h"

#include "entalloc.h"
#include "entconn.h"
#include "world.h"
#include "main.h"
#include "texloader.h"
#include "client.h"
#include "lighting.h"
#include "network.h"
#include "animation.h"
#include "strcasecmp.h"

#include "program.h"
#include "nettables.h"
#include "delta.h"

#include "game/main.h"
#include "game/client.h"

using namespace std;


//enable this to force the packets to be sent even through the local loopback connection
//otherwise the same memory is used for both server state writing and client state reading
static CVari serverForceLoopbackSend("serverForceLoopbackSend", 1);
static CVari serverReportChanges("serverReportChanges", 0);
static CVari connectionCount("connectionCount", 0);



//one sv_listener per socket type
SocketListen *sv_listener[] = {
	new SocketListenLoopback(),
#ifndef OMIT_NETWORKING
	new SocketListenINET(),
#endif //OMIT_NETWORKING
};

class EntitySystemImpl {
public:
	bool sysInit();
	void sysKill();
	bool sessionInit();
	void sessionKill();
	
	bool reconnect();
	void update();
	void sendMessage(entptr_t ep, const std::vector<std::string> &params);
	
	void sendEntDeltaInfo(World *world);
	void sendEntDeltaForConn(World *world, Connection *conn, bool updateClientEntityDelta);
	void recvMessages();
	bool getJoinRequest(int count, sockaddr_in *src_addr, int conn_type);
	bool setRemoteEntity(Connection *c);
	Connection *getConnForAddr(const sockaddr_in *src_addr);
	void sendChangeLevel(Connection *c);
	bool connCanSeeEntity(Entity *e, Connection *conn, MapTagSet *tags);
	void initListen(int port = 0);
	void killListen();
};


// **** system code

void EntitySystemImpl::sendChangeLevel(Connection *c) {
	try {
		World *world = worldSys.getWorld();
		if (!world) throw Exception() << "we are not in a level";
		if (!world->filename.length()) throw Exception() << "our level has no associated filename";

		assert(c->sender);
		//don't change the level if we're already connected =)
		if (c->sender->getType() == SOCKET_LOOPBACK) {
			c->ready = true;
			return;
		}

		//send the "you are connected!"
		//and the "this is your level!"
		//packets separately . . .because they are handled in separate methods on the client side
		//(which we /could/ change . . . )
	//	char buffer[256];
	//	int bufferlen = 0;
		Message<> m;

		//NOTICE - i havent tested this yet =)  make sure messages work with strings . . .
		
		//now to send a string of the map =) hmm... null-terminated? lets try...
		m << (char)NETCMD_SV_SETLEVEL;
		m << world->filename;
		assert(!m.overflow);
		c->ready = false;	//stop sending entity data while we wait for the level to change...
		c->sender->send(m);
	} catch (Exception &t) {
		con << "EntitySystemImpl::sendChangeLevel failed: " << t << endl;
	}
}


// ************ server connection code


//call just after remote connection
//call just before changing levels
bool EntitySystemImpl::setRemoteEntity(Connection *c) {
	Message<> m;	//	char buffer[256];
	m << (char)NETCMD_SV_SET_CLIENTS;
	m << (char)c->count;
	//short *sp = (short *)(buffer + 2);
	//int bufferlen = 2;

	for (int i = 0; i < c->count; i++) {

		//is the index passed through spawnPlayer used anywhere else?
		//because there is a disconnect between conn# and client-per-conn# ...
		entptr_t e = game.spawnPlayer(c - serverConnSys.connections, i);

//		con.print("setRemoteEntity(): connecting client %d:%d to entity %d\n", c - connections, i, e.i);
		
//connection_s::ent write
		c->local[i].ent.i = e.i;
		m << (short)e.i;	//		*sp = e.i		sp++;		bufferlen += 2;
	}
	assert(!m.overflow);
	c->sender->send(m);

	sendChangeLevel(c);

	//set this connection to the state of 'loading level, waiting to send server 'got my stuff''
	return true;
}

bool EntitySystemImpl::getJoinRequest(int count, sockaddr_in *src_addr, int conn_type) {
#ifndef OMIT_NETWORKING
//	con.print("getJoinRequest(%d,%s)\n", count, inet_ntoa(src_addr->sin_addr));
#else //OMIT_NETWORKING
//	con.print("getJoinRequest(%d)\n", count);
#endif //OMIT_NETWORKING

	//TODO - 'c' here should already be created by the 'get new connection' code
	//it should have the 'active' flag set - since it was created
	//the local flag clear - because its not local
	//the ready flag clear - til we send it the map
	//and the 'count' set to zero - until we give counts here...
	//BUT it should 
	Connection *c = serverConnSys.create();
	if (!c) {
		con << "couldn't find any more free client connections!" << endl;
		return false;
	}

	c->count = count;

	if (c->sender && c->sender->isSending()) {
		con << "what is up?  we got a new connection with a socket already open!" << endl;
		c->sender->kill();
	}

//	con.print("getJoinRequest(): initializing a send connection to the client...\n");
	if (c->sender && c->sender->getType() != conn_type) {
		delete c->sender;
		c->sender = NULL;
	}

	switch (conn_type) {
#ifndef OMIT_NETWORKING
	case SOCKET_INET:
		if (!c->sender) c->sender = new SocketSendINET();
		break;
#endif //OMIT_NETWORKING

	case SOCKET_LOOPBACK:
		if (!c->sender) c->sender = new SocketSendLoopback();
		break;

	default:
		con << "Tried to create an unsupported send socket type " << conn_type << "!" << endl;
		serverConnSys.destroy(c);
		return false;
	}

	if (!c->sender) {
		con << "getJoinRequest():  failed to alloc!" << endl;
		//make very sure we dont leave this routine with a connection with null sender
		serverConnSys.destroy(c);
		return false;
	}

	switch (conn_type) {
#ifndef OMIT_NETWORKING
	case SOCKET_INET:
		if (!((SocketSendINET*)c->sender)->init(inet_ntoa(src_addr->sin_addr), 0)) {
			con << "getJoinRequest():  failed to init!" << endl;
			serverConnSys.destroy(c);
			return false;
		}
		break;
#endif //OMIT_NETWORKING
	case SOCKET_LOOPBACK:
		if (!((SocketSendLoopback*)c->sender)->init(LOOPBACK_TYPE_SERVER)) {
			con << "getJoinRequest():  failed to init!" << endl;
			serverConnSys.destroy(c);
			return false;
		}
		break;
	}

	if (!setRemoteEntity(c)) {
		con << "getJoinRequest(): failed to setup the entity" << endl;
		return false;
	}

	//if we're a loopback then tell the client to shift its local pointer to the connection index
	//this way the loopback client and server share the cl_entity and cl_info data
	//and we dont have to send packets between the two
	if (conn_type == SOCKET_LOOPBACK) {
		clientSys.setLocalArrays(c - serverConnSys.connections);
	}

	return true;
}

Connection *EntitySystemImpl::getConnForAddr(const sockaddr_in *src_addr) {
	//pre - check the listen port to make sure its from any connected clients
	for (Connection *conn = serverConnSys.connections; conn < endof(serverConnSys.connections); conn++) {
		if (!conn->active) continue;

		SocketSend *s = conn->sender;
		if (!s) {	//making sure...
			con.print("found an active server connection with no send socket!\n");
			continue;
		}

		switch (s->getType()) {
#ifndef OMIT_NETWORKING
		case SOCKET_INET:
			if (!memcmp(
					&((SocketSendINET*)conn->sender)->getDestAddr()->sin_addr,
					&src_addr->sin_addr,
					sizeof(src_addr->sin_addr)))
			{
				return conn;
			}
			break;
#endif //OMIT_NETWORKING
		case SOCKET_LOOPBACK:
			{
				sockaddr_in test_addr;
				memset(&test_addr, 0, sizeof(test_addr));
				if (!memcmp(src_addr, &test_addr, sizeof(test_addr))) {
					return conn;
				}
			}
			break;
		default:
			con << "getConnForAddr(): unsupported socket type " << s->getType() << "!" << endl;
			continue;
		}
	}
	return NULL;
}


void EntitySystemImpl::recvMessages() {
	Message<PACKET_LENGTH> bm;

	lua_State *L = progSys.getState();

	for (int i = 0; i < numberof(sv_listener); i++) {
		SocketListen *l = sv_listener[i];

		if (!l->isListening()) continue;	//can't listen if we're not listening =) obviously

		sockaddr_in src_addr;

		for(;;) {
			if (l->recv(bm, &src_addr) != 1) break;
			if (!bm.size) break;

			Connection *conn = getConnForAddr(&src_addr);
			if (!conn) {
				if (bm.data[0] == NETCMD_CL_JOIN_REQUEST) {
					//could we be getting a new connection?
					if (bm.size >= 2) {
						int count = bm.data[1];
						//con << "recvMessages(): got join request for " << count << " clients" << endl;
						getJoinRequest(count, &src_addr, i);
					} else {
						con << "recvMessages(): got connect msg with size too small: " << bm.size << endl;
					}
				} else {
#ifndef OMIT_NETWORKING
					con << "recvMessages(): we got an unknown command (" << bm.data[0] << ") from an unknown connection: " << inet_ntoa(src_addr.sin_addr) << endl;
#else //OMIT_NETWORKING
					con << "recvMessages(): we got an unknown command (" << bm.data[0] << ") from an unknown connection" << endl;
#endif //OMIT_NETWORKING
				}

				continue;	//get the next packet from the sv_listener
			}

			int conn_index = conn - serverConnSys.connections;
			assert(conn_index >= 0 && conn_index < numberof(serverConnSys.connections));

	//con.print("getting packet for connection %d\n", conn_index);

			for (char *p = bm.data; p < bm.data + bm.size; ) {
				int len = bm.size - (int)(p - bm.data);

				switch (p[0]) {
				case NETCMD_CL_JOIN_REQUEST:
					con << "recvMessages(): got a join request from connection " << conn_index << "!" << endl;
					p++;
					conn->count = *p;
					if (!setRemoteEntity(conn)) {
						con << "recvMessages(): setRemoteEntity failed" << endl;
					}
					p++;
					break;
				case NETCMD_CL_READY:
					p++;
					conn->ready = true;
					con << "recvMessages(): got ready cmd from client " << conn_index << endl;
					break;
				case NETCMD_CL_SEND_INPUT:
					p++;
					if (len >= 16) {
						entptr_t dest_ent = {*(short *)p}; p += 2;

						ServerLocalPlayer *lp = NULL;
						for (int i = 0; i < MAX_LOCAL_CLIENTS; i++) {
							if (conn->local[i].ent.i == dest_ent.i) {
								lp = conn->local + i;
								break;
							}
						}
						if (lp) {
							lp->input.cmd = *(int *)p; p += 4;
							lp->input.mouseActive = (*p != 0); p++;
							lp->input.mousePos.x = *(float *)p; p += 4;
							lp->input.mousePos.y = *(float *)p; p += 4;
						} else {
							con.print("recvMessages(): got input method from connection %d for unaccessible entity %d\n", conn_index, dest_ent.i);
							p += 13;
						}
					} else {
						con.print("recvMessages(): got input msg with size too small: %d\n", len);
					}
					break;
				case NETCMD_CL_UPDATE_ENT:
					/*
					this final piece of the client->server entity update will be lengthy
					it'll go like follows:
					1) 
					*/
					p++;
					if (len < 5) {
						con << "are you sure you got an update ent message?  there isn't any room for the ent index and message (len = " << len << ")\n";
					} else {
	
						int index = *(short *)p; p += 2;
						
						if (index < 0 || index >= ENTITY_BLOCK_COUNT) {
							con << "got an update for an out of bounds entity " << index << endl;
						} else {
#ifdef ONE_DELTA_ENTITY_BUFFER
							ClientEntity *cl_ent = (ClientEntity*)(cl_entity + sizeof(ClientEntity) * i);
#else
							ClientEntity *cl_ent = (ClientEntity*)(conn->cl_entity + sizeof(ClientEntity) * i);
#endif
							char original[sizeof(ClientEntity)];
							memcpy(original, cl_ent, sizeof(ClientEntity));
#if 0
							con.print("cl->sv recv ");
							for (char *u = p;;) {
								int ofs = *u; u++;
								int size = *u; u++;
								if (!size) break;
								con.print(" %2x:", ofs);
								for (int j = 0; j < size; j++) {
									con.print(" %2x->%2x", ((char*)cl_ent)[ofs+j], *u); u++;
								}
							}
							con.print("\n");
#endif
							char *deltamsgstart = p;
							char deltamask[sizeof(ClientEntity)];
							memset(deltamask,0, sizeof(deltamask));
							if (!deltaUncompress(p, cl_ent, deltamask, sizeof(ClientEntity))) {
								con << "tried to write an out of bound block to the entity " << index << endl;
								memcpy(cl_ent, original, sizeof(ClientEntity));
							} else {
								EntClass *ec = netTablesSys.getEntClass(cl_ent->entClassID);
								if (ec && ec->netTableID != -1) {
									/*
									got a deltauncompress successfully
									now xor deltabuffer with cl_ent
									if anything comes up, those are the fields we need to update
									*/
									NetTableInstance net(ec->netTableID, cl_ent, index);
									//now have netTableSys sweep through all fields
									//	and only update those that differ from cl_ent to deltabuffer
									
									//make sure the entity's table is on top of the stack before deltaWrite
									lua_getglobal(L, "_ents");
									lua_pushinteger(L, i);
									lua_rawget(L, -2);
									if (lua_istable(L, -1)) {
										net.deltaWrite(L, deltamask);
									}
									lua_pop(L,2);
									//the problem with deltaWrite is it's set up to only write what changes from original to the current table ptr
									//because the client and server share data on a local connection, this data will already be updated
									//	and there will be NO differences
									//	and deltaWrite won't find any fields
									//	and it won't write anything.
									//fix: pick apart the delta message inside deltawrite.
									//this stretches the inderdependency of consistency of the delta structure
									//NOW we write back to lua over the fields that lie in the delta data region
									// even if we don't see a visible difference in the updated data
								}
							}
						}
					}
					break;
				case NETCMD_CL_SEND_MESSAGE:
					p++;
					if (len < 3) {
						con << "got a NETCMD_CL_SEND_MESSAGE with no room for an ent (len=" << len << ")\n";
					} else {
						int index = *(short*)p; p += 2;
						if (index < 0 || index >= ENTITY_BLOCK_COUNT) {
							con << "NETCMD_CL_SEND_MESSAGE got an update for an out of bounds entity " << index << "\n";
						} else {
							vector<string> params;
							bool broken = false;
							for (;;) {
								int len = strlen(p);
								if (!len) {	//a \0 in place of a string is the signal that it's over
									p++;
									break;
								}
								params.push_back(p);
								p += len+1;
								if (p > bm.data + bm.size) {
									con << "NETCMD_CL_SEND_MESSAGE got a broken message" << endl;
									broken = true;
									break;
								}
							}
							if (!broken) {
								lua_State *L = progSys.getState();
								int top = lua_gettop(L);
								lua_getglobal(L, "_ents");		//[_ents]
								lua_pushinteger(L, index);		//[_ents] [index]
								lua_rawget(L, -2);				//[_ents] [e]
								lua_getfield(L, -1, "serverRecvMessage");	//[_ents] [e] [e.serverRecvMessage]
								if (!lua_isfunction(L, -1)) {
									con << "NETCMD_SV_SEND_MESSAGE class did not possess serverRecvMessage member function" << endl;
								} else {
									lua_pushvalue(L, -2);			//[_ents] [e] [e.serverRecvMessage] [e]
									for (vector<string>::iterator i = params.begin(); i != params.end(); ++i) {
										lua_pushstring(L, i->c_str());
									}
									if (lua_pcall(L, params.size()+1, 0, 0)) {
										con << "serverRecvMessage: lua_pcall got error: " << lua_tostring(L, -1) << endl;
									}
								}
								lua_settop(L, top);
							}
						}
					}
					break;
				case NETCMD_CL_DISCONNECT:
					p++;
					con << "interpreting NETCMD_CL_DISCONNECT...\n";
					serverConnSys.destroy(conn);
					break;
				default:
					p++;
					break;
				}

				if (!conn->active) break;
			}
		}
	}
	
	for (Connection *conn = serverConnSys.connections; conn < endof(serverConnSys.connections); conn++) {
		if (!conn->active) continue;
		for (int i = 0; i < conn->count; i++) {
			ServerLocalPlayer *lp = conn->local + i;
			//update deltas while we're here?  or move this somewhere else?
			//recalculate press/release flags
			//	00 = up
			//	10 = press
			//	01 = down
			//	11 = release
			//so the high bit is the xor of the low bit and the last low bit
			int cmdBeforeMessingWithBits = lp->input.cmd;
			lp->input.cmd = 0;
			//lp->input.cmd &= 0x55555555;	//clear all high bits
			//int diffmask = (lp->input.cmd ^ lp->lastInput.cmd) & 0x55555555;
			//lp->input.cmd |= diffmask << 1;
			//TODO - do this as a bit operation, in one command
			for (int i = 0; i < 32; i += 2) {
				int thiscmd = (cmdBeforeMessingWithBits >> i) & 3;
				int lastcmd = (lp->lastInput.cmd >> i) & 3;
				bool thisset = thiscmd == BUTTON_STATE_DOWN || thiscmd == BUTTON_STATE_PRESS;
				bool lastset = lastcmd == BUTTON_STATE_DOWN || lastcmd == BUTTON_STATE_PRESS;
				int newcmd = 0;
				if (thisset) {
					if (lastset) {
						newcmd = BUTTON_STATE_DOWN;
					} else {
						newcmd = BUTTON_STATE_PRESS;
					}
				} else {
					if (!lastset) {
						newcmd = BUTTON_STATE_UP;
					} else {
						newcmd = BUTTON_STATE_RELEASE;
					}
				}
				lp->input.cmd |= newcmd << i;
			}
			if (lp->input != lp->lastInput) {
				game.clientSetInput(lp->ent, &lp->input);
			}
			lp->lastInput = lp->input;
			lp->lastInput.cmd = cmdBeforeMessingWithBits;	//preserve before we mess the bits up to send to lua
		}
	}
}

bool EntitySystemImpl::connCanSeeEntity(Entity *e, Connection *conn, MapTagSet *tags) {
	box2f b;

	//can the engine access virtual members from the mod?
	const AnimInfo *anim = e->getAnim();
	if (!anim) return false;
	anim->getBBox(&b);	//does nothing to b if it fails

	//scale too
	Renderable *r;
	if (r = dynamic_cast<Renderable*>(e)) {
		b.min = vecScaleElem(r->scale, b.min);
		b.max = vecScaleElem(r->scale, b.max);
	}

	b += e->getPos();
	
	//what if the ent has no vis area but still needs to be seen! (like ClientView)

	box2i ib = b;
	if (b.min.x < 0) ib.min.x--;
	if (b.min.y < 0) ib.min.y--;
	if (b.max.x < 0) ib.max.x--;
	if (b.max.y < 0) ib.max.y--;
	
	//get the tags of the corners
	short tag1 = tags->maploc(ib.min.x, ib.min.y);
	short tag2 = tags->maploc(ib.min.x, ib.max.y);
	short tag3 = tags->maploc(ib.max.x, ib.max.y);
	short tag4 = tags->maploc(ib.max.x, ib.min.y);

	//check it to all clients on this connection
	for (int i = 0; i < conn->count; i++) {
		if (conn->local[i].ent.i == -1) continue;
		Entity *cl = entAllocSys.entForPtr(conn->local[i].ent);
		if (!cl || !cl->active) continue;
		short tag = tags->maploc(floor_ftoi(cl->getPos().x), floor_ftoi(cl->getPos().y));

		if (tag == tag1 || tag == tag2 || tag == tag3 || tag == tag4) return true;
	}

	return false;
}

bool EntitySystemImpl::sysInit() { return serverConnSys.init(); }
void EntitySystemImpl::sysKill() { serverConnSys.shutdown(); }

//hmm ... rethink when to disconnect
bool EntitySystemImpl::sessionInit() {
	entAllocSys.clearAll();

	//set the server flag:
	gameflags |= GAME_SERVER;

//	con.print("GAME_SERVER set\n");

	//get our listening sockets up:
	initListen();
	
#if 0	//this is messing up, so for now lets let the client connect manually
	//reconnects all local entities
	//depends on spawnObjSessionInit() for creating an instance of the "player_start" spawn object entity
	con.print("reconnecting...\n");
	if (!reconnect()) return false;
#endif

	return true;
}

void EntitySystemImpl::initListen(int port) {
#ifndef OMIT_NETWORKING
	//inet
	if (!((SocketListenINET*)sv_listener[SOCKET_INET])->init(port)) {
		con.print("cfuncInitListen(): failed to init inet sv_listener on the server\n");
//	} else {
		//con.print("cfuncInitListen(): inet sv_listener working!\n");
	}
#endif //OMIT_NETWORKING

	//loopback
	if (!((SocketListenLoopback*)sv_listener[SOCKET_LOOPBACK])->init( LOOPBACK_TYPE_SERVER )) {
		con.print("cfuncInitListen(): failed to init loopback sv_listener on the server\n");
//	} else {
		//con.print("cfuncInitListen(): loopback sv_listener working!\n");
	}
}

void EntitySystemImpl::killListen() {
	for (int i = 0; i < numberof(sv_listener); i++) {
		sv_listener[i]->kill();
	}
}

void EntitySystemImpl::sessionKill() {
	//unlink all world links *HERE*
	//i dont think i have any yet ...
	
	killListen();
	
	//delete them all anyways - to perform any ~() operations on stuff (strings, etc)
	//ever since introducing this code (and the virtual ~() on Entity) weve been getting crashes upon app term
	for (int i = entAllocSys.getCount()-1; i >= 0; i--) {
		Entity *e = entAllocSys.entForIndex(i);
		assert(e);
		assert(e->active);
		delete e;
	}

	//set the server flag:
	gameflags &= ~GAME_SERVER;

}


//and this is a huge freakin mess
//it reconnects all entities ... supposedly upon a server level change / entity flush / etc
//but right now its also doubly spawning the client entity
bool EntitySystemImpl::reconnect() {
	con << "entitySys.reconnect()..." << endl;

	//send a 'respawn' to the ents of anyone connected

	//for all connections on the server
	for (int i = 0; i < numberof(serverConnSys.connections); i++) {
		//if it is active
		if (serverConnSys.connections[i].active) {
			//for all clients on the connection
			for (int j = 0; j < serverConnSys.connections[i].count; j++) {
				setRemoteEntity(serverConnSys.connections + i);
			}
		}
	}
	return true;
}

void EntitySystemImpl::sendEntDeltaInfo(World *world) {
	connectionCount = 0;

	//order usually doesn't matter
	//but with the one laptop (delta buffer) per child (client)
	//we need to know when the last client is
	//thats usually hard to tell with the whole active flag issue
	//but i'll fix it by just assuming conn 0 is active and cycle in reverse order
	//to fix this: use a vector
	//assert(connections->active);
	for (Connection *conn = serverConnSys.connections + (numberof(serverConnSys.connections)-1); conn >= serverConnSys.connections; conn--) {
		if (!conn->active) continue;
		connectionCount = connectionCount + 1;
#ifdef ONE_DELTA_ENTITY_BUFFER
		//with one delta buffer only update our delta test buffer after comparing for all clients
		bool updateClientEntityDelta = conn == serverConnSys.connections;
#else
		//with many always do this
		bool updateClientEntityDelta = true;
#endif

		sendEntDeltaForConn(world, conn, updateClientEntityDelta);
	}
}

//a mess...
//	todo - merge with the 'net' global in lua
//	 in fact, make 'net' in lua a param to the updateNetInfo function
//	  annd any client functions who use it (updateClientWhatever)
static NetTableInstance *currentNetTableInst = NULL;
static Entity *currentEnt = NULL;

//TODO - generalize the RLE encoding code
// ALSO - make sure - for matching segments in excess of 255 bytes
// - that the RLE byte counters DONT OVERFLOW!
//
/*
here's a new dilemma
lua-side:
clientUpdateRemote() calls render() with our netObj that has been encoded
but clientUpdate() calls render() with self, no net info at all
that means for local connections, for render() we don't need to net encode anything ...

but for client view tracking we seem to ... hmm ...
so I need to find some way to minimize or turn off the net encode stuff for local connections
for speed's sake and what not.
*/
void EntitySystemImpl::sendEntDeltaForConn(World *world, Connection *conn, bool updateClientEntityDelta) {

	if (!conn->sender->isSending()) {
		con.print("tried to send client info through a disconnected socket %d!\n", conn - serverConnSys.connections);
		return;
	}

	//FROM HERE ON OUT IS THE cl_info / cl_entity SENDING CODE
	//if we're a loopback connection, just assume its auto-written, and dont send ent packets
	bool local = (conn->sender->getType() == SOCKET_LOOPBACK);
	//if we're local and not testing the loopback code then override a few key things
	bool localOverride = !serverForceLoopbackSend && local;

	//big message - holding all the smaller messages
	Message<PACKET_LENGTH> bm;	//	unsigned char buffer[PACKET_LENGTH];	int bufferlen = 0;	unsigned char *p = buffer;

	//cycle through every block
	//if its inactive - send a 'block inactive' message
	//if its active - send the updated entity data
	//WE FINALLY MAKE USE OF THE getEntBlockFlags() METHOD!

	//loop across all entities, combining all into a packet (as big as we can get - one per frame)
	for (; conn->next_ent < ENTITY_BLOCK_COUNT; conn->next_ent++) {
		int i = conn->next_ent;

//first - clear the structure

		char new_ent[sizeof(ClientEntity)];		
		memset(&new_ent, 0, sizeof(ClientEntity));		//new_ent.active = false;

//second - test fill conditions
		size_t deltaSize = sizeof(ClientEntity);

		//if its a header to an entity...
		char blockflags = entAllocSys.getBlockFlags(i);
		if (blockflags == 3) {
			
			entptr_t ep = {i};
			Entity *e = entAllocSys.entForPtr(ep);

			//first make sure the entity is a renderable entity
			//only send the ent if its a renderable entity.
			//otherwise, why does the client need it anyways?
			//
			//next make sure it exists in the same tag area as the client:
			//for all corners of this entity,
			//for all the connection's local client entities,
			//if there is no corner that any client can see
			//then consider this entity to be removed (full zero) and send away
			//(if any entity can see any corner then consider the ent not removed)
			//
			//TODO - consider building in the canBeSeen stuff into the script fill routine
			//that way overrides can cancel the canBeSeen requirement
			//...or pass it as a flag?  all in all it looks confusing if we're going to allow 
			//multiple clients per connection (what if dif clients end up seeing dif entities -- what would we send?)
			if (e->active && connCanSeeEntity(e, conn, world->tags)) {


				//here's how things are going to go ...
				//we're gonna run a script-side net fill function
				//it will return a nil if it cant get anything -- upon which we skip the entity --
				//or it will return a table corresponding with a previously (upon script load) registered
				//net table which matches keys (string) to values (specified)
				//this code will convert the table values into the specified net values,
				//then compress and send as it already is designed to

				//single-thread-only!
				//set this as the current ptr
				//to-mt-safe-this: pass 'serverSend' an obj that links to this ent and the client conn and allows safe writes: 
				//serverSend = function(self,net)
				//	net.active = net.client.canSee(self)
				//	if not net.active then return end
				//	net.pos:set(self.pos)
				//	net.weapons = self.weapons
				//etc etc
				//...or even automate that so stuff is just picked from lua and fed into the net table ... tempting ...
				//but that would be less flexible - disallowing any client-side exclusion from specific stuff ( if this then send that )
				
//con << " new ent pos: " << ((ClientEntity*)new_ent)->e.pos << "\n";

				{

					//run other:netSendTo(connIndex)	-- or a connection object with a _ptr index field
					lua_State *L = progSys.getState();
					int top = lua_gettop(L);

					lua_getglobal(L, "_ents");				//[_ents]
					lua_pushinteger(L, i);						//[_ents] [int i]
					lua_rawget(L, -2);							//[_ents] [entity e]
					if (lua_istable(L, -1)) {					//if entity e exists ...

						//todo: eventually optimize the lua vars that are just accessors held in C ...

						EntClass *ec = netTablesSys.getEntClass(e->entClassID);
						if (ec && ec->netTableID != -1) {
							NetTableInstance net(ec->netTableID, new_ent, i);
							NetTable *table = net.getCurrentTable();

							currentNetTableInst = &net;	//keeps getting messier and messier
							currentEnt = e;

							//here: get the current net table
							//cycle through all its entries
							//for each one, get the index from self
							//then set it to 'net' <-> i.e. write it out via PF_netTable()

							if (!table) {
								con << "got ent " << i << " with a bad net table index " << ec->netTableID << "\n";
							} else {
							
								//while we're here, run the entity's updateNetInfo member function
								//this'll convert engine and lua stuff into the lua data
								//which will -- in the next collection of code -- be converted into net table data
								//so to fix this we're gonna short-cut that data

								if (lua_istable(L, -1)) {	//[e]
									lua_getfield(L, -1, "updateNetInfo");	//[e] [updateNetInfo]
									if (lua_isfunction(L, -1)) {
										lua_pushvalue(L, lua_gettop(L)-1);		//[updateNetInfo] [e]
										if (lua_pcall(L, 1, 0, 0)) {		//
											con << "lua_pcall failed to call updateNetInfo:" << endl
												<< lua_tostring(L, -1) << endl
												<< lua_tostring(L, -2) << endl;
										}
									} else {
										lua_pop(L,1);
									}
								}

								deltaSize = table->getSize();
								//cycle through all entries in the map ...
								//since they're stored in an array i might as well make a parallel vector ...
								//and do a lua rw for each
								for (map<string, NetTableEntry *>::iterator it = table->entries.begin(); it != table->entries.end(); it++) {
									const char *key = it->first.c_str();
									NetTableEntry *entry = it->second;
									
									//this flag says that an engine-hardwired callback will be writing to the net table value
									// - rather than this, which picks values out of the lua tables
									if (entry->engineUpdate) continue;
									
									//here we need to see if the key is in lua or in C++
									//actually taht should be done when the net table is created
									//	then it should be given either a C++ or Lua accessor method
									//and here we should call it.
									
									//when the net table is created ...
									//search through the pre-registered (game/program.cpp) entclass tables
									//(we need to map from the entClasses table to the entclass tables in game/program.cpp)
									//when we get a request to make a new nettable entry
									//cycle back through all the parent classes and see if there are any enttable game/program.cpp bindings
									//if there are then give it a special 'readFromServer' code or something ...
									//if (ec->entTable) {
									//} else {
									
									//get the thing from lua
									lua_getfield(L, -1, key);
									if (lua_isnil(L,-1)) {
										con << "entity " << i << " net table id " << ec->netTableID << " got bad key " << key << "\n";
									} else {
										//and store it in the net table
										char *fieldPtr = (char*)new_ent + entry->ofs;
										entry->type.readFromLua(L, lua_gettop(L), fieldPtr);
									}
									lua_pop(L, 1);
								}
							}
							
							currentNetTableInst = NULL;
							currentEnt = NULL;
						}
					}
					lua_settop(L,top);
				}
			}
		}

//fourth - test send/update conditions

		//todo - go from per-client buffers to one unified buffer for the server
		//and for the client a separate one
#ifdef ONE_DELTA_ENTITY_BUFFER
		char *cl_ent = cl_entity + sizeof(ClientEntity) * i;
#else
		char *cl_ent = conn->cl_entity + sizeof(ClientEntity) * i;
#endif

		bool update = false;

		//if force loopback send is off then only update auto if we're local
		//if force loopback send is on then do RLE on everything, local and all
		if (localOverride) {
			update = true;
		} else if (memcmp(&new_ent, cl_ent, deltaSize)) {
			//build the message from the entity
			Message<> m;

			//send a remove message if only now we're inactive
			if (!((ClientEntity *)(&new_ent))->active ) {
				
				//only send the remove command if we were not removed before
				if ( ((ClientEntity *)cl_ent)->active) {
					//if we're only sending this then we're not zero-ing anything clientside, right?
					//what if the client expects things to remain zero later?
					//how about if this makes an assumed zero on everything?
					memset(&new_ent, 0, sizeof(ClientEntity));
					update = true;

					m << (char)NETCMD_SV_REMOVE_ENT;
					m << (short)i;
				}
			} else {
				m << (char)NETCMD_SV_UPDATE_ENT;
				m << (short)i;
				deltaCompress(&new_ent, cl_ent, sizeof(ClientEntity), m);
			}

			//if we're out of room then send it off . . .
			if (!bm.fits(m.size)) { //p + thismsglen > buffer + sizeof(buffer) ) {
				//break to the end of the entity cycler and send through that.
#if 0	//limit the number of updated entities.  throttles traffic, but stutters input if too many ents are sent to the client even on localhost
				break;
#else	//otherwise? just keep on sending...
				conn->sender->send(bm);
				bm.reset();
#endif

				//assert(bufferlen);	//we should at least have something . . . one ent wont fill the whole buffer
				//conn->sender->send((char *)buffer, bufferlen);
				//p = buffer;
				//bufferlen = 0;
			}

			//and increment the buffer pointer
			//bufferlen += thismsglen;
			//memcpy(p, thismsg, thismsglen);
			//p += thismsglen;
			bm << m;
			assert(!bm.overflow);

			//then update the server data
			update = true;
		}

		if (update) {

			if (serverReportChanges) {	//print updated blocks:
				if (memcmp(cl_ent, &new_ent, sizeof(ClientEntity))) {
					con.print("sending update for ent %d:", i);
					for (int q = 0; q < sizeof(ClientEntity); q++) {
						if ( ((unsigned char *)&new_ent)[q] != ((unsigned char *)cl_ent)[q]) {
							con.print(" %2x:%2x", q, ((unsigned char *)&new_ent)[q]);
						}
					}
					con.print("\n");
				}
			}

			//for many delta buffers always update
			//for one delta buffer only update on the last pass
			//determine between the two by the updateClientEntityDelta var
			if (updateClientEntityDelta) {
				memcpy(cl_ent, &new_ent, sizeof(ClientEntity));
			}
		}
	}

	if (conn->next_ent == ENTITY_BLOCK_COUNT) conn->next_ent = 0;

	assert(!bm.overflow);
	//if there's anything more remaining . . .
	if (bm.size) { // bufferlen ) {
		conn->sender->send(bm);	//(char *)buffer, bufferlen);
		bm.reset();	//bufferlen = 0;
	}

	//issues with the current method . . .
	//we send over ClientEntity data for the client to use
	//it is indexed based upon the header index
	//this is dynamic
	//so when a connected client's entity changes order in the headers
	//the client will end up looking at a wrong entity...
	//
	//solutions . . .
	//  send the block index instead of the header index with the ClientEntity
	//	(this would make the client entity remain constant)
	//problems:
	//	we would also have to send updates on ents that become inactive 
	// . . . and that is difficult considering the headers cycle through only active entities...
	//	solutions:
	//		cycle through all entity blocks, testing the flags for each as we go for whether an ent exists here or not . . .

	//what we STILL need to send...
	//all lighting data
	//all traceline data
	//should we subclass Entity for these?  it seems conveniant server-side
	//but for clientside, we're recieving only renderable entities
	//while tracelines *somewhat* cover this, lights definately don't
	//so just send them separate *HERE*
}

template<typename T>
static void safesetvalue(const char *field, const T& value) {
	void *ptr = currentNetTableInst->rawAccess(field);
	if (!ptr) luaL_error(progSys.getState(), "failed to get ptr for field %s", field);
	*(T*)ptr = value;
}

static void entUpdateNetInfo(Entity *e) {
	safesetvalue<bool>("active", e->active);
	safesetvalue<int>("entClassID", e->entClassID);
	safesetvalue<vec2f>("pos", e->getPos());
}

int PF_entUpdateNetInfo(lua_State *L) {
	if (!currentNetTableInst) luaL_error(L, "we have no currentNetTableInst");
	EXPOSE_ENTITY_PARAM_IN_LUA__GRAB_ENT_PARAM(animUpdateNetInfo, Entity, 1, ep, e)
	entUpdateNetInfo(e);
	return 0;
}

static void animUpdateNetInfo(Renderable *r) {
	entUpdateNetInfo(r);
	safesetvalue<int>("absFrame", r->anim.getAbsFrame());
	safesetvalue<NetTypeAngle::angleType>("angle", NetTypeAngle::floatToAngle(r->angle));
	safesetvalue<bool>("left", r->left);
	safesetvalue<vec2f>("scale", r->scale);
	safesetvalue<unsigned char>("blend", r->blend);
	safesetvalue<vec4uc>("color", r->color);
}

int PF_animUpdateNetInfo(lua_State *L) {
	if (!currentNetTableInst) luaL_error(L, "we have no currentNetTableInst");
	EXPOSE_ENTITY_PARAM_IN_LUA__GRAB_ENT_PARAM(animUpdateNetInfo, Renderable, 1, ep, r)
	animUpdateNetInfo(r);
	return 0;
}

static void colUpdateNetInfo(Collidable *c) {
	animUpdateNetInfo(c);
	//do collision updates here... which there are none of
}

int PF_colUpdateNetInfo(lua_State *L) {
	if (!currentNetTableInst) luaL_error(L, "we have no currentNetTableInst");
	EXPOSE_ENTITY_PARAM_IN_LUA__GRAB_ENT_PARAM(animUpdateNetInfo, Collidable, 1, ep, c)
	colUpdateNetInfo(c);
	return 0;
}

void EntitySystemImpl::update() {
	PROFILE();

	//no server running
	if (!(gameflags & GAME_SERVER)) return;

	//get all messages
	//	new client connections
	//	old client input updates
	recvMessages();
	
	/*
	we don't hold the net table between entSystemUpdate and sendEntDeltaInfo ...
	how about we move the updateNetInfo from think (called in entSystemUpdate)
	...to sendEntDeltaInfo...
	...to be called after a net buffer is allocated and before the send...
	
	then within updateNetInfo have callbacks to C code for the base classes
	that should shave 15fps or so off
	*/

	//call the game's entity system update code
	//this calls think
		//and think contains the code that sends data from the ent binary to ent script...
	game.entSystemUpdate();

	World *world = worldSys.getWorld();

	//send out all entity data here
	sendEntDeltaInfo(world);
}

void EntitySystemImpl::sendMessage(entptr_t ep, const vector<string> &params) {
	/*
	options for overflow
	one: max size on messages, and just 'return false' if it doesn't fit
	two: keep track of a fatty list clientside and patch them together afterwards
		... that means three way handshakes and timestamps and stuff ...
	*/
	for (Connection *conn = serverConnSys.connections; conn < endof(serverConnSys.connections); conn++) {
		if (!conn->active) continue;
		for (int i = 0; i < numberof(conn->local); i++) {
			if (conn->local[i].ent.i == ep.i) {
				Message<> m;

				m << (char)NETCMD_SV_SEND_MESSAGE;
				m << (short)ep.i;
				
				//you know, this is UDP...
				//that means order isn't ensured ...
				//I should probably just make a TCP connection and let the protocol figure all this out...
				for (vector<string>::const_iterator j = params.begin(); j != params.end(); ++j) {
					int neededSize = j->length() + 2;	//two 'cuz one for null term, one for msg null term
#if 1	//handle splits
					if (neededSize >= m.capacity() - 2) {
						//set some error here
						m.overflow = true;
						con << "message overflow" << endl;
						break;
					}
					if (!m.fits(neededSize)) {
						//then we need to give it a "more to come" flag...
						//don't forget this is datagram...
						//also - error if one string in the vector won't fit in an entire buffer
						m << (char)1;
						m << '\0';
						conn->sender->send(m);
						m.reset();
						m << (char)NETCMD_SV_SEND_MESSAGE;
						m << (short)ep.i;
					}
#endif
					m << j->c_str();
				}
				m << '\0';
				if (m.overflow) {
					con << "EntitySys::sendMessage got an overflow" << endl;
					con << "message data was" << endl;
					for (vector<string>::const_iterator j = params.begin(); j != params.end(); ++j) {
						con << *j << endl;
					}
					con << endl;
				} else {
					conn->sender->send(m);
				}
				return;
			}
		}
	}
	con << "EntitySys::sendMessage unable to find connection for entity " << ep.i << endl;
}

EntitySystem::EntitySystem() : impl(new EntitySystemImpl()) {}
EntitySystem::~EntitySystem() { delete impl; }
bool EntitySystem::sysInit() { return impl->sysInit(); }
void EntitySystem::sysKill() { impl->sysKill(); }
bool EntitySystem::sessionInit() { return impl->sessionInit(); }
void EntitySystem::sessionKill() { impl->sessionKill(); }
bool EntitySystem::reconnect() { return impl->reconnect(); }
void EntitySystem::update() { impl->update(); }
void EntitySystem::sendMessage(entptr_t ep, const std::vector<std::string> &params) { impl->sendMessage(ep, params); }
EntitySystem entitySys;


 void cfuncSetLevel(int argc, char **argv) {
	if (argc < 2) {
		con.print("usage: setlevel <filename>\n");
		return;
	}

	char *filename = argv[1];

	//load the bitmap
	if (!worldSys.loadWorld(filename)) {
		con.print("failed to load world %s\n", filename);
		return;
	}

	mainSys.sessionReset();	//inits the level, inits the entities, inits the connections
}

void cfuncInitListen(int argc, char **argv) {
	int port = 0;
	if (argc > 1) port = atoi(argv[1]);
	entitySys->initListen(port);
}

void cfuncKillListen(int argc, char **argv) {
	entitySys->killListen();
}

static CVarw *localCvars[] = {
	new CVarw(0,	"setlevel",		cfuncSetLevel),
	new CVarw(0,	"initListen",	cfuncInitListen),
	new CVarw(0,	"killListen",	cfuncKillListen),
};

//game variables

