/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/#include <math.h>
#include <assert.h>

#include "system/exception.h"

#include "inspiration/game/client.h"
#include "inspiration/entalloc.h"
#include "inspiration/program.h"
#include "inspiration/nettables.h"
#include "inspiration/delta.h"

#include <sstream>
using namespace std;

//expects a table on the stack
//grabs its instance function and puts it before it
//in: [obj]
//out: [obj:func] [obj]
void lua_getInstFunc(lua_State *L, const char *func) {
	if (!lua_istable(L,-1)) throw Exception() << "couldn't find object";
	lua_getfield(L, -1, func);				//[obj] [func]
	if (!lua_isfunction(L, -1)) throw Exception() << "couldn't find object function " << func;
	lua_insert(L, lua_gettop(L)-1);		//[func] [obj]
}

//returns GameAPI[func] GameAPI
void lua_getGameAPIfunction(lua_State *L, const char *func) {
	try {
		lua_getglobal(L,"GameAPI");		//[GameAPI]
		lua_getInstFunc(L, func);
	} catch (Exception &t) { throw Exception() << "GameAPI object: " << t; }
}

void lua_pushentity(lua_State *L, entptr_t ep) {
	lua_getglobal(L, "_ents");			//_ents
	if (lua_istable(L, -1)) {
		lua_pushinteger(L, ep.i);		//_ents entIndex
		lua_gettable(L, -2);			//_ents ent
		lua_remove(L, lua_gettop(L)-1);	//ent
	}
}

void GameAPI::clientSetInput(entptr_t dest_ent, const ClientInput *in) {
	lua_State *L = progSys.getState();
	int top = lua_gettop(L);
	try {
		lua_getGameAPIfunction(L, "clientSetInput");
		lua_pushentity(L, dest_ent);
		lua_pushinteger(L, in->cmd);
		lua_pushboolean(L, in->mouseActive);
		lua_pushnumber(L, in->mousePos.x);
		lua_pushnumber(L, in->mousePos.y);
		if (lua_pcall(L, 6, 0, 0)) throw Exception() << "failed to run. got errors:\n" << lua_tostring(L, -1);
	} catch(Exception &t) {
		//hmm, this is def. a session-worthy error.  i.e. drop out of the session now
		con << "failed to run GameAPI:clientSetInput: " << t << "\n";
	}
	lua_settop(L, top);
}

/*
okay here's the deal
this is called from the clientside
we need a way to test local/remote connect
if we're local then do this
if we're remote then run a separate method

 if we're remote then ... 
  well, store 2N nettable records for each entity
		for client-solo's N=1; for servers N=#connections
 then during our client update
	pick up network packets and interpret them
	un-delta-compress them into their assoc client-side entity things...

 the client-side entity arrays ...
 will hold:
	0) entity #
	1) class # (enumerate on both client and server side - assume the both have the same script)
	2) active flag
	3) net table content

then instead of cycling across the _ents array we will just do so over the cl_entity array


our new clientside structure will be fully-nettable-based
basic entries:
	active
	entClassID
	pos
	etc...
allocate [numConnections]x[maxClientEnts]
and for each allocation, allocate a pointer or something ...
-- we could cap the nettable size and then just do one big allocation ...
hmm...
either that or lots of little dynamic allocations as we get varying nettables
doesnt malloc() do so in increments of 512 or something anyways? thats more than I'll need I believe
*/
void GameAPI::clientUpdateEntities(ClientEntity *e, int count, SocketSend *cl_sender) {
	Message<PACKET_LENGTH> bm;
	char original[sizeof(ClientEntity)];
	lua_State *L = progSys.getState();
	int basetop = lua_gettop(L);
	for (int i = 0; i < count; i++, e++) {
		if (!e->active) continue;
		EntClass *ec = netTablesSys.getEntClass(e->entClassID);
		if (ec && ec->netTableID != -1) {
			int top = lua_gettop(L);
			NetTableInstance net(ec->netTableID, e, i);
			
			//take the entClassID from the structure
			//and look up that class in the '_entClasses' table
			//this asserts the structure will be the same between the client and server
			//which will be true so long as they have the same scripts and lua impl
			lua_getglobal(L, "_entClasses");		//[_entClasses]
			lua_pushinteger(L, e->entClassID);		//[_entClasses] [e->entClassID]
			lua_gettable(L, -2);					//[_entClasses] [class]
			
			memcpy(original, e, sizeof(ClientEntity));

			if (!lua_istable(L, -1)) {
				con << "couldn't find lua class table for entClassID " << e->entClassID << "!\n";
				break;
			} else {
				lua_getfield(L, -1, "clientUpdate");	//[_entClasses] [class] [class.clientUpdate]
				//dont complain if the function is not there
				//just assume it wasnt provided
				if (lua_isfunction(L, -1)) {
					lua_pushvalue(L, lua_gettop(L)-1);				//[_entClasses] [class] [class.clientUpdate] [class]
					if (lua_pcall(L, 1, 0, 0)) {
						con << "clientUpdate: lua_pcall got error: " << lua_tostring(L, -1) << "\n";
						lua_printbacktrace(L);
					}
				} else {
					//call a default otherwise?
					//but to have nil be the specifier for a default call would invalidate any Renderable inheritence 
					//	any class.super.clientUpdate() calls would break ...
				}
			}

			//does anyone use this?
			//and if times don't match up between packets sent back and forth, won't this desync with the server?
			//..and its original intention was replaced with client/server message passing
			//it's only used now for client acknoweldgements -- which should be built into the client connection anyways
			if (net.isDirtyFlagSet()) {
				//then delta compress
				Message<> m;
				m << (char)NETCMD_CL_UPDATE_ENT;
				m << (short)i;
				deltaCompress(e, original, net.getCurrentTable()->getSize(), m);

				if (!bm.fits(m.size)) {
					cl_sender->send(bm);
					bm.reset();
				}
				
				bm << m;
				assert(!bm.overflow);
			}

			lua_settop(L,top);
		}
	}
	lua_settop(L,basetop);
	
	assert(!bm.overflow);
	if (bm.size) {
		cl_sender->send(bm);
		bm.reset();
	}
}
