/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <map>
#include <string>

#include "inspiration/game/main.h"	//gameapi
#include "inspiration/game/programmacros.h"	//all those lua macros
#include "inspiration/game/nettables.h"	//hack in progress

using namespace std;

//this forces the meta index and newindex calls to make a C call
//which was only optional before depending on the key
//the forced C call is more of a performance hit then it originally seemed worth
//i don't know if things are made up for in the network code
//until then i'm just zeroing this out
#ifdef HANDLE_ENTITY_META_IN_ENGINE

//NetTable no longer suits the title
//more like reflection table? (stealing Java's word)
NetTable keyTable_Entity;
NetTable keyTable_Renderable;
NetTable keyTable_Collidable;

//now for each of these make an engine callback function for the meta indexes
//entptr, key[, value]
//if value is provided then we are taking in params
//if not we are returning our value
#define CREATE_META_TABLE(entclass)	\
int PF_engine##entclass##MetaTable(lua_State *L) {	\
	int argc = lua_gettop(L);	\
	if (argc != 2 && argc != 3) {	\
		con << "PF_engine" #entclass "MetaTable expected an ent ptr and a key\n";	\
		return 0;	\
	}	\
	/*get our key*/	\
	const char *key = lua_tostring(L, 2);	\
	if (!key) {	\
		con << "PF_engine" #entclass "MetaTable expected a key as the second param\n";	\
		return 0;	\
	}	\
	\
	/*check for key validity first before we see if our ent is there*/	\
	/*sometimes reads are done on entptr -1 for key 'active'*/	\
	/*so we should be grateful we don't have to handle them.*/	\
	/*note to self: active gets its own special type*/	\
	map<string,NetTableEntry*>::iterator it = keyTable_Entity.entries.find(key);	\
	if (it == keyTable_Entity.entries.end()) return 0;	/*i said false, but nil works too*/	\
	NetTableEntry *entry = it->second;	\
		\
	/*get our entity*/	\
	EXPOSE_ENTITY_PARAM_IN_LUA__GRAB_ENT_PARAM(engine##entclass##MetaTable, Entity, 1, ep, e)	\
	\
	char *fieldPtr = (char *)e + entry->ofs;	\
\
	lua_pushboolean(L, true);		/*found it so put a true on the stack*/	\
	int result = 1;	\
		\
	if (argc == 2) {	/*read: return true/false if its found, then return the valuer*/	\
		result += entry->type.writeToLua(L, fieldPtr);	\
	} else {	/*argc == 3, write: so just return true/false*/	\
		entry->type.readFromLua(L, 3, fieldPtr);	\
	}	\
	\
	return result;	\
}
CREATE_META_TABLE(Entity)
CREATE_META_TABLE(Renderable)
CREATE_META_TABLE(Collidable)
#else
int PF_engineEntityMetaTable(lua_State *L) { return 0; }
int PF_engineRenderableMetaTable(lua_State *L) { return 0; }
int PF_engineCollidableMetaTable(lua_State *L) { return 0; }
#endif

bool GameAPI::sysInit() {
#ifdef HANDLE_ENTITY_META_IN_ENGINE
	//for each engine-bound class, make our bindings and our metatables
	//now don't we need unique callback functions for each binding?
	//not necessarily ... but then we'd need to pass some unique lua-class identifier
	//and i don't know how to make that mapping from lua to C++ ...

#define ADDENTRY(cl,type,name)	keyTable_##cl.addEntry(#name, type, OFFSETOF(cl, name) )
	
	//here's a catch: the server-side tables and net(i.e.client-side) tables must match up!
	//having 1 transfer func for 2 permutations is bad enuf.  i don't want 3 funcs for 3 perms.
	
	//so just map for now
//	ADDENTRY(Entity, netTypeActive, active);		//read only means custom nettype.  well, read-only from lua at least.  not for writing through network.
//	ADDENTRY(Entity, netTypeSpawnObj, spawnobj);	//still todo
	ADDENTRY(Entity, netTypeShort, entClassID);
//	ADDENTRY(Entity, netTypeVec2f, pos);			//vec2f will be special for writing.  same with all structs
	ADDENTRY(Entity, netTypeFloat, nextthink);
	ADDENTRY(Entity, netTypeInt, layer);
	
//	ADDENTRY(Renderable, netTypeAnimGroup, group);		//very custom.  its anim.group ftw
//	ADDENTRY(Renderable, netTypeAnimSeq, seq);			//likewise, anim.seq
//	ADDENTRY(Renderable, netTypeAnimIdleSeq, idleseq);	//and anim.idleseq
	keyTable_Renderable.addEntry("frame", netTypeInt, OFFSETOF(Renderable, anim.frame));
	keyTable_Renderable.addEntry("frameNextTime", netTypeFloat, OFFSETOF(Renderable, anim.frameNextTime));
	keyTable_Renderable.addEntry("frameLastTime", netTypeFloat, OFFSETOF(Renderable, anim.frameLastTime));
//	ADDENTRY(Renderable, netTypeAnimAbsFrame
	ADDENTRY(Renderable, netTypeBool, left);
//	ADDENTRY(Renderable, netTypeFloat, angle);	//debate this.  its stored server as float, client as ushort
//	ADDENTRY(Renderable, netTypeVec2f, scale);
	ADDENTRY(Renderable, netTypeUChar, blend);
//	ADDENTRY(Renderable, netTypeColor, color);
	
	ADDENTRY(Collidable, netTypeBool, noPhysics);
	ADDENTRY(Collidable, netTypeFloat, rotation);
//	ADDENTRY(Collidable, netTypeVec2f, vel);
	ADDENTRY(Collidable, netTypeFloat, massInv);
	ADDENTRY(Collidable, netTypeBool, useGravity);
//	ADDENTRY(Collidable, netTypeBox2f, bbox);
	ADDENTRY(Collidable, netTypeFloat, restitution);
	ADDENTRY(Collidable, netTypeFloat, friction);
	ADDENTRY(Collidable, netTypeBool, collIgnoreEnts);
	ADDENTRY(Collidable, netTypeBool, collIgnoreWorld);
	ADDENTRY(Collidable, netTypeBool, collTouchOnly);
	ADDENTRY(Collidable, netTypeBool, collDiffClass);
//	ADDENTRY(Collidable, netTypeEntity<Collidable>, collIgnore);
//	ADDENTRY(Collidable, netTypeVec2f, contactNormal);
	ADDENTRY(Collidable, netTypeFloat, contactTime);
//	ADDENTRY(Collidable, netTypeEntity<Collidable>, contactEnt);
//	ADDENTRY(Collidable, netTypeVec2i, contactTile);
#endif
	return true;
}
