/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <math.h>
#include <assert.h>
#include <iostream>

#include "system/gl.h"
#include "system/exception.h"
#include "system/singletons.h"
#include "system/memory.h"

#include "base/base.h"
#include "base/main.h"

#include "menu/sys.h"

#include "entalloc.h"
#include "world.h"
#include "texloader.h"
#include "client.h"
#include "main.h"
#include "lighting.h"
#include "draw.h"
#include "program.h"
#include "spawnobj.h"
#include "network.h"
#include "animation.h"
#include "shader.h"
#include "sound.h"
#include "nettables.h"
#include "filesys.h"
#include "menu.h"
#include "game/main.h"
#include "editor/editor.h"

using namespace std;

//hack it into a module
class MenuSysModule : public IModule, public pwn::ResizeListener {
public:
	virtual bool sysInit() {
		pwn::menuSys.init();
		pwn::menuSys.getRoot()->addResizeListener(this);
		return true;
	}
	virtual void sysKill() {
		pwn::menuSys.shutdown();
	}
	virtual void callback_onResize(pwn::Object *o) {
		pwn::menuSys.getRoot()->setScale(vec2f(
			disp.getSize().x / 60.f,
			disp.getSize().y / 40.f));
	}
};
static MenuSysModule menuSysModule;	

static bool system_init_d = false;
static bool session_init_d = false;

//globals to cut back on param passing & make it easier on the script system
float system_thistime;
float system_lasttime;
float system_deltatime;

float game_thistime;
float game_lasttime;
float game_deltatime;

float game_min_dt = 0.001f;
float game_max_dt = 0.1f;
float game_time_scalar = 1.f;

gamestate_t gameState = GS_MENU;

//all it takes to expose a function through the console system
void CFunc_Reset(int, char**)	{mainSys.gameReset();}
static CVarw cv_reset(0, "reset", CFunc_Reset);

//composite cvar commands would fix this ...
static void cfuncResetWithConnection(int, char**)	{
	mainSys.gameReset(); 
	clientSys.connect(1, 0);
}
static CVarw cv_resetWithConnection(0, "resetWithConnect", cfuncResetWithConnection);

static CVarw *local_cvars[] = {
	new CVarw(&game_min_dt,			"game_min_dt"),
	new CVarw(&game_max_dt,			"game_max_dt"),
	new CVarw(&game_time_scalar,	"game_time_scalar"),
};

void CFunc_Pause(int,char**) { gameState = GS_PAUSED; }
static CVarw cv_pause(0, "pause", CFunc_Pause);

void CFunc_Play(int,char**) { gameState = GS_RUNNING; }
static CVarw cv_play(0, "play", CFunc_Play);

void cfuncMenu(int,char**) { gameState = GS_MENU; }
static CVarw cv_menu(0, "menu", cfuncMenu);

//0 = menu splash screen
//1 = dedicated server
//2 = connected to remote server
//3 = local game

int gameflags = 0;

class MainModule : public IModule {
	typedef IModule super;
public:

	//trying to organize it all ...
	class ModulePair {
	public:
		ModulePair(const char *_name, IModule *_module) : name(_name), module(_module) {}
		const char *name;
		IModule *module;
	};
//this is why I love structs ...
#define MODULEPAIR(x)	ModulePair(#x,&x),

	class ModuleArray {
	public:
		ModuleArray() : v(0), size(0) {}
		ModulePair *v;
		int size;
	};
	
	ModuleArray sysModules;
	ModuleArray gameModules;
	ModuleArray sessionModules;
	
#define BEGIN_MODULEARRAY(x)	\
	{	\
		/*built into the begin module stuff to make sure the coder matches the names up right*/	\
		static const char *_hackcheck = #x; \
		static ModulePair _##x##Modules[] = {

#define END_MODULEARRAY(x)	\
		};	\
		x##Modules.v = _##x##Modules;	\
		x##Modules.size = numberof(_##x##Modules);	\
		assert(!strcmp(#x, _hackcheck));	\
	}
	
#define MODULEARRAY_INIT(x)	\
	for (int i = 0; i < x##Modules.size; i++) {	\
		con << x##Modules.v[i].name << "."#x"Init...\n";	\
		if (!x##Modules.v[i].module->x##Init()) return false;	\
	}
	
#define MODULEARRAY_KILL(x)	\
	for (int i = x##Modules.size-1; i >= 0; i--) {	\
		x##Modules.v[i].module->x##Kill();	\
	}

	virtual bool sysInit() {
		//all the modules that sys uses
		//will these be properly defined even if the linker puts this function's root code before the insanciation of those objects?
		//probably not, so do them at sysInit after all the global namespace has been run
		//statics dont get alloc'd/init'd until their line is hit, right?
		//i wish i could have this locally alloc'd but still class namespace so i could use numberof() on it
		BEGIN_MODULEARRAY(sys)
			MODULEPAIR(shaderSys)		//for extension loading
			MODULEPAIR(fileSys)
			MODULEPAIR(soundSys)
			MODULEPAIR(networkSys)
			MODULEPAIR(texSys)
			MODULEPAIR(menuSysModule)
			MODULEPAIR(lightSys)
			MODULEPAIR(progSys)
			MODULEPAIR(entAllocSys)		//needs to be before entitySys (for conn disconnect).  this was after game.
			MODULEPAIR(entitySys)		
			MODULEPAIR(spawnObjSys)
			MODULEPAIR(renderSys)
			MODULEPAIR(editor)
			MODULEPAIR(worldSys)
			MODULEPAIR(clientSys)
			MODULEPAIR(game)
			MODULEPAIR(netTablesSys)		//whenever.  no dependencies either way.
		END_MODULEARRAY(sys)
		
		BEGIN_MODULEARRAY(game)
			MODULEPAIR(mainMenuSys)
			MODULEPAIR(editor)			//fix up the editor now that the types are gone
			MODULEPAIR(game)
			MODULEPAIR(progSys)
			MODULEPAIR(worldSys)
			MODULEPAIR(clientSys)		//just resets vars atm
		END_MODULEARRAY(game)
		
		BEGIN_MODULEARRAY(session)
			MODULEPAIR(shaderSys)		//clears the shader buffers for accum'ing handles in
			MODULEPAIR(clientSys)		//inits the local cl_entity and cl_info buffers
			MODULEPAIR(entitySys)		//clears all entity tables
			MODULEPAIR(animSys)
			MODULEPAIR(spawnObjSys)
			
			//run this before progSys so we create Renderable's netTable before any classes are created
			//(rather than after -- because if we did so after, then all the params we filled the net table in with wouldn't be inherited)
			MODULEPAIR(netTablesSys)

			//this is dependant upon Editor_FlushProgram
			//
			//then calls progStop and progClear to ensure nothing else is running at the moment
			//well, then I have it call progSys.compile which calls the spawn object's add globals routine
			//
			//so here's where we run the root lua script
			//and that's gonna create the base classes, like those bound to the engine
			//if we then later call netTablesSys.sessionInit then netTablesSys can pick out the engine-bound classes
			//	... and give them net tables
			MODULEPAIR(progSys)

			//now that we've run the script
			//lets refresh all our map textures
			MODULEPAIR(worldSys)
			
			//refresh the editor
			//this gives it the new spawntypes
			//also by now we can assert textures are good too
			MODULEPAIR(editor)
			
			//run the main() script function ...
			//todo - call this 'GameAPI:onSessionInit' or something else more appropriate ...
			MODULEPAIR(game)
		END_MODULEARRAY(session)
		
		cvarSys->exec("log");

		//adjust foundation
			//adjust console
		con.font.load2D("base/res/font.png");

		base.useCameraKeys = false;

			//engine module:
		MODULEARRAY_INIT(sys)
		
			//once the game is up, load a level with it!
		con.print("gameInit...\n");
		if (!gameInit()) return false;

		system_init_d = true;

			//optional?

		con.print("system initialized!\n");
		
		con << "mainSys.sysInit good!\n";

		return true;
	}
	
	virtual void sysKill() {
		gameKill();	//contains the sessionShutdown code

		MODULEARRAY_KILL(sys)

		system_init_d = false;
		
#ifdef USE_PROFILER
		profileSys->report();
#endif
	}

	//contains any post-mod system init / shutdown code
	//as well as the session init / shutdown code
	virtual bool gameInit() {

		MODULEARRAY_INIT(game)
		
		//this really shouldn't even be called here
		//it should have a session object or something ... i think ...
		//but either way, it shouldn't be called when the main menu first goes up
		//... not til they click 'new game' at least

		con << "sessionInit...\n";
		if (!sessionInit()) {
			con << "mainSys.gameInit(): sessionInit failed!\n";
			//don't fail if session init fails
			//because session is independent of system
		} else {
			con << "mainSys.gameInit good!\n";
		}

		return true;
	}

	virtual void gameKill() {
		//first shut down the session
		mainSys.sessionKill();
		
		MODULEARRAY_KILL(game)
	}

	virtual bool gameReset() {
		bool res = super::gameReset();
		if (!res) con << "g_mainSys.gameReset() failed!\n";
		return res;
	}

	virtual bool sessionInit() {
		//engine session init
		game_thistime = 0;
		
		MODULEARRAY_INIT(session)

		con.print("session initialized!\n");

		session_init_d = true;

		return true;
	}

	virtual void sessionKill() {

		if (!session_init_d) return;
		
		MODULEARRAY_KILL(session)

		session_init_d = false;
	}

	virtual bool sessionReset() {
		bool res = super::sessionReset();
		if (!res) con << "mainSys.sessionReset failed!\n";
		
#if 1
		//now do reconnections...
		//do them more official than this please
		//which means before calling reset, save the conn state
		//then after, restore it
		clientSys.connect(1, 0);
#endif

		return res;
	}

	void update() {
		PROFILE();
		if (!system_init_d) return;
		
		{
			PROFILE();
			mainMenuSys.update();
		}

#if 0	//variable frametime physics integration

			//intro timer
		system_lasttime = system_thistime;
		system_thistime = (float)timer.getTime();
		system_deltatime = system_thistime - system_lasttime;
		if (gameState == GS_RUNNING) {
			PROFILE();
			game_deltatime = system_deltatime;
			//clamp delta time
			if (game_deltatime < game_min_dt) game_deltatime = game_min_dt;
			if (game_deltatime > game_max_dt) game_deltatime = game_max_dt;
			game_deltatime *= game_time_scalar;
			game_lasttime = game_thistime;
			game_thistime = game_lasttime + game_deltatime;
			//server
			entitySys.update();
			progSys.update();
		}
		
		{	//allow the client to run even whe the game isnt (for the editor)
			//client
			PROFILE();
			clientSys.update();	//gathers input, reads server state, displays system
		}

#else	//variable time client, fixed time server

			//intro timer
		system_lasttime = system_thistime;
		system_thistime = (float)timer.getTime();
		system_deltatime = system_thistime - system_lasttime;
		if (gameState == GS_RUNNING) {
			PROFILE();
			static int last_system_frame = -1;
			const float system_framerate = 60.f;	//1/N-th of a second
			int this_system_frame = (int)(system_thistime * system_framerate);
			game_deltatime = 1.f / system_framerate;

			//as byproduct of this method.
			// keyboard press/release states are set only one frame
			//since occasionally frames dont calculate game state updates,
			//keyboard press/release state reads get skipped.
			//SOLUTION: don't use the keyboard press/release states,
			//but keep track of keyboard down/up states ourselves
			//and calculate relative press/releases ourselves as well
			for (int i = last_system_frame; i < this_system_frame; i++) {
				PROFILE();
				game_thistime = i * game_deltatime;
				game_lasttime = game_thistime - game_deltatime;
				entitySys.update(); //server update
				progSys.update();
			}
			last_system_frame = this_system_frame;
		}

		//allow the client to run even whe the game isnt (for the editor)
		{
			PROFILE();
			//ANOTHER STICKY SITUATION
			//input needs to be gathered/interpretted per server update frame . . .
			//except editor input - that needs to be done per client frame . . .
			//i see a need to separate the two into separate routines =)
			//one to be called post-entitySys.update() in the physics loop above
			//and the other (the client displays and the editor) to be called here.

			//BASICALLY to break down the separation
			//the input interpretter - whether a key has been pressed or not
			// - needs to operate every server frame
			//the display interpreter and the editor interpreter
			// - needs to occur every client frame

			//so why dont we just move the input interpretter into the server code?
			//give the user two variables:
			//	input_down	for the up/down states of the keys
			//	input_press	for the press/release states of the keys
			//the client will give the server 'input_down' every frame
			//and the server will discern the press/release keys from this . . 

			//how did quake do it?
			//they collected press/release keys throughout all client updates
			//and passed them at once to server
			//and from there, the server determined what was down or not
			//hmm...

			//so we would have to send an arbitrary sized list of messages
			//rather than a fixed size array of key flags . . .
			//h m m m. . . fixed-size the list; the user won't expect the game
			//to store much more than a dozen key-cmds every 10ms, will they?

			//client
			clientSys.update();	//gathers input, reads server state, displays system
		}
#endif
			//exit timer
	}
};
static MainModule g_mainSys;
IModule &mainSys = g_mainSys;

//this is the kind of module thats registerred with the foundation stuff
//different from the modules that go on inside of the engine
class Base_MainModule : public BaseModule {
	typedef BaseModule super;
public:
	virtual void init() {
/*	
set up size

Option 1:
add in yet another lua script somewhere for window setup
 have it specify size, resizability, (position, color/stencil/depth info, bufferedness?)
if I wanted this to be game-centric thennn....
then I wouldn't want windows to be created until after the game was init'd
and that'd cause us a window close / open every game restart
not pretty

Option 2:
issue resize functions
this means exposing resize code to the game
and the ability to set resizability to the game
both of which sound good.  now will mac cooperate?
*/
		super::init();

		con.setIdleVisible(false);
		
		if (!mainSys.sysInit()) throw Exception() << "failed to init mainSys";

		//dont allow any render calls
		base.clearAllViewports();
	}
	
	virtual void shutdown() {
		g_mainSys.sysKill();
		super::shutdown();
	}
	
	//and i mean no render calls
	virtual void render(const Viewport *, const View *) { assert(false); }
	
	virtual void update() {
		glViewport(0,0,VEC2ELEM(disp.getSize()));
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
		g_mainSys.update();
	}
};
static Base_MainModule base_mainSys;
BaseModule *RegisterApplication() { return &base_mainSys; }