/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <iostream>
#include <stdio.h>
#include <assert.h>

#include "base/base.h"
#include "defs/defs.h"	//numberof
#include "network.h"

using namespace std;
#define MY_DEFAULT_PORT	22501 

//client architecture:
	//we have one sending socket - sending data to port X on the server
	//we have one listening socket - listening for data from port X on our machine
	//
	// upon connect request:
	// kill all sending sockets
	// init listening socket
	// init sending socket 0 for sending to our dest host

//server architecture
	//we have one listening socket - listening for all data from anywhere	
	//we have N sending sockets - one for every connected client
	//
	// every time a new join message is recieved from an unknown client,
	// we find an unused socket and use it to reply & connect to the joiner
	//
	// every time no packets arrive from a client after N seconds or so, we drop the connection
	//
	// add in some cvars for kicking connections too
	//
	// upon net server init:
	// kill all sending sockets
	// init listening socket
	// if we get a packet, track the client, verify the protocol, and init a new sending socket for that client

#ifndef OMIT_NETWORKING


#ifdef WIN32
#define errno	WSAGetLastError()
#define ioctl	ioctlsocket
typedef size_t	socklen_t;
#else
#include <sys/ioctl.h>
//how do you close sockets?
#define closesocket close
#define SOCKET_ERROR	-1
#define INVALID_SOCKET	-1
#endif
	
// ******************************** INET SERVER-SIDE

SocketListenINET::SocketListenINET() {
	listening = 0;
	listen_sockfd = INVALID_SOCKET;	//invalid socket
}

SocketListenINET::~SocketListenINET() {
	if (isListening()) kill();
}

bool SocketListenINET::init(int dest_port) {
	if (listening) {
		con << "SocketListenINET::init(): already listening\n";
		return false;
	}

	if (!dest_port) dest_port = MY_DEFAULT_PORT;

	con << "listening on port " << dest_port << "...\n";

	if ((listen_sockfd = socket(AF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET) {
		con.print("socket: got error %x\n", errno);
		return false;
	}

	u_long _true = 1;
//	if (ioctlsocket (listen_sockfd, FIONBIO, &_true) == -1)
	if (ioctl(listen_sockfd, FIONBIO, &_true) == -1)
	{
		con.print("ioctlsocket: got error %x\n", errno);
		closesocket(listen_sockfd);
		return false;
	}

	listen_my_addr.sin_family = AF_INET;         // host byte order
	listen_my_addr.sin_port = htons(dest_port);  // short, network byte order
	listen_my_addr.sin_addr.s_addr = INADDR_ANY; // automatically fill with my IP
	memset(&(listen_my_addr.sin_zero), '\0', 8); // zero the rest of the struct

	if (bind(listen_sockfd, (struct sockaddr *)&listen_my_addr, sizeof(struct sockaddr)) == SOCKET_ERROR) {
		con.print("bind: got error %x\n", errno);
		closesocket(listen_sockfd);
		return false;
	}

	listening = 1;
	con.print("successfully listening!\n");

	return true;
}

bool SocketListenINET::kill() {
	if (listening) {
		closesocket(listen_sockfd);
		listening = 0;
		con.print("end listen session\n");
	} else {
		con.print("not listening\n");
		return false;
	}

	return true;
}

int SocketListenINET::isListening() {
	return listening;
}

//returns 0 = nothing, 1 = something (length), -1 = error
int SocketListenINET::recv(char *destbuf, int *destbufsize, sockaddr_in *destaddr) {
	if (!listening) return -1;

#ifdef WIN32
	int addr_len = sizeof(struct sockaddr);
#else
	socklen_t addr_len = sizeof(struct sockaddr);
#endif
	struct sockaddr_in their_addr; // connector's address information

	int numbytes = recvfrom(
		listen_sockfd,
		destbuf,
		*destbufsize,
		0,
		(struct sockaddr *)&their_addr,
		&addr_len);

	if (destaddr) {
		memcpy(destaddr, &their_addr, sizeof(sockaddr_in));
	}

	if (numbytes == SOCKET_ERROR) {
		int last_error = errno;

#ifdef WIN32
#define EWOULDBLOCK		WSAEWOULDBLOCK
#define ECONNREFUSED	WSAECONNREFUSED
#endif
		//nothing to hear just yet...
		if (last_error == EWOULDBLOCK || last_error == ECONNREFUSED)
			return 0;

		//i get WSAEMSGSIZE when i connect to a server sometimes . . .

		//got a real error...
		con.print("recvfrom: got error %x\n", errno);
		return -1;
	}

//	con.print("listen: got packet from %s\n",inet_ntoa(their_addr.sin_addr));
//	con.print("source sockaddr_in:\n");
//	con.print("  family: %d\n",	their_addr.sin_family);
//	con.print("  port: %d\n",		their_addr.sin_port);
//	con.print("  sin_addr: %s\n", inet_ntoa(their_addr.sin_addr));

//	con.print("packet is %d bytes long\n",numbytes);
	if (numbytes < *destbufsize) destbuf[numbytes] = '\0';
//	con.print("packet contains \"%s\"\n",destbuf);

	*destbufsize = numbytes;

	return 1;
}

// ******************************** INET CLIENT-SIDE

SocketSendINET::SocketSendINET() {
	sending = 0;
	send_sockfd = -1;
}

SocketSendINET::~SocketSendINET() {
	if (isSending()) kill();
}

int SocketSendINET::isSending() { return sending; }

bool SocketSendINET::init(const char *hostname, int dest_port) {

	if (sending) {
		con.print("already connected\ntype 'kill_send' to disconnect\n");
		return false;
	}

	if (!hostname) hostname = "127.0.0.1";

	if (!dest_port) dest_port = MY_DEFAULT_PORT;

	//same with this one
    struct hostent *he = gethostbyname(hostname);
    if (!he) {
		con.print("gethostbyname: got error %x\n", errno);
        return false;
    }

	send_sockfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (send_sockfd == -1) {
		con.print("socket: got error %x\n", errno);
        return false;
    }

	send_dest_addr.sin_family = AF_INET;     // host byte order
	send_dest_addr.sin_port = htons(dest_port); // short, network byte order
	send_dest_addr.sin_addr = *((struct in_addr *)he->h_addr_list[0]);
	memset(&(send_dest_addr.sin_zero), '\0', 8); // zero the rest of the struct

	sending = 1;
	con.print("successfully sending!\n");

	//out the host data
	con.print("host info:\n");
	con.print("h_name: %s\n", he->h_name);
#ifdef WIN32
	for (char FAR * FAR *ch = he->h_aliases; *ch; ch++)
#else
	for (char **ch = he->h_aliases; *ch; ch++)
#endif
		con.print("h_alias: %s\n", ch);
	con.print("h_addrtype: %d\n", he->h_addrtype);
	con.print("h_length: %d\n", he->h_length);
	con.print("h_addr: %x\n", *(unsigned long *)he->h_addr_list[0]);

	return true;
}

bool SocketSendINET::kill() {
	if (sending) {
		closesocket(send_sockfd);
		sending = 0;
		con.print("end send session\n");
	} else {
		con.print("not sending\n");
		return false;
	}
	return true;
}

bool SocketSendINET::send(const char *msg, int length) {
	if (!sending) {
		con.print("not connected\ntype 'init_send'\n");
		return false;
	}

	if (!msg) return false;

	int numbytes = sendto(
		send_sockfd,
		msg,
		length,
		0,
        (struct sockaddr *)&send_dest_addr,
		sizeof(struct sockaddr));

    if (numbytes == -1) {
		con.print("sendto: got error %x\n", errno);
        return false;
    }

	if (numbytes < length) {
		con.print("sendto only sent %d of %d bytes\n", numbytes, length);
		return false;
	}

//	con.print("send: sent %d bytes to %s\n", numbytes, inet_ntoa(send_dest_addr.sin_addr));
    return true;
}

#endif // OMIT_NETWORKING

// ******************************** LOOPBACK SERVER-SIDE

class LoopbackBuffer {
public:
	class LoopbackEntry {
	public:
		char packet[1400];	//fix this up, replace the 16x1400 with a single size buffer
		int length;
	} ;

	LoopbackEntry msg[256];	//was 64...
	int start;
	int end;

	LoopbackBuffer() {
		start = 0;
		end = 0;
	}
};

static LoopbackBuffer *loopback[LOOPBACK_TYPE_COUNT] = {
	new LoopbackBuffer(),	//server
	new LoopbackBuffer()	//client
};

SocketListenLoopback::SocketListenLoopback() { type = -1; }
SocketListenLoopback::~SocketListenLoopback() {}

bool SocketListenLoopback::init(int type) {
	this->type = type;
//	con.print("SocketListenLoopback::init() %p ->type = %d\n", this, type);
	return true;
}

bool SocketListenLoopback::kill() { return true; }
int SocketListenLoopback::isListening() { return type >= 0; }

int SocketListenLoopback::recv(char *destbuf, int *destbufsize, sockaddr_in *destaddr) {
	assert(type >= 0 && type < LOOPBACK_TYPE_COUNT);
	LoopbackBuffer *lb = loopback[type];
	//got nothin
	if (lb->start == lb->end) return 0;

	int msgsize = lb->msg[lb->start].length;
	if (*destbufsize >= msgsize) {
		*destbufsize = msgsize;
	} else {
		//i havent coded up 'get half the packet' yet
		con.print("loopback recv overflow: buf size %d msg size %d\n", *destbufsize, msgsize);
		return -1;
	}

	memcpy(destbuf, lb->msg[lb->start].packet, msgsize);

	//increment and wrap
	lb->start = (lb->start + 1) % numberof(lb->msg);

	if (destaddr) memset(destaddr, 0, sizeof(*destaddr));

//	con.print("recv type %d msg: ", type, destbuf);
//	for (int i = 0; i < msgsize; i++) {
//		con.print("%2.x", destbuf[i]);
//	}
//	con.print("\n");

	return 1;
}

SocketSendLoopback::SocketSendLoopback() { type = -1; }
SocketSendLoopback::~SocketSendLoopback() {}

bool SocketSendLoopback::init(int type) {
	this->type = type;
	return true;
}

bool SocketSendLoopback::kill() { return true; }
int SocketSendLoopback::isSending() { return type >= 0; }

bool SocketSendLoopback::send(const char *msg, int length) {

	assert(type >= 0 && type < LOOPBACK_TYPE_COUNT);
	LoopbackBuffer *lb = loopback[1 - type];	//switch

	int next_loopback_end = (lb->end + 1) % numberof(lb->msg);
	if (next_loopback_end == lb->start) {
		//TODO - find what messages are filling the buffer up before we fully connect
		con << "SocketSendLoopback::send(): loopback buffer " << (1 - type) << " full" << endl;
		return false;
	}

	if (length > numberof(lb->msg->packet)) {
		con << "SocketSendLoopback::send(): tried to send a loopback packet too big" << endl;
		return false;
	}

	lb->msg[lb->end].length = length;
	memcpy(lb->msg[lb->end].packet, msg, length);
	lb->end = next_loopback_end;

//	con.print("send type %d msg: ", type);
//	for (int i = 0; i < length; i++) {
//		con.print("%2.x", msg[i]);
//	}
//	con.print("\n");

	return true;
}

// ******************************** GENERAL NETWORK COD3

static int network_init_d = 0;

bool networkSystemInit() {

	if (network_init_d) return true;	//already initialized - return true?

#ifndef OMIT_NETWORKING
#ifdef WIN32
	WSADATA wsaData;

	//attempt to init
	if (WSAStartup(MAKEWORD(1, 1), &wsaData)) {	//returns 0 on succeed
		con.print("WSAStartup failed!\n");
		return false;
	}
#endif
#endif //OMIT_NETWORKING

	network_init_d = true;

//	listener = new SocketListenINET();
//	if (!listener) return false;
//	sender = new SocketSendINET();
//	if (!sender) return false;

	return true;
}

void networkSystemShutdown() {
//	if (listener) delete listener;
//	if (sender) delete sender;
//	listener = NULL;
//	sender = NULL;

	if (network_init_d) {
#if !defined(OMIT_NETWORKING) && defined(WIN32)
		if (WSACleanup()) {
			con.print("WSACleanup failed");
		} else
#endif
		{
			network_init_d = 0;
		}
	}
}

class NetworkSys : public IModule {
public:
	virtual bool sysInit() { return networkSystemInit(); }
	virtual void sysKill() { networkSystemShutdown(); }
};
static NetworkSys g_networkSys;
IModule &networkSys = g_networkSys;
