/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <math.h>

#ifdef WIN32
#include <direct.h>
#define getcwd _getcwd
#endif

#include "system/singletons.h"
#include "system/exception.h"
#include "system/memory.h"

#include "base/base.h"
#include "base/cvar.h"
#include "base/filebuffer.h"

#include "spawnobj.h"
#include "animation.h"
#include "program.h"
#include "main.h"
#include "client.h"
#include "game/main.h"
#include "world.h"
#include "spawnobj.h"	//PF_spawnType
#include "nettables.h"	//PF_netTable, PF_registerEntClass
#include "filesys.h"

using namespace std;

//maybe we should put this in the table, and just give LuaThreads callbacks to the table?
static lua_State *mainState = NULL;

void lua_printbacktrace (lua_State *L) {
	assert(L);
	con << "Lua stack trace:\n";
	lua_Debug ar;
	memset(&ar, 0, sizeof(ar));
	for (int level = 0; lua_getstack(L, level, &ar); level++) {  /* check function at level */
		lua_getinfo(L, "Snl", &ar);  /* get info about it */
		con << "level " << level << "\n";
		con << "  event:       " << ar.event << "\n";
		con << "  name:        " << ar.name << "\n";
		con << "  namewhat:    " << ar.namewhat << "\n";
		con << "  what:        " << ar.what << "\n";
		con << "  source:      " << ar.source << "\n";
		con << "  currentline: " << ar.currentline << "\n";
		con << "  nups:        " << ar.nups << "\n";
		con << "  linedefined: " << ar.linedefined << "\n";
		con << "  lastlinedefined: " << ar.lastlinedefined << "\n";
		con << "  short_src:   " << ar.short_src << "\n";
		//    if (ar.currentline > 0) {  /* is there info? */
		//    con.print(L, "%s:%d: ", ar.short_src, ar.currentline);
		memset(&ar, 0, sizeof(ar));
	}
}



static void threadTable_hookFunc(lua_State *L, lua_Debug *ar);

class LuaThread {
public:
	lua_State *L;
	double 	elapsedTime;	//elapsed time in the spotlight - used for callbacks to determine how long we've been running
	double 	lastTime;	//the last read time for this thread.  set this when a thread starts/resumes
	int		hookCount;	//for kicks, the # of hook callbacks we got since running in our timeslice
	
	int index;
	
	//later:
	double sleepEndTime;// - if this is set above our app timer then we're still asleep and dont wake us - used with sleep()
	int waitingOn;// - if this is a valid index of a thread in the table then we're waiting for that thread to die before continuing - used with wait()
	
	LuaThread() : L(NULL), elapsedTime(0), lastTime(0), hookCount(0), sleepEndTime(0), waitingOn(-1), index(0xdeadbeef) {}
	
	//create right off the bat
//	LuaThread(const char *fn)
//	: L(NULL), elapsedTime(0), lastTime(0), hookCount(0), sleepEndTime(0), waitingOn(-1)
//	{ create(fn); }
	
	//creates a thread from the base state
	//and clears that new thread entry from the base state stack (where its pushed by default)
	//inits the new state to a specified function
	//leaves the function sitting on top of the stack
	//sets up hook callback
	//
	//the thread will be the function on top of src_state's stack
	//
	// src_state most likely won't be the mainState (where I make my threads from.  why do i do this? i dont know)
	bool create(lua_State *src_state) {
	
		if (!lua_isfunction(src_state, -1)) {
			con << "LuaThread::create - didn't get a function as a parameter!\n";
			return false;
		}
		
		L = lua_newthread(mainState);	//make a new thread
		
		//how are we gonna get rid of threads?
		//I'm gonna move it to the base of the stack 
		//because I think removing it makes GC tweak out
		//lua_insert(mainState, 1);			//and take it off the stack
		//lua_pop(mainState, 1);
		
		{	//[thread]
			lua_getglobal(mainState, "_threads");	//[thread] [table/nil]
			if (lua_isnil(mainState,-1)) {
				lua_pop(mainState,1);		//[thread]
				lua_newtable(mainState);	//[thread] [table]
				lua_setglobal(mainState, "_threads");	//[thread]
				lua_getglobal(mainState, "_threads");	//[thread] [table]
			}
			//[thread] [table]
			lua_pushinteger(mainState, lua_objlen(mainState, -1)+1);	//[thread] [table] [index]
			lua_insert(mainState, -3);	//[index] [thread] [table]
			lua_insert(mainState, -3);	//[table] [index] [thread]
			lua_rawset(mainState, -3);	//[table]
			lua_pop(mainState, 1);
		}

		//move the function from src_state to L
		lua_xmove(src_state, L, 1);	//from, to, n (why not to,from like mem/strcpy? )
		
		if (!lua_isfunction(L, -1)) {
			con << "LuaThread::create - couldnt find a function\n";
			lua_pop(L,1);
			return false;
		}
		
		//set hooks - only allow so many ops to run
		//lets think about this
		//if we get a PF_sleep then the thread lua_yield's
		//if we get a PF_wait then the thread lua_yield's
		//for what other reason do we want our thread to yield?
		//either way, currently all our line counter does is count the time
		//so why don't we insert a time counter instead?
		// . . . because lua doesn't have any time-based hooks . . .
		lua_sethook(L, threadTable_hookFunc, LUA_MASKCOUNT, 24);	// 8 instructions / loop
		return true;
	}

	//runs the function - called at the beginning of each timeslice
	//expects the function to be on top of the stack, as create() leaves it
	//prereq: create()
	bool run() {
		hookCount = 0;		//clear the callback ticker
		elapsedTime = 0;	//clear its timeslice timer
		lastTime = timer.getTime();	//set its last time for callbacks' sake

		//run it
		int status = lua_resume(L, 0);

		if (status == 0) {	//thread finished
			return false;
		}
		if (status == LUA_YIELD) {	//still going
			return true;
		}
		//else...
		con << "run func: lua_pcall got error " << status << "\n";
		string errorstr = lua_tostring(L, -1);
		con << " : " << errorstr << "\n";
		lua_pop(L, 1);
		
		lua_printbacktrace(L);
		return false;

	}
};

class LuaThreadTable {
	vector<LuaThread *> threads;
	int lastUpdated;	//the last updated thread
public: 
	
	LuaThreadTable() : lastUpdated(0) {}
	
	int addThread(LuaThread *th) {
		if (!th) return -1;		//me being lazy at my error handling
		threads.push_back(th);
		th->index = threads.size()-1;
		return threads.size()-1;
	}
	
//how long the table is to run
#define SCRIPT_DURATION		0.01		//10ms
//how long each thread in the table can run (max)
#define SCRIPT_THREAD_DURATION	0.003		//3ms
	void run() {

		if (isEmpty()) return;
	
		int threadCount = threads.size();
		
		double startTick = timer.getTime();
		double thisTick;
		for (lastUpdated %= threadCount;	//make sure we're in bounds incase the thread vector size changes
			(thisTick = timer.getTime()) < (startTick + SCRIPT_DURATION);
			lastUpdated = (lastUpdated+1) % threadCount)
		{
			LuaThread *t = threads[lastUpdated];

			//null entries are allowed for uniqueness of index in the thread table
			if (!t) continue;
			if (t->waitingOn != -1) continue;
			if (t->sleepEndTime >= game_thistime) continue;
			
			//EVEN WHEN OUR THREADS ARE SLEEPING THEY'RE TAKING FOREVER!!!!
			
				//here's where we run our thread

//			con.print("thread table running index %d\n", lastUpdated);
			//if this returns false then we are done with this thread
			if (!t->run()) {
				
				//destroy the thread
				//there is no official Lua destroy thread function - they just drift away with the rest of GC'd objs
				delete t;
				threads[lastUpdated] = NULL;
				
				//now find all waiting on this thread
				//and wake them
				bool found = false;
				for (int j = 0; j < threadCount; j++) {
					LuaThread *wt = threads[j];
					if (wt) {
						found = true;
						if (wt->waitingOn == lastUpdated) {
							wt->waitingOn = -1;
						}
					}
				}
				//if we finished a thread and didnt find any more, dont bother stay in the table loop
				if (!found) {
					break;
				}
			}
		}
	}
	
	LuaThread *find(lua_State *L) {
		for (int i = 0; i < threads.size(); i++) {
			LuaThread *t = threads[i];
			if (!t) continue;
			if (t->L == L) return t;
		}
		return NULL;
	}
	
	bool inbounds(int i) { return i >= 0 || i < threads.size(); }
	LuaThread *getByIndex(int i) { return inbounds(i) ? threads[i] : NULL; }
	LuaThread *operator[](int i) { return getByIndex(i); }
	
	bool isEmpty() {
		for (int i = 0; i < threads.size(); i++) {
			if (threads[i]) return false;
		}
		return true;
	}
	
	void clear() {
		VECTOR_DELETE_ALL(threads);
	}
};

static LuaThreadTable threadTable;

static void threadTable_hookFunc(lua_State *L, lua_Debug *ar) {
	//TODO - if this is hit upon any lua code called within C code called from a lua child thread
	//(which happens more often than you'd think)
	//then it will break.
	
	//SOLUTION: when we run a lua child thread -> c function (any PF_*** funcs) then
	//we should set a semaphore
	//if that semaphore is set here then dont yield!
	//DOWNSIDE: if it's calling a C function not in my engine (i.e. any of the builtin functions)
	//then it'll still crash
	
	//SOLUTION: if lua keeps track of c stack depth then, here, if thats set then dont yield
	//this is what I'm doing at the minute
	//DOWNSIDE: it requires modifying the lua library
	//by adding a function for determining whether a lua_yield will work or not
	//this is necessary because the lua_State structure is never exposed to the outside world...
	//... though I could redeclare it here ... I wonder if the linker would notice or care ...

	//1) lookup the lua state in our thread table
	//2) read its last time, subtract the current time, to calc the elapsed time
	//3) add that to its current time in the spotlight, set its last time
	//4) if its time in the spotlight exceeds our per-thread maximum then
	//		yield this thread
	//		and resume another one.  can we do that from a callback function like this?
	LuaThread *t = threadTable.find(L);
	assert(t);
	t->hookCount++;
	double thisTime = timer.getTime();
	double dt = thisTime - t->lastTime;
	t->elapsedTime += dt;

	//TODO - keep up to date with lua_yield's fail condition
	//since I don't think there's a better way to check (no lua_canyield or whatever)
	if (!lua_canyield(L)) {
//		con << "got a L->nCcalls yield stall\n";
		return;
	}

	//if we haven't finished our time then bail
	if (t->elapsedTime < SCRIPT_THREAD_DURATION) return;
	
	//if we've ate up all our time then yield this thread
	//from there, we'll return to the lua_pcall to continue on our merry way
	//otherwise we'll return to the lua code executer
	lua_yield(L, 0);
}

//TODO - use refs!

static const char *timeoutTableName = "_timeouts";

static std::string *loadWorldRequest = NULL;
static bool resetSessionRequest = false;
static bool resetGameRequest = false;

//my attempt at a library:

static int PF_fork(lua_State *L) {
	try {
		int threadIndex = -1;
		if (!lua_gettop(L)) throw Exception() << "didnt get a parameter!";

		lua_settop(L, 1);	//move our first param to the top
		//if its a string then try reading it from the global function table
		//convert it from a string into the main table entry .
		if (lua_isstring(L, 1)) lua_rawget(L, LUA_GLOBALSINDEX);
		if (!lua_isfunction(L, 1)) throw Exception() << "didn't find a function!";

		LuaThread *th = new LuaThread();
		if (!th) throw Exception() << "couldnt create thread object";
		if (!th->create(L)) throw Exception() << "wasnt able to fork to that function";

		threadIndex = threadTable.addThread(th);
		//con.print("PF_fork creating fork thread %d\n", threadIndex);

		lua_pushinteger(L, threadIndex);
		return 1;
	} catch (Exception &t) {
		con << "PF_fork error: " << t << "\n";
		//push -1 if we get a problem.  thtas the equiv of a bad thread.  i think that might session-kill PF_wait ...
		lua_pushinteger(L, -1);
	}
	return 1;
}

static int PF_wait(lua_State *L) {
	try {
		int threadIndex = 0;
		if (lua_gettop(L)) threadIndex = (int)lua_tonumber(L, 1);
		
		//current thread
		LuaThread *waiter = threadTable.find(L);
		if (!waiter) throw Exception() << "called by an unknown thread";

		//to be waited on thread
		if (!threadTable.inbounds(threadIndex)) {
			throw Exception() << "couldnt find thread " << threadIndex;
		}

		LuaThread *waitingOn = threadTable[threadIndex];
		if (!waitingOn) return 0;	//already done
		
	//	con.print("PF_wait %p waiting on index %d\n", waiter, (int)threadIndex);

		//ref by handle
		waiter->waitingOn = threadIndex;
		//tell lua to stop running as well
		return lua_yield(L, 0);	//return 0;
	} catch (Exception &t) {
		con << "PF_wait error: " << t << "\n";
		//here's another doozie error that tends to lock up our engine
		mainSys.sessionKill();
	}
	return 0;
}

static int PF_sleep(lua_State *L) {
	try {
		lua_Number endTime = 0;
		if (lua_gettop(L)) endTime = lua_tonumber(L, 1);
		endTime += game_thistime;
		
		//current thread
		LuaThread *sleeper = threadTable.find(L);
		if (!sleeper) throw Exception() << "called by an unknown thread";
		
	//	con.print("PF_sleep %p until %f\n", sleeper, endTime);

		sleeper->sleepEndTime = endTime;
		//tell lua to stop running as well
		return lua_yield(L, 0);	//return 0;
	} catch (Exception &t) {
		con << "PF_sleep error: " << t << "\n";
		//kill the session if this happens.
		//why?  because it should never ordinarily happen
		//and 9 times out of 10 that it does, it is associated with a lockup due to a loop in the lua script
		//killing the session will stopthe lua engine which will stop the lockup
		//how do we do that anyways?
		//mainSys.sessionKill();
		//though this destroys the lua state and then crashes everything... grr...
		lua_printbacktrace(L);
	}
	return 0;
}

//TURNS OUT this isnt the best
//but for the sake of it already being coded, i'll keep it around
//
//okay the story here is
//require() in lua
//if it pulls something from another directory
//it still wont change the working directory
//so what include here is going to do is
//2) store the working dir
//1) pick the dir change out of the path (look for /'s) and chdir
//3) run the require() call
//4) pop the working dir
static int PF_include(lua_State *L) {
	const char *str = lua_tostring(L,1);
	if (!str) {
		con << "PF_include expected a string!\n";
		return 0;
	}
	
	char require[4096];
	strncpy(require, str, sizeof(require));
	char srcrequire[4096];
	strncpy(srcrequire, require, sizeof(require));

	char olddir[4096];
	getcwd(olddir, sizeof(olddir));
	
	int changed = 0;
	int skipped = strlen(require) - 1;
	for (; skipped > 0 && require[skipped] != '/'; skipped--);
	if (require[skipped] == '/') {	//if we happened to run into a slash then save our working dir and set the new one
		skipped++;
		char newdir[4096];
		assert(skipped <= sizeof(newdir));
		memcpy(newdir, require, skipped);
		newdir[skipped] = '\0';
		chdir(newdir);	//how do we know if it works?
		changed = 1;
		
		//fix the module while you're here
		char newrequire[4096];
		strncpy(newrequire,  require+skipped, sizeof(newrequire));
		memcpy(require, newrequire, sizeof(require));
	}

	//now run the require() method
	//and how do we know how many args it returns?
	lua_getglobal(L, "require");	//require function
	lua_pushstring(L, require);		//require function, require string
	if (lua_pcall(L, 1, 1, 0)) con << "failed to include file " << srcrequire << "\n"
		<< "current working directory is " << olddir << "\n"
		<< " got errors:\n" << lua_tostring(L, -1) << "\n" << lua_tostring(L, -2);

	if (changed) chdir(olddir);
	
	return 1;
}

static int PF_bitand(lua_State *L) {
	int argc = lua_gettop(L);
	int param = (int)lua_tonumber(L,1);
	for (int i = 2; i <= argc; i++) {
		param &= (int)lua_tonumber(L,i);
	}
	lua_pushinteger(L,param);
	return 1;
}

static int PF_bitor(lua_State *L) {
	int argc = lua_gettop(L);
	int param = (int)lua_tonumber(L,1);
	for (int i = 2; i <= argc; i++) {
		param |= (int)lua_tonumber(L,i);
	}
	lua_pushinteger(L,param);
	return 1;
}

static int PF_bitxor(lua_State *L) {
	int argc = lua_gettop(L);
	int param = (int)lua_tonumber(L,1);
	for (int i = 2; i <= argc; i++) {
		param ^= (int)lua_tonumber(L,i);
	}
	lua_pushinteger(L,param);
	return 1;
}

static int PF_bitnot(lua_State *L) {
	lua_pushinteger(L, ~(int)lua_tonumber(L, 1));
	return 1;
}

static int PF_bitshiftleft(lua_State *L) {
	lua_pushinteger(L, (int)lua_tonumber(L,1) << (int)lua_tonumber(L,2));
	return 1;
}
static int PF_bitshiftright(lua_State *L) {
	lua_pushinteger(L, (int)lua_tonumber(L,1) >> (int)lua_tonumber(L,2));
	return 1;
}

static int PF_getTime		(lua_State *L) { lua_pushnumber(L, game_thistime ); return 1; }
static int PF_getDeltaTime	(lua_State *L) { lua_pushnumber(L, game_deltatime); return 1; }
static int PF_getLastTime	(lua_State *L) { lua_pushnumber(L, game_lasttime);	return 1; }
static int PF_getSysTime	(lua_State *L) { lua_pushnumber(L, system_thistime); return 1; }
static int PF_getSysDeltaTime(lua_State*L) { lua_pushnumber(L, system_deltatime); return 1; }
static int PF_getSysLastTime(lua_State *L) { lua_pushnumber(L, system_lasttime); return 1; }

static int PF_gameTimeScalar(lua_State *L) {
	int argc = lua_gettop(L);
	float oldtime = game_time_scalar;
	if (argc == 1) game_time_scalar = lua_tonumber(L, -1);
	lua_pushnumber(L, oldtime);
	return 1;
}

//void print(string)
static int PF_print(lua_State *L) {
	int argc = lua_gettop(L);
	ostringstream s;
	const char *sep = "";
	for (int i = 1; i <= argc; i++) {
		s << sep;
		sep = "\t";
		lua_getglobal(L, "tostring");
		lua_pushvalue(L, i);
		lua_pcall(L, 1, 1, 0);
		s << lua_tostring(L, -1);
		lua_pop(L, 1);
	}
	string str = s.str();
	con << str << "\n";
	return 0;
}

//the best i can do for variable access right now
static int PF_viewTileSize(lua_State *L) {
	lua_Number ret = viewTileSize;	//keep track of the old
	if (lua_gettop(L)) viewTileSize = lua_tonumber(L, 1);	//if so, set the new
	lua_pushnumber(L, ret);	//and return the old
	return 1;
}

static int PF_gameState(lua_State *L) {
	lua_Number ret = (lua_Number)gameState;	//keep track of the old
	if (lua_gettop(L)) gameState = (gamestate_t)(int)lua_tonumber(L, 1);	//if so, set the new
	lua_pushnumber(L, ret);	//and return the old
	return 1;
}

static int PF_gameFlags(lua_State *L) {
	lua_pushnumber(L, gameflags);	//and return the old
	return 1;
}

//atm we are mapping unique indices into functions to call back
//what we need to do is map unique indices into two-element tables:: 1: the function to callback, & 2: when to call it back

static int PF_setTimeout(lua_State *L) {
	if (lua_gettop(L) != 2) {
		con << "usage: PF_setTimeout(function, delay)\n";
		return 0;
	}
	
	if (!lua_isfunction(L, 1)) {
		con << "PF_setTimeout expected a function\n";
		return 0;
	}
	
	//grab the timeout from the top - for later
	lua_Number runtime = lua_tonumber(L, 2);
	runtime += game_thistime;
	
	lua_settop(L, 1);	//cut out all else but the function
	//stack: [function]
	
	//create a table w/elem 0 as the function and elem 1 as the timeout
	lua_newtable(L);			// [function] [table]
	lua_insert(L, 1);			// [table] [function]
	lua_rawseti(L, 1, 0);		// [table={0=>function}]
	lua_pushnumber(L, runtime);	// [table={0=>function}] [number:timeout]
	lua_rawseti(L, 1, 1);		// [table={0=>function, 1=>timeout}]
			
	//now get the master timeout table
	//if it doesnt exist then create it
	
	//so we're gonna make a new table off of globals for holding all our timeout function calls
	lua_getglobal(L, timeoutTableName);	// [new entry] [timeout master table]
	
	//if nothing's there then we create a new one
	if (!lua_istable(L, -1)) {
	
		//get that miscreated table off the stack!
		lua_pop(L, 1);						// [new entry]
		
		//and make a new one
		lua_newtable(L);				// [new entry] [new table]

		//this method pops our table, so we gotta put it back on the stack
		lua_setglobal(L, timeoutTableName);	//[new entry]
		
		//is the table still on top? nope - put it back there!
		lua_getglobal(L, timeoutTableName);	//[new entry] [timeout master table]

		//btw, what is the difference between lua_setfield and lua_settable ?
		//maybe setfield is a shortcut for pushing the key string onto the stack?
	}
			
	//now we push the stuff onto the master stack
	
	lua_insert(L, 1);					// [timeout table] [new entry] 
	
	//TODO - refs!
	static int id = 0;
	id++;
	lua_rawseti(L, 1, id);
	
	//con << "pushed a new timeout entry w/id " << id << "\n";
			
	//and have faith that that worked

	return 0;
}

static int PF_removeTimeout(lua_State *L) {
	con << "PF_removeTimeout not implemented yet!\n";
	//I hope this isn't being used quite yet ...
	return 0;
}

static int PF_resetGame(lua_State *L) {
	resetGameRequest = true;
	return 0;
}

static int PF_resetSession(lua_State *L) {
	resetSessionRequest = true;
	return 0;
}

static int PF_loadWorld(lua_State *L) {
	int argc = lua_gettop(L);
	string filename;
	ostringstream s;
	try {
		if (!argc) { //reload the current map
			World *world = worldSys.getWorld();
			if (!world) throw Exception() << "cannot reload the world -- none present at the moment";
			if (!world->filename.length()) throw Exception() << "cannot reload the world -- it has no associated filename.  It needs to be saved first.";
			filename = world->filename;
		} else {
			const char *_filename;
			if (!(_filename = lua_tostring(L, 1))) throw Exception() << "got a null string!";
			filename = _filename;
		}
		if (loadWorldRequest) delete loadWorldRequest;	//overwrite any old requests
		loadWorldRequest = new string(filename);		//use the pointer's existence as a flag
	} catch (Exception &t) {
		con << "PF_loadWorld failed: " << t << "\n";
		lua_pushboolean(L, false);
		return 1;
	}
	lua_pushboolean(L, true);
	return 1;
}

//the session variable map stores inter-level stuff inside itself.
//ironic name since its really session-transcendant
//since I named all my sub-session super-level stuff 'session'
//i need more distinction in later versions...
//	engine start (hw init)
//		game start (mod-specific stuff: init.lua, gui specific stuff)
//			profile start (key mappings, custom tags, etc)
//				session start (new/load game - game-state variables (below))
//					level start (clear entity arrays, init spawn objects, etc)
//right now I have this:
//	engine start
//		game start
//			session start (which is basically a level start)
static map<string,string> sessionVarMap;
static int PF_sessionVar(lua_State *L) {
	int argc = lua_gettop(L);
	try {
		if (argc == 0) throw Exception() << "expected a key, maybe a value after that";
		const char *key = lua_tostring(L,1);
		if (!key) throw Exception() << "your key couldnt turn into a string";
		if (argc > 1) {	//we have a value.  one value.  no more than one. for now.
			if (lua_isnil(L,2)) {
				con << "clearing spawn var " << key << "\n";
				map<string,string>::iterator it = sessionVarMap.find(key);
				if (it == sessionVarMap.end()) return 0;	//nil if there was no var
				lua_pushstring(L, it->second.c_str());	//put it on the stck in case the returner wants a copy
				sessionVarMap.erase(it);	//and clear it
				return 1;
			} else {
				const char *value = lua_tostring(L,2);
				if (!value) throw Exception() << "your value couldn't turn into a string";
				sessionVarMap[key] = value;
				lua_pushstring(L, value);
				return 1;		//return the value we set
			}
		}
		map<string,string>::iterator it = sessionVarMap.find(key);
		if (it == sessionVarMap.end()) return 0;	//nil if there was no var
		lua_pushstring(L, it->second.c_str());
		return 1;	//var otherwise
	} catch (Exception &t) {
		con << "PF_sessionVar error: " << t << "\n";
	}
	return 0;
}

//zero this on session init
//use this for looking back and forth with ents and their classes
//so we only need to transport a single variable through networks
static int nextEntClassID = 0;
static CVarw cv_nextEntClassID(&nextEntClassID, "nextEntClassID");

static int PF_cvarExec(lua_State *L) {
	const char *cmd = lua_tostring(L, 1);
	if (cmd) cvarSys->exec((char *)cmd);
	return 0;
}

//int spriteLoad(string filename)
static int PF_spriteLoad(lua_State *L) {
	int argc = lua_gettop(L);
	if (argc != 1) {
		con.print("usage: PF_spriteLoad(filename)\n");
		return 0;
	}
	TexPtr tex;
	const char *fn = lua_tostring(L,1);
	string result;
	if (!fileSys.search(fn, game.getSpritePaths(), result)) {
		con << "PF_spriteLoad " << fn << " failed to find file" << endl;
	} else {
		tex = texSys.load(result.c_str(), TEXTURE_SPRITE, 0);		//i dont know what other flag we'd load them under ...
	}
	lua_pushinteger(L, tex.i);
	return 1;
}

static void CFunc_lua(int argc, char** argv) {
	//gotta add quotes for spaces in arguments to cvar system . . .
	if (argc < 2) return;
	char *buffer = argv[1];
	lua_State *L = mainState;
	int top = -1;
	
	try {
		if (!L) throw Exception() << "couldn't find the lua state";
		top = lua_gettop(L);
		int err = 0;
		if (err = luaL_loadbuffer(L, buffer, strlen(buffer), "script")) {
			throw Exception() << "for including script code:\n" << buffer << "\n\n"
				<< " we got some errors:\n"
				<< lua_tostring(L, -1) << "\n"
				<< lua_tostring(L, -2) << "\n";
		}
		if (lua_pcall(L, 0, 0, 0)) {
			throw Exception() << "lua got error\n"
				<< " : " << lua_tostring(L, -1) << "\n";	//display the error
		}
	} catch (Exception &t) {
		con << "CFunc_lua got error: " << t << "\n";
		if (L) lua_printbacktrace(L);
	}

	if (top != -1) lua_settop(L, top);
}
static CVarw cv_lua(0, "lua", CFunc_lua);

//game-independent
static const luaL_reg frameworkGlueFuncs_[] = {
//base class stuff
	{"print",				PF_print},					//void print(string)
	{"cvarExec",			PF_cvarExec},				//no way to return values, just exec
//binary operations
	//TODO - bit
	{"bitand",				PF_bitand},
	{"bitor",				PF_bitor},
	{"bitxor",				PF_bitxor},
	{"bitnot",				PF_bitnot},
	{"bitshiftleft",		PF_bitshiftleft},
	{"bitshiftright",		PF_bitshiftright},

	{NULL, NULL}
};
//cuz c isn't smart
const luaL_reg *frameworkGlueFuncs = frameworkGlueFuncs_;

//game-dependent
static const luaL_reg engineGlueFuncs_[] = {

//multithread
	//TODO - namespace thread
	{"fork",				PF_fork},					//thread fork( void(*)() )
	{"wait",				PF_wait},					//void wait(thread)
	{"sleep",				PF_sleep},					//void sleep(float)
	{"setTimeout",			PF_setTimeout},				//void setTimeout( void(*)(), float )	-- our single-thread fork alternative
	{"removeTimeout",		PF_removeTimeout},			//void removeTimeout( void(*)() ) -- and its alternative
//require ammendment
//	{"include",				PF_include},				//[whatever require returns] include(string fn)
//engine calls
	//TODO - timer
	{"getTime",				PF_getTime},				//float getTime()
	{"getDeltaTime",		PF_getDeltaTime},			//float getDeltaTime()
	{"getLastTime",			PF_getLastTime},			//float getLastTime()
	{"getSysTime",			PF_getSysTime},
	{"getSysDeltaTime",		PF_getSysDeltaTime},
	{"getSysLastTime",		PF_getSysLastTime},
	{"gameTimeScalar",		PF_gameTimeScalar},
	//TODO - view
	{"viewTileSize",		PF_viewTileSize},			//float viewTileSize([float newvalue])
	{"gameState",			PF_gameState},				//enum too?
	{"gameFlags",			PF_gameFlags},				//enum too?
//spawn opeartions
	//TODO - spawnobj?  or abstract them completely so it doesn't matter?  just use lua refs?
	{"spawnType",			PF_spawnType},
	{"spawnObjMetaNewIndex",PF_spawnObjMetaNewIndex},
	{"spawnObjMetaIndex",	PF_spawnObjMetaIndex},
//world
	//TODO - game
	{"resetGame",			PF_resetGame},
	{"resetSession",		PF_resetSession},
	{"loadWorld",			PF_loadWorld},
//supersession vars
	//TODO - session
	{"sessionVar",			PF_sessionVar},
//animation
	//TODO - anim
	{"animLoad",			PF_animLoad},//animation.h
//sprites
	//TODO - sprite
	{"spriteLoad",			PF_spriteLoad},
//net tables starting up
	//TODO - net
	{"netTableCreate",		PF_netTableCreate},	//nettable.h
	{"netTable",			PF_netTable},	//nettable.h
	{"registerEntClass",	PF_registerEntClass},	//nettable.h
//client input
	{"getClientKeyValue",	PF_getClientKeyValue},	//client.h
	{NULL, NULL}
};
const luaL_reg *engineGlueFuncs = engineGlueFuncs_;

void luaStateAddFuncs(lua_State *L, const luaL_reg *reg) {
	for (; reg->func; reg++) lua_register(L, reg->name, reg->func);
}

//currently only referenced by progSys.sessionInit
//this means the mod will have been loaded already
//this means its safe to call a game api function to load its predef functions
void progBuildPredefined() {
	luaL_openlibs(mainState);

	luaStateAddFuncs(mainState, frameworkGlueFuncs);
	luaStateAddFuncs(mainState, engineGlueFuncs);

	//these methods will gain access to the mainState
	//by teh progState() callback
	game.progAddGameLevelPredefined(mainState);
	game.progAddSessionLevelPredefined(mainState);
	spawnObjSys.addScriptGlobals();
}

//stop all threads ...
static void progStop() {
	threadTable.clear();
}

static int panic(lua_State *L) {
	con.print("Lua Paniced!\n");
	con.print("%s\n", lua_tostring(L, -1));
	con.print("%s\n", lua_tostring(L, -2));
	return 0;
}

//run all base level code, and thus register all functions and what not in the global table
static bool runbase(lua_State *L) {
	bool result = true;
	
	//begin rip from lua.exe: - i guess this is for the callback for pcall?
	int base = lua_gettop(L);  /* function index */
	lua_pushliteral(L, "_TRACEBACK");
	lua_rawget(L, LUA_GLOBALSINDEX);  /* get traceback function */
	lua_insert(L, base);  /* put it under chunk and args */
	
	if (lua_pcall(L, 0, 0, 0)) {
		con << "lua_pcall got error during runbase\n";
		con << " : " << lua_tostring(L, -1) << "\n";	//display the error
		lua_pop(L, 1);	//pop the error
		lua_printbacktrace(L);
		result = false;
	}

	lua_remove(L, base);  /* remove traceback function */
	//end rip
	
	return result;
}

class ProgSysImpl {
public:
	bool sysInit();
	void sysKill();
	bool gameInit();
	bool sessionInit();
	void sessionKill();

	lua_State *getState();
	bool compile();
	void update();
};

bool ProgSysImpl::sysInit() { threadTable.clear(); return true; }
void ProgSysImpl::sysKill() { threadTable.clear(); }

bool ProgSysImpl::gameInit() { sessionVarMap.clear(); return true; }

bool ProgSysImpl::sessionInit() {
	cout << "ProgSysImpl::sessionInit" << endl;
	
	//reset the ent classes
	nextEntClassID = 0;
	
	//reconstruct the main state here
	mainState = lua_open();
	if (!mainState) return false;
	//maybe the main state should only be created at the session?
	lua_atpanic(mainState, panic);

	progStop();//1) stop all threads
	
	progBuildPredefined();	//2) re-register any predefined functions
	
	//now compile our code 
	if (!progSys.compile()) return false;
	
	//and the ol' faithful, run-base to register all our base-level functions
	if (!runbase(mainState)) return false;
	
	return true;
}
	
void ProgSysImpl::sessionKill() {
	progStop();//1) stop all threads
	
	if (mainState) {
		lua_close(mainState);
		mainState = NULL;
	}
}

lua_State *ProgSysImpl::getState() { return mainState; }
		
bool ProgSysImpl::compile() {
	lua_State *L = mainState;
	
	ostringstream ss;
	
	World *world = worldSys.getWorld();
	assert(world);
	
	//create our init lua vars:
	ss << "MAPTYPE = {}\n";
	for (int i = 0; i < mapTypes.size(); i++) {
		ss << "\tMAPTYPE." << mapTypes[i] << " = " << i << "\n";
	}

	//append to package.path the luaPaths vector
	for (int i = 0; i < game.getLuaPaths().size(); i++) {
		ss << "package.path = '" << game.getLuaPaths()[i] << "?.lua;'..package.path\n";
	}

	//start with the base files:
	for (int i = 0; i < game.getLuaBaseIncludes().size(); i++) {
		ss << "require(\"" << game.getLuaBaseIncludes()[i] << "\")\n";
	}

	//and add all our custom files
	char linebuf[256];
	for (int i = 0; worldSys.getScriptFileName(i, linebuf, sizeof(linebuf)); i++) {
		ss << "require(\"" << linebuf << "\")\n";
	}
	//for some odd reason without moving our stringstream into a separate stream
	//it tends to get garbled before we can use it ...
	string sstr = ss.str();
	const char *buffer = sstr.c_str();

	int err = 0;
	if (err = luaL_loadbuffer(L, buffer, strlen(buffer), "script")) {
		con << "for including script code:\n" << buffer << "\n\n";
		con << " we got some errors:\n";
		con.print("%s\n", lua_tostring(L, -1));
		con.print("%s\n", lua_tostring(L, -2));
		return false;
	}
	
	return true;
}

//if it needs this for updating threads or what not
void ProgSysImpl::update() {
	PROFILE();

	if (!mainState) return;	//we probably shouldn't be updating

	if (!lua_checkstack(mainState, 100)) {
		//for whatever its worth ...
		con.print("lua_checkstack failed!\n");
		mainSys.sessionKill();
		return;
	}

	threadTable.run();
	
	//now we run the timeout tables, if any are valid
	{
		lua_State *L = mainState;
		lua_getglobal(L, timeoutTableName);
		int t = lua_gettop(L);	//get our table's index in the stack
		if (!lua_istable(L, -1)) {
			//con << "timeout table is non-table: " << lua_typename(L, lua_type(L,-1)) << "\n";
			lua_pop(L, 1);
			//lua_pushnil(L);
			//lua_setglobal(L, timeoutTableName);
		} else {
		
			//keep track of all the functions we gotta delete
			//cap it though - instead of doing dyn alloc per frame
			int rememberToDelete[256];
			int rememberCount = 0;
		
			//[timeout table]
			lua_pushnil(L);					//[timeout table] [nil]
			while (lua_next(L, t) != 0) {
				// `key' is at index -2 and `value' at index -1
				// @t [timeout table] [key] [value]
				
				bool deleteThis = false;
				
				if (!lua_istable(L, -1)) {
					con << "FOUND A NON-TABLE OBJECT IN OUR TIMEOUT TABLE!\n";
					deleteThis = true;
				} else {
				
					// now get elem 1 of the current value -- that'll be the next execution time
					lua_rawgeti(L, -1, 1);					// @t [timeout table] [key] [entry] [timeout time]
					lua_Number nexttime = lua_tonumber(L, -1);
					lua_pop(L, 1);							// @t [timeout table] [key] [entry]

					if (nexttime <= game_thistime) {
					
						//con << "calling timeout index " << (int)lua_tonumber(L, -2) << "\n";
						
						//better execute it
						lua_rawgeti(L, -1, 0);				//@t [timeout] [key] [entry] [timeout function]
						int status = lua_pcall(L, 0, 0, 0);	//@t [timeout] [key] [entry] 
						if (status) {
							con.print("timeout execution: lua_pcall got error %d\n");
							con.print(" : %s\n", lua_tostring(L, -1));	//display the error
							lua_pop(L, 1);	//pop the error
						}
						//@t [timeout] [key] [entry] 
						
						deleteThis = true;
					}
				}
				
				if (deleteThis) {
					
					if (rememberCount < numberof(rememberToDelete)) {
						rememberToDelete[rememberCount] = (int)lua_tonumber(L, -2);	//convert the key to a lua number
						rememberCount++;
					} else {
						con << "timeout delete list overflow!\n";
					}
				}
				
				lua_pop(L, 1);  /* removes `value'; keeps `key' for next iteration */
			}
			
			//now we just have the table on top
			//now we go through and remove what of it we need to
			
			for (int i = 0; i < rememberCount; i++) {
				int id = rememberToDelete[i];
				//con << "removing timeout index " << id << "\n";
				//looks like this doesn't remove entries from the table
				//TODO - refs!
				lua_pushnil(L);
				lua_rawseti(L, -2, id);
			}
			
			//if (rememberCount) con << "done removing!\n";
			
			lua_pop(L, 1);	//now get rid of the table
		}
	}
	
	
	//finally handle any load level requests we mightve got:
	if (loadWorldRequest) {
		if (!worldSys.loadWorld(loadWorldRequest->c_str())) {
			con << "PF_loadWorld request failed to load the map " << *loadWorldRequest << "\n";
		} else {
			//now session reset doesn't reload init.lua
			//but it does reload the map code, right?
			mainSys.sessionReset();
		}
		delete loadWorldRequest;
		loadWorldRequest = NULL;
	} else if (resetGameRequest) {
		resetGameRequest = false;
		mainSys.gameReset();
	} else if (resetSessionRequest) {
		resetSessionRequest = false;
		mainSys.sessionReset();
	}
}

ProgSys::ProgSys() : impl(new ProgSysImpl()) {}
ProgSys::~ProgSys() { delete impl; }
bool ProgSys::sysInit() { return impl->sysInit(); }
void ProgSys::sysKill() { impl->sysKill(); }
bool ProgSys::gameInit() { return impl->gameInit(); }
bool ProgSys::sessionInit() { return impl->sessionInit(); }
void ProgSys::sessionKill() { impl->sessionKill(); }
lua_State *ProgSys::getState() { return impl->getState(); }
bool ProgSys::compile() { return impl->compile(); }
void ProgSys::update() { return impl->update(); }
ProgSys progSys;
