/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <iostream>
#include <sstream>
#include <map>

#include "defs/defs.h"
#include "math/vec.h"
#include "system/exception.h"
#include "base/opengl.h"
#include "base/filebuffer.h"
#include "program.h"
#include "shader.h"
#include "game/programmacros.h"

using namespace std;

ShaderObj::ShaderObj(GLhandleARB _progh, GLhandleARB _verth, GLhandleARB _fragh, const std::string &_name)
: name(_name), progh(_progh), verth(_verth), fragh(_fragh),
timeloc((GLint)-1),
spritetexloc((GLint)-1),
lighttexloc((GLint)-1),
bumptexloc((GLint)-1),
lastscreentexloc((GLint)-1),
currentscreentexloc((GLint)-1)
{
	//keep track of our uniform locations - or if they exist at all
	//i'm trusting the shaders to give me -1 if they don't
	//and i'm really trusting (performancewise) our shaders to optimize these out if the aren't used
	//i'll probably make a bigger standard set later (i.e. dot3 basis, flip, light origins, etc)
	timeloc = glGetUniformLocationARB(progh, "time");
	spritetexloc = glGetUniformLocationARB(progh, "spritetex");
	lighttexloc = glGetUniformLocationARB(progh, "lighttex");
	bumptexloc = glGetUniformLocationARB(progh, "bumptex");
	lastscreentexloc = glGetUniformLocationARB(progh, "lastscreentex");
	currentscreentexloc = glGetUniformLocationARB(progh, "currentscreentex");
}

ShaderObj::~ShaderObj() {	//deleteobject the handles
	if (progh) glDeleteObjectARB(progh);
	if (verth) glDeleteObjectARB(progh);
	if (fragh) glDeleteObjectARB(progh);
}


//it's so tempting to make use of my Shader class and Program class and everything
//but this cut from glush is so tempting ...
//i'm just going to stick with it for now

//a vector of all objects to-be-deleted on session kill
static vector<ShaderObj*> shaderObjs;

//tiles can be tied to shaders, but how to ref them?
//they'll have a string.  script'll do createShader with a param as "name"
//and we will match the two up
typedef map<string,int> shaderMap_t;
static shaderMap_t shaderIndexForName;

bool ShaderSys::sysInit() {
	//this becomes another singleton wrapper
	//one with a slightly abiguous name
	//'shaderSystem' is the object defined in my libs/shader/shaderext.cpp file
	//	for init/killing the shader extension
	//in that dir are also some wrappers for other shader stuff
	//but as long as i have all the functionality here ... meh, why bother?
	shaderSystem.init();	//atm it always returns true
	//todo - in draw.init or something have us check for gl extensions
	//if we don't have any then change the renderMethod to 1
	return true;	//if it fails then just fall back on different render pathways
}
	
bool ShaderSys::sessionInit() {
	shaderObjs.push_back(NULL);	//the first one will be null - to correspond with clear/zero indexes
	shaderIndexForName.clear();
	return true;
}

void ShaderSys::sessionKill() {
	//progs are inserted before vtx & frag shaders, so deleting front-to-back will cause progs to unlink their frags/vtx before those are individually deleted
	VECTOR_DELETE_ALL(shaderObjs);
	shaderIndexForName.clear();
}

const ShaderObj *ShaderSys::getIndex(int index) {
	if (index < 0 || index >= shaderObjs.size()) return NULL;
	return shaderObjs[index];
}

int ShaderSys::getIndexForName(const char *name) {
	shaderMap_t::iterator it = shaderIndexForName.find(name);
	if (it == shaderIndexForName.end()) return 0;
	return it->second;
}

ShaderSys shaderSys;

//windows sucks
static const char *safestr(const char *str) { return str ? str : "(null)"; }


static bool printHandleLog(GLhandleARB handle, const char *code = NULL) {

	GLsizei charsWritten  = 0;
	char *infoLog;

	GLint infologLength = 0;
	glGetObjectParameterivARB(handle, GL_OBJECT_INFO_LOG_LENGTH_ARB, &infologLength);

	if (infologLength <= 0) return true;

	infoLog = new char[infologLength];
	if (!infoLog) {
		con << "failed to alloc text buffer!\n";
		return false;
	}

	glGetInfoLogARB(handle, infologLength, &charsWritten, infoLog);
	con << "log:\n";
	con << infoLog << "\n";
	
	if (code) {
		con << "shader code was\n";
		con << code << "\n";
	}

	delete[] infoLog;
	return true;
}

int PF_createShader(lua_State *L) {
	int argc = lua_gettop(L);
	GLhandleARB progh = 0, verth = 0, fragh = 0;
	string shaderName;
	try {
		if (argc < 1) throw Exception() << "expected a table with vertexFile and fragmentFile params";
		
		lua_getfield(L,1,"name");
		if (!lua_isnil(L, -1)) shaderName = lua_tostring(L,-1);
		lua_pop(L,1);

		//start off by making a program ...
		//on exception we should delete it you know
		progh = glCreateProgramObjectARB();
		
		{
			ostringstream ss;

			lua_getfield(L, 1, "vertexCode");
			if (!lua_isnil(L,-1)) ss << safestr(lua_tostring(L, -1)) << endl;
			lua_pop(L,1);

			lua_getfield(L, 1, "vertexFile");
			if (!lua_isnil(L,-1)) {
				const char *fn = lua_tostring(L, -1);
				FileBuffer fb(fn);
				if (fb.getData()) ss << fb.getData();
			}
			lua_pop(L,1);
			
			string s = ss.str();
			const char *str = s.c_str();
			if (str && strlen(str) > 1) {
				verth = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
				glShaderSourceARB(verth, 1, &str, NULL);
				glCompileShaderARB(verth);
				glAttachObjectARB(progh, verth);
				printHandleLog(verth, str);
			}
		}

		{
			ostringstream ss;

			lua_getfield(L, 1, "fragmentCode");
			if (!lua_isnil(L,-1)) ss << safestr(lua_tostring(L, -1)) << endl;
			lua_pop(L,1);

			lua_getfield(L, 1, "fragmentFile");
			if (!lua_isnil(L,-1)) {
				const char *fn = lua_tostring(L, -1);
				FileBuffer fb(fn);
				if (fb.getData()) ss << fb.getData();
			}
			lua_pop(L,1);
			
			string s = ss.str();
			const char *str = s.c_str();
			if (str && strlen(str) > 1) {
				fragh = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
				glShaderSourceARB(fragh, 1, &str, NULL);
				glCompileShaderARB(fragh);
				glAttachObjectARB(progh, fragh);
				printHandleLog(fragh, str);
			}
		}
		
		glLinkProgramARB(progh);
		printHandleLog(progh);

		int err;
		if (err = glGetError()) fprintf(stderr, "error: %d\n", err);
		
		//here - go through all uniforms
		//get their type
		//if they are a sampler
		//	do a lua lookup on the global with the same name
		//		if it exists (as a texture reference?  i could save some special gl info on the id type ... )
		//			cycle through all mtex units and see if it is bound anywhere
		//				if it is then set the uniform to that unit #

		lua_getfield(L, 1, "uniforms");
		if (lua_istable(L, -1)) {
			int t = lua_gettop(L);
			//now cycle across the entries of the table
			// table is in the stack at index 't'
			lua_pushnil(L);  // first key
			while (lua_next(L, t) != 0) {
				// 'key' is at index -2 and 'value' at index -1
				//fprintf(stderr, "%s - %s\n", lua_typename(L, lua_type(L, -2)), lua_typename(L, lua_type(L, -1)));
				const char *name = lua_tostring(L, -2);
				//const char *value = lua_tostring(L, -1);
				//fprintf(stderr, "setting uniform %s = %s\n", name, value);
				
				//setup the program
				glUseProgramObjectARB(progh);
				
				//get the uniform location
				int loc = glGetUniformLocationARB(progh, name);
				if (loc == -1) {
					con << "unable to find uniform " << safestr(name) << "\n";
				} else {
				
					//get info on it
					GLsizei length = 0;
					GLsizei size = 0;
					GLenum type = 0;
					glGetActiveUniformARB(progh, loc, 0, &length, &size, &type, 0);
					switch(type) {
						case GL_INT:
						case GL_SAMPLER_1D:
						case GL_SAMPLER_2D:
						case GL_SAMPLER_3D:
						case GL_SAMPLER_CUBE:
						case GL_SAMPLER_1D_SHADOW:
						case GL_SAMPLER_2D_SHADOW:
							{
								int num = (int)lua_tonumber(L,-1);
//								fprintf(stderr, "setting param %s loc %d to %d\n", name, loc, num);
								glUniform1iARB(loc, num);
							}
							break;
							
						case GL_FLOAT:
//							fprintf(stderr, "setting param %s loc %d to %f\n", name, loc, lua_tonumber(L,-1));
							glUniform1fARB(loc, lua_tonumber(L,-1));
							break;

						case GL_FLOAT_MAT2:
							{
								if (!lua_istable(L,-1)) throw Exception() << "mat2 parameter expected a lua table ... \n";
								vec4f v = luaReadVec4f(L);
//								fprintf(stderr, "setting param %s loc %d to %f %f %f %f\n", name, loc, VEC4ELEM(v));
								glUniformMatrix2fvARB(loc, 1, GL_FALSE, v);
							}
							break;
							
						case GL_FLOAT_VEC2:
							{
								if (!lua_istable(L,-1)) throw Exception() << "vec2 parameter expected a lua table ... \n";
								vec2f v = luaReadVec2f(L);
//								fprintf(stderr, "setting param %s loc %d to %f %f\n", name, loc, VEC2ELEM(v));
								glUniform2fvARB(loc, 2, v);
							}
							break;
						case GL_FLOAT_VEC3:
							{
								if (!lua_istable(L,-1)) throw Exception() << "vec3 parameter expected a lua table ... \n";
								vec3f v = luaReadVec3f(L);
//								fprintf(stderr, "setting param %s loc %d to %f %f %f\n", name, loc, VEC3ELEM(v));
								glUniform3fvARB(loc, 4, v);
							}
							break;
						case GL_FLOAT_VEC4:
							{
								if (!lua_istable(L,-1)) throw Exception() << "vec4 parameter expected a lua table ... \n";
								vec4f v = luaReadVec4f(L);
//								fprintf(stderr, "setting param %s loc %d to %f %f %f %f\n", name, loc, VEC4ELEM(v));
								glUniform4fvARB(loc, 4, v);
							}
							break;
						case GL_INT_VEC2:
						case GL_INT_VEC3:
						case GL_INT_VEC4:
						case GL_BOOL:
						case GL_BOOL_VEC2:
						case GL_BOOL_VEC3:
						case GL_BOOL_VEC4:
						case GL_FLOAT_MAT3:
						case GL_FLOAT_MAT4:
							con << "tried to set uniform " << safestr(name) << " loc " << loc << " of an unhandled type: " << type << "\n";
							break;
						default:
							con << "tried to set a uniform of an unknown type\n";
							break;
					};
					if (err = glGetError()) con << "error: " << err << "\n";
				}
				glUseProgramObjectARB(0);
				
				lua_pop(L, 1);  // removes 'value'; keeps 'key' for next iteration
			}
		}
		lua_pop(L,1);
		
		if (err = glGetError()) fprintf(stderr, "error: %d\n", err);
		
		//store our shaders
		ShaderObj *so = new ShaderObj(progh,fragh,verth, shaderName);	//grab the timeloc
		int index = shaderObjs.size();
		shaderObjs.push_back(so);
		//map it by name
		if (shaderName.length()) shaderIndexForName[shaderName] = index;

		lua_pushinteger(L, index);	//push the shader index. this'll be our ref from the outside world
		return 1;
	} catch (Exception &t) {
		con << "createShader: " << t << "\n";

		//delete any created objects
		if (progh) glDeleteObjectARB(progh);
		if (fragh) glDeleteObjectARB(fragh);
		if (verth) glDeleteObjectARB(verth);
		
		return 0;
	}
	return 0;
}